/*
 *  cmpile.c(pp) - converts script into bytecode
 *
 *  Yazoo
 *  Copyright (C) 2005 Brian C. Ross
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of the License, or any later version.

 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */


#include <float.h>
#include <math.h>
#include "yazooh.h"
#include "cmpile.h"
#include "hobbsh.h"
#include "extrnl.h"


// global variables

yz_compile_global_struct yz_compile_globals;



//   ************** Translation routines ***********


// InitCompile() creates linked lists used by the compiler; call this before running Compile()

parse_retrn InitCompile(Ulong names_num)
{
    LL_retrn linkedLretrn;
    
    linkedLretrn = FNewLinkedList(&AllUserNames, names_num, sizeof(void *), 100, uncleared);
    if (linkedLretrn != passed)  return memory_err;

    linkedLretrn = FNewLinkedList(&NewlinePositions, 0, 2*sizeof(Ulong), 100, uncleared);
    if (linkedLretrn != passed)  {  FDeleteLinkedList(&AllUserNames);  return memory_err;  }

    NullNamesNum = 0;
    doublesize_in_longs = (int) ceil(1.*sizeof(double)/sizeof(long));
    
    return passed;
}


// FinishCompile() frees up memory used by the compiler; call this after running Compile()

void FinishCompile()
{
    Ulong counter;
    
    if (AllUserNames.memory == NULL)  return;
    
    for (counter = 1; counter <= AllUserNames.elementnum; counter++)    {
//        if (*(const char **) LL_Ptr(&AllUserNames, counter) != empty_str)   {
        free(*(void **) LL_Ptr(&AllUserNames, counter));
    }//   }
}


// Compile():  translates a script (text) into bytecode.  Expects a null character at the end (eof).
// This routine calls all the subsequent functions in this file.

parse_retrn Compile(char **scanPtr, linkedlist *CodeLLPtr)
{
    LL_retrn linkedLretrn;
    parse_retrn retrn;
    Ulong CodeElement = 1;

    errPtr = *scanPtr;    // just in case we somehow try to read errPtr without its having been set


        // set up memory
        
    linkedLretrn = FNewLinkedList(CodeLLPtr, 0, sizeof(Slong), 100, uncleared);
    if (linkedLretrn != passed)  return memory_err;
    

        // call the main parsing loop
        
    DocStartPtr = *scanPtr;
    retrn = ParseStatements(scanPtr, CodeLLPtr, &CodeElement, not_a_Ccmd, 0);
    if (retrn != passed)  return retrn;


        // add a null terminating character to the bytecode
        
    retrn = InsertCodeWord(CodeLLPtr, &CodeElement, null_cmd);      // to signify the end
    if (retrn != passed)  {  errPtr = *scanPtr;  return retrn;  }


        // tidy up loose ends

    if (lettertype(**scanPtr) != a_null)   {
        errPtr = *scanPtr;
        return eof_expected_err;    }
        
    linkedLretrn = FDefragmentLinkedList(CodeLLPtr);     // must be defragmented for the interpreter
    if (linkedLretrn != passed)   {
        errPtr = *scanPtr;
        return memory_err;   }
    
    return passed;
}


// ParseStatements() reads in a list of commands (composed of symbols, functions, constants, etc.) and generates bytecode from it.
// Specifically, it iteratively steps into each line and calls ParseExpression() to expand the expression.
// Expands if/while/do/for constructs.
// Stops at the end of the code block -- either the end of the file, or a closing brace '}'.
// Note that ParseExpression() recursively calls ParseStatements() to translate the contents of {...} and f(...).

parse_retrn ParseStatements(char **scanPtr, linkedlist *CodeLLPtr, Ulong *CodeElement, Ulong prev_Ccmd, Ulong prev_conditional_jump)
{
    parse_retrn retrn;
    Ulong control_cmd, FirstCmdElement;
    Slong first_cmd, first_flags, DG_Flags;


        // each iteration of the main loop has to start at the beginning character of the line
    
    retrn = get_start_of_line(scanPtr, CodeLLPtr, CodeElement);
    if (retrn != passed)  return retrn;
    
    
        // main loop:  iterates over each sentence, until we hit the end of the code block

    while (if_expression_terminator(**scanPtr) == yz_false)     {
    
        Ulong if_alwaysJumpElement = 0, if_condJumpElement = 0;      // the two locations where jump statements may be inserted
        char *LineBeginningPtr, *ForwardPtr;


            // update the NewlinePositions[] array (which allows runtime errors to be traced back to the original script)
        
        LineBeginningPtr = *scanPtr;
        if (CodeLLPtr != NULL)  {
            FirstCmdElement = *CodeElement;
            retrn = AddLine(*scanPtr, FirstCmdElement);
            if (retrn != passed)  return retrn;     }


            // see if we're dealing with a control command (if/while/etc.)
		
        control_cmd = get_control_command(scanPtr);         // g_c_c() takes care of get_next_word if a match is found


            // *** an if-endif block ***

        if ((control_cmd == if_Ccmd) || (control_cmd == elseif_Ccmd) || (control_cmd == else_Ccmd))  {
            char *StartOfConditional;

            if ( ((control_cmd == elseif_Ccmd) || (control_cmd == else_Ccmd)) &&
                    ((prev_Ccmd != if_Ccmd)) && (prev_Ccmd != elseif_Ccmd))  {
                errPtr = LineBeginningPtr;
                return must_follow_if_err;    }


                // make room for the unconditional jump from the last block preceding an elseif or else block...
            
            if (((control_cmd == elseif_Ccmd) || (control_cmd == else_Ccmd)) && (CodeLLPtr != NULL))  {
                if_alwaysJumpElement = *CodeElement+1;
                retrn = InsertCodeWord(CodeLLPtr, CodeElement, jump_always);
                if (retrn == passed)  retrn = InsertCodeWord(CodeLLPtr, CodeElement, 0);    // where the jump element will go
                if (retrn != passed)  {  errPtr = *scanPtr;  return retrn;  }    }
                    

                // and fill in the offset for the conditional jump at the beginning of the last block
            
            if ((CodeLLPtr != NULL) && (prev_conditional_jump != 0) && (control_cmd != if_Ccmd))  {
                if (*LL_SLong(CodeLLPtr, prev_conditional_jump) == 0)   {
                    *LL_SLong(CodeLLPtr, prev_conditional_jump) = (Slong) (*CodeElement)-prev_conditional_jump;
            }   }


                // write out the conditional jump for the present block
            
            if (control_cmd != else_Ccmd)   {
            
                    // write out the jump command and make room for the offset
                
                if (CodeLLPtr != NULL)  {
                    retrn = InsertCodeWord(CodeLLPtr, CodeElement, jump_if_false);
                    if (retrn != passed)  {  errPtr = *scanPtr;  return retrn;  }

                    if_condJumpElement = *CodeElement;
                    retrn = InsertCodeWord(CodeLLPtr, CodeElement, 0);       // where the jump element will go
                    if (retrn != passed)  {  errPtr = *scanPtr;  return retrn;  }   }


                    // parse the conditional expression
                
                StartOfConditional = *scanPtr;
                retrn = ParseExpression(scanPtr, NULL, &first_cmd, &first_flags, CodeLLPtr, CodeElement);
                if (retrn != passed)  return retrn;

                if (first_cmd == not_a_command)
                    {  errPtr = *scanPtr;  return conditional_expected_err;  }
                if (AllowedCMDJumpTable[log_arg][first_cmd] == &IllegalCmd)
                    {  errPtr = StartOfConditional;  return conditional_expected_err;  }
            }

            if (if_expression_terminator(**scanPtr) == yz_false)
                {  errPtr = *scanPtr;  return eol_expected_err;  }


                // now process the code inside the conditional block
            
            retrn = ParseStatements(scanPtr, CodeLLPtr, CodeElement, control_cmd, if_condJumpElement);
            if (retrn != passed)  return retrn;
            LineBeginningPtr = *scanPtr;


                // finally, make sure the appropriate end __ is present, and write out the last jump offset if necessary
            
            if (get_control_command(scanPtr) != endif_Ccmd)
                {  errPtr = LineBeginningPtr;  return endif_expected_err;  }
            if (control_cmd != if_Ccmd)  *scanPtr = LineBeginningPtr;

            if (((control_cmd == elseif_Ccmd) || (control_cmd == else_Ccmd)) && (CodeLLPtr != NULL))
                *(Slong *) Element(CodeLLPtr, if_alwaysJumpElement) = (*CodeElement)-if_alwaysJumpElement;

            if (control_cmd != if_Ccmd)  return passed;
        }


            // *** a while-endw block ***

        else if (control_cmd == while_Ccmd)  {

            char *StartOfConditional;


                // write out the conditional jump at the beginning; make room for the offset
                
            if (CodeLLPtr != NULL)  {
                retrn = InsertCodeWord(CodeLLPtr, CodeElement, jump_if_false);
                if (retrn == passed)  retrn = InsertCodeWord(CodeLLPtr, CodeElement, 0);    // where the jump element will go
                if (retrn != passed)  {  errPtr = *scanPtr;  return retrn;  }   }


                // parse the conditional expression
            
            StartOfConditional = *scanPtr;
            retrn = ParseExpression(scanPtr, NULL, &first_cmd, &first_flags, CodeLLPtr, CodeElement);
            if (retrn != passed)  return retrn;

            if (first_cmd == not_a_command)
                {  errPtr = *scanPtr;  return conditional_expected_err;  }
            if (AllowedCMDJumpTable[log_arg][first_cmd] == &IllegalCmd)
                {  errPtr = StartOfConditional;  return conditional_expected_err;  }
            if (if_expression_terminator(**scanPtr) == yz_false)
                {  errPtr = *scanPtr;  return eol_expected_err;  }


                // process the statements in the while block
            
            retrn = ParseStatements(scanPtr, CodeLLPtr, CodeElement, control_cmd, 0);
            if (retrn != passed)  return retrn;
            LineBeginningPtr = *scanPtr;


                // make sure the block ends with an endw

            if (get_control_command(scanPtr) != endwhile_Ccmd)
                {  errPtr = LineBeginningPtr;  return endwhile_expected_err;  }


                // write out the unconditional jump at the end, and fill in the offset for the conditional jump
            
            if (CodeLLPtr != NULL)   {
                retrn = InsertCodeWord(CodeLLPtr, CodeElement, jump_always);
                if (retrn == passed)  retrn = InsertCodeWord(CodeLLPtr, CodeElement, 0);
                if (retrn == passed)  *(Slong *) Element(CodeLLPtr, (*CodeElement)-1) = FirstCmdElement-(*CodeElement-1);
                if (retrn != passed)  {  errPtr = *scanPtr;  return retrn;  }

                *(Slong *) Element(CodeLLPtr, FirstCmdElement+1) = (*CodeElement)-(FirstCmdElement+1);
            }   }


            // *** a do-until block ***

        else if (control_cmd == do_Ccmd)  {
            char *StartOfConditional;

            if (if_expression_terminator(**scanPtr) == yz_false)
                {  errPtr = *scanPtr;  return eol_expected_err;  }


                // process the statements (no jump preceding block)
            
            retrn = ParseStatements(scanPtr, CodeLLPtr, CodeElement, control_cmd, 0);
            if (retrn != passed)  return retrn;

            LineBeginningPtr = *scanPtr;
            if (get_control_command(scanPtr) != until_Ccmd)
                {  errPtr = LineBeginningPtr;  return until_expected_err;  }


                // write out the conditional jump command and offset
            
            if (CodeLLPtr != NULL)  {
                retrn = InsertCodeWord(CodeLLPtr, CodeElement, jump_if_false);
                if (retrn == passed)  retrn = InsertCodeWord(CodeLLPtr, CodeElement, (Slong) FirstCmdElement-*CodeElement);
                if (retrn != passed)  {  errPtr = *scanPtr;  return retrn;  }   }


                // process the conditional expression
            
            StartOfConditional = *scanPtr;
            retrn = ParseExpression(scanPtr, NULL, &first_cmd, &first_flags, CodeLLPtr, CodeElement);
            if (retrn != passed)  return retrn;

            if (first_cmd == not_a_command)
                {  errPtr = *scanPtr;  return conditional_expected_err;  }
            if (AllowedCMDJumpTable[log_arg][first_cmd] == &IllegalCmd)
                {  errPtr = StartOfConditional;  return conditional_expected_err;  }
        }


            // *** a for-endf loop ***

        else if (control_cmd == for_Ccmd)  {
            Ulong for_JumpElement = 0, for_condJumpElement = 0;
            char *VarPtr, *StartIndexPtr, *EndIndexPtr, *StepPtr = NULL;
            Ulong CompareIndex = 0;      // avoid uninitialized-variable warning
            
            
                // Read in the different parts of the for expression, marking the locations of expressions we will compile later

            VarPtr = *scanPtr;       // read the looping variable
            retrn = ParseExpression(scanPtr, NULL, &first_cmd, &first_flags, NULL, NULL);
            if (retrn != passed)  return retrn;
            if (first_cmd == not_a_command)
                {  errPtr = *scanPtr;  return variable_expected_err;  }
            if (AllowedCMDJumpTable[var_arg][first_cmd] == &IllegalCmd)
                {  errPtr = VarPtr;  return variable_expected_err;  }

            ForwardPtr = *scanPtr;         // read the 'in'
            if (CompareString(&ForwardPtr, (char *) ControlCommandNames[in_Ccmd]) == yz_false)
                {  errPtr = ForwardPtr;  return in_expected_err;  }
            *scanPtr = ForwardPtr;
            retrn = get_next_word(scanPtr);
            if (retrn != passed)  return retrn;

            if (**scanPtr != '[')           // read the '['
                {  errPtr = *scanPtr;  return left_bracket_expected_err;  }
            retrn = get_next_word(scanPtr);
            if (retrn != passed)  return retrn;

            StartIndexPtr = *scanPtr;       // read the starting index
            retrn = ParseExpression(scanPtr, NULL, &first_cmd, &first_flags, NULL, NULL);
            if (retrn != passed)  return retrn;
            if (first_cmd == not_a_command)
                {  errPtr = *scanPtr;  return var_or_num_expected_err;  }
            if (AllowedCMDJumpTable[data_arg][first_cmd] == &IllegalCmd)
                {  errPtr = StartIndexPtr;  return var_or_num_expected_err;  }

            if (**scanPtr != ',')           // read the ','
                {  errPtr = *scanPtr;  return comma_expected_err;  }
            next_char(scanPtr);
            if (lettertype(**scanPtr) == a_space)   {
                retrn = get_next_word(scanPtr);
                if (retrn != passed)  return retrn;   }

            EndIndexPtr = *scanPtr;         // read the ending index
            retrn = ParseExpression(scanPtr, NULL, &first_cmd, &first_flags, NULL, NULL);
            if (retrn != passed)  return retrn;
            if (first_cmd == not_a_command)
                {  errPtr = *scanPtr;  return var_or_num_expected_err;  }
            if (AllowedCMDJumpTable[data_arg][first_cmd] == &IllegalCmd)
                {  errPtr = EndIndexPtr;  return var_or_num_expected_err;  }

            if (**scanPtr != ']')           // read the ']'
                {  errPtr = *scanPtr;  return right_bracket_expected_err;  }
            retrn = get_next_word(scanPtr);
            if (retrn != passed)  return retrn;

            
            ForwardPtr = *scanPtr;          // read the 'step'
            if (CompareString(&ForwardPtr, (char *) ControlCommandNames[step_Ccmd]) == yz_true)  {
                retrn = get_next_word(scanPtr);
                if (retrn != passed)  return retrn;

                StepPtr = *scanPtr;        // read the step index
                retrn = ParseExpression(scanPtr, NULL, &first_cmd, &first_flags, NULL, NULL);
                if (retrn != passed)  return retrn;
                if (first_cmd == not_a_command)
                    {  errPtr = *scanPtr;  return var_or_num_expected_err;  }
                if (AllowedCMDJumpTable[data_arg][first_cmd] == &IllegalCmd)
                    {  errPtr = StepPtr;  return var_or_num_expected_err;  }    }

            if (if_expression_terminator(**scanPtr) == yz_false)
                {  errPtr = *scanPtr;  return eol_expected_err;  }

            if (CodeLLPtr != NULL)   {
                char *tempVarPtr = VarPtr, *tempStartIndexPtr = StartIndexPtr, *tempEndIndexPtr = EndIndexPtr;
                
                
                    // write out the initial assignment of the counter variable

                retrn = InsertCodeWord(CodeLLPtr, CodeElement, def_general);
                if (retrn == passed)  retrn = InsertCodeWord(CodeLLPtr, CodeElement, equ_flags);
                if (retrn == passed)  retrn = ParseExpression(&tempVarPtr, NULL, &first_cmd, &first_flags, CodeLLPtr, CodeElement);
                if (retrn != passed)  {  errPtr = VarPtr;  return retrn;  }
                retrn = ParseExpression(&tempStartIndexPtr, NULL, &first_cmd, &first_flags, CodeLLPtr, CodeElement);
                if (retrn != passed)  return retrn;
                
                
                    // write the conditional jump

                tempVarPtr = VarPtr;
                for_JumpElement = *CodeElement;
                retrn = InsertCodeWord(CodeLLPtr, CodeElement, jump_if_true);
                for_condJumpElement = *CodeElement;
                if (retrn == passed)  retrn = InsertCodeWord(CodeLLPtr, CodeElement, 0);
                CompareIndex = *CodeElement;
                if (retrn == passed)  retrn = InsertCodeWord(CodeLLPtr, CodeElement, if_gt);
                if (retrn != passed)  {  errPtr = VarPtr;  return retrn;  }
                retrn = ParseExpression(&tempVarPtr, NULL, &first_cmd, &first_flags, CodeLLPtr, CodeElement);
                if (retrn != passed)  return retrn;
                retrn = ParseExpression(&tempEndIndexPtr, NULL, &first_cmd, &first_flags, CodeLLPtr, CodeElement);
                if (retrn != passed)  return retrn;
            }
                
                
                // now process the statements inside the loop, and check the endf

            retrn = ParseStatements(scanPtr, CodeLLPtr, CodeElement, control_cmd, 0);
            if (retrn != passed)  return retrn;
            LineBeginningPtr = *scanPtr;

            if (get_control_command(scanPtr) != endfor_Ccmd)
                {  errPtr = LineBeginningPtr;  return endfor_expected_err;  }

            if (CodeLLPtr != NULL)    {
                char *tempVarPtr = VarPtr, *tempStepPtr = StepPtr;


                    // write the command to increment the counter variable
                
                retrn = InsertCodeWord(CodeLLPtr, CodeElement, def_general);     // increment var = var+step
                if (retrn == passed)  retrn = InsertCodeWord(CodeLLPtr, CodeElement, equ_flags);
                if (retrn != passed)  {  errPtr = VarPtr;  return retrn;  }
                retrn = ParseExpression(&tempVarPtr, NULL, &first_cmd, &first_flags, CodeLLPtr, CodeElement);
                if (retrn != passed)  return retrn;

                tempVarPtr = VarPtr;
                retrn = InsertCodeWord(CodeLLPtr, CodeElement, addf);      // the addf command
                if (retrn != passed)  {  errPtr = VarPtr;  return retrn;  }
                retrn = ParseExpression(&tempVarPtr, NULL, &first_cmd, &first_flags, CodeLLPtr, CodeElement);
                if (retrn != passed)  return retrn;

                if (StepPtr == NULL)   {      // no step -- increment by 1
                    retrn = WriteNum(CodeLLPtr, CodeElement, 1, yz_false);
                    if (retrn != passed)  {  errPtr = *scanPtr;  return retrn;  }    }

                else    {     // step given -- use that expression instead
                    signed char StepDirection = 0;
                    Ulong StepIndex;
                    
                    StepIndex = *CodeElement;
                    retrn = ParseExpression(&tempStepPtr, NULL, &first_cmd, &first_flags, CodeLLPtr, CodeElement);
                    if (retrn != passed)  return retrn;

                    if (*LL_SLong(CodeLLPtr, StepIndex) == constant_Slong)    {
                        if (*LL_SLong(CodeLLPtr, StepIndex+1) > 0)  StepDirection = 1;
                        else  StepDirection = -1;       }
                    else if (*LL_SLong(CodeLLPtr, StepIndex) == constant_float)      {
                        double TempFloat;
                        FGetElements(CodeLLPtr, StepIndex+1, StepIndex+doublesize_in_longs, &TempFloat);
                        if (TempFloat > 0)  StepDirection = 1;
                        else  StepDirection = -1;        }


                        // we can have decreasing counter variables only if the step is an inlined (negative) constant
                    
                    if (StepDirection == -1)  *LL_SLong(CodeLLPtr, CompareIndex) = if_lt;
                }


                    // write out the terminating jump-if-<= command
                
                tempVarPtr = VarPtr;    
                retrn = InsertCodeWord(CodeLLPtr, CodeElement, jump_always);
                if (retrn == passed)  retrn = InsertCodeWord(CodeLLPtr, CodeElement, 0);


                    // fill in miscellaneous offsets
                
                *(Slong *) Element(CodeLLPtr, (*CodeElement)-1) = for_JumpElement-((*CodeElement)-1);
                *(Slong *) Element(CodeLLPtr, for_condJumpElement) = *CodeElement-for_condJumpElement;
            }

        }


            // in the case of end __, do some basic checks, fill in some jump offsets and return
        
        else if ((control_cmd == endif_Ccmd) || (control_cmd == endfor_Ccmd) || (control_cmd == endwhile_Ccmd) || (control_cmd == until_Ccmd))   {

            if ((if_expression_terminator(**scanPtr) == yz_false) && (control_cmd != until_Ccmd))
                {  errPtr = *scanPtr;  return eol_expected_err;  }
            if (prev_Ccmd == not_a_Ccmd)         // mismatched terminators are caught in the if, for, etc. routines
                {  errPtr = LineBeginningPtr;  return unexpected_control_terminator_err;  }
                    
            if ((CodeLLPtr != NULL) && (prev_conditional_jump != 0))    {       // fill in any previous jf (e.g. from an if to the else)
                if (*LL_SLong(CodeLLPtr, prev_conditional_jump) == 0)   {
                    *LL_SLong(CodeLLPtr, prev_conditional_jump) = (*CodeElement)-prev_conditional_jump;
            }   }

            *scanPtr = LineBeginningPtr;     // they are handled separately
            return passed;    }      // fall out and let the next level read over it


            // the last possibility -- it's an ordinary Yazoo sentence (no flow control word)
        
        else if (control_cmd == not_a_Ccmd)  {
        
                // read and process the normal way; make sure it ends in a newline or comma; fix up a couple of miscellaneous things
                
            retrn = ParseExpression(scanPtr, NULL, &first_cmd, &first_flags, CodeLLPtr, CodeElement);
            if (retrn != passed)  return retrn;

            if (lettertype(**scanPtr) == a_letter)    // then we hit a control command in the middle of a line
                {  errPtr = *scanPtr;  return eol_expected_err;  }
                
            if (first_cmd == negatef)  first_cmd = mulf;            // equivalent for our purposes here
            if (first_cmd != built_in_function)  first_flags = 0;   // to prevent an error


                // Check to see if we have to insert a token.  Various possibilities below..

            if (first_cmd != not_a_command)    {
                if ((AllowedCMDJumpTable[start_arg][first_cmd] == &IllegalCmd) || (first_cmd == user_function)
                            || ((first_cmd == built_in_function) && (BIF_CanStartSentence[first_flags] == yz_false)))    {
                            

                        // bare code blocks get hidden-define operators
                    
                    if ((first_cmd == append_code) || ((first_cmd >= code_block) && (first_cmd <= string_cmd)))
                        DG_Flags = def_flags;       // second_cmd = user_function;  }
                        

                        // variables/user functions/etc. get a hidden define-equate operator
                        
                    else if ((AllowedCMDJumpTable[var_arg][first_cmd] != &IllegalCmd)  // don't point to registers; a(top(), size());
                                && (first_cmd != built_in_function))                   // that would change R_slong
                        DG_Flags = dqa_flags-dg_update_members;        // user_f:  return value can be overwritten; e.g. print(f(a), f(b))
                        
                        
                        // other expressions (Yazoo functions) copy their data directly
                    
                    else if (AllowedCMDJumpTable[data_arg][first_cmd] != &IllegalCmd)  DG_Flags = deq_flags-dg_update_members;
                    
                    
                        // anything else is a legitimate bug in the source code
                        
                    else  DG_Flags = dg_bad_flags;


                        // Now write out the token.
                    
                    if ((CodeLLPtr != NULL) && (DG_Flags != dg_bad_flags) )  {
                        Ulong loopCmdElement = FirstCmdElement, IndicesTestCmdElement;
                        
                            // this part's a little hokey -- try to infer if the token needs to be an array or not
                            
                        while ((*LL_SLong(CodeLLPtr, loopCmdElement) == step_to_memberID) || (*LL_SLong(CodeLLPtr, loopCmdElement) == step_to_index))  {
                            if (loopCmdElement >= CodeLLPtr->elementnum)  break;
                            loopCmdElement++;         }
                        IndicesTestCmdElement = *LL_SLong(CodeLLPtr, loopCmdElement);
                        
                        
                            // write out the token:  either $1 := .. or $1[*] := ...  (the operators may also be :: or := @)
                        
                        if ((IndicesTestCmdElement == step_to_indices) || (IndicesTestCmdElement == step_to_all) ||
                                (IndicesTestCmdElement == resize) || (IndicesTestCmdElement == insert_indices_up) ||
                                (IndicesTestCmdElement == insert_indices_down))   {
                            retrn = InsertPrefix(CodeLLPtr, &FirstCmdElement, DG_Flags, yz_true);
                            (*CodeElement) += 5;        }       // to account for the prefix insertion
                        else    {
                            retrn = InsertPrefix(CodeLLPtr, &FirstCmdElement, DG_Flags, yz_false);
                            (*CodeElement) += 4;        }       // to account for the prefix insertion
                        if (retrn != passed)  {  errPtr = LineBeginningPtr;  return retrn;  }       }
                    
                    if (DG_Flags != dg_bad_flags)  first_cmd = def_general;
                }

                if (AllowedCMDJumpTable[start_arg][first_cmd] == &IllegalCmd)   {
                    errPtr = LineBeginningPtr;
                    return not_a_statement_err;    }     }
        }


            // The last possibility is that our sentence began with something that doesn't fit into any category.
        
        else    {
            errPtr = *scanPtr;
            return illegal_control_command_err;   }

        retrn = get_start_of_line(scanPtr, CodeLLPtr, CodeElement);     // so we will reuse the endif for all elseifs and the if
        if (retrn != passed)  return retrn;
    }
    
    
        // Add the final line to the NewlinePositions[] array.
    
/*    if (CodeLLPtr != NULL)  {
        retrn = AddLine(*scanPtr, *CodeElement);
        if (retrn != passed)  return retrn;     }*/

    return passed;
}


// AddLine() updates the NewlinePositions[] array for each new sentence on a new line.
// NewlinePositions[] is used for mapping runtime errors to the culpable line of text in the script source code.

parse_retrn AddLine(char *scanPtr, Ulong FirstCmdElement)
{
    Ulong line_to_insert, char_num;
    LL_retrn linkedLretrn;
    char *PrevEOL = scanPtr-1;
    Boolean if_add;
    
    if (PrevEOL < DocStartPtr)  PrevEOL = DocStartPtr;
    //    while (((lettertype(*PrevEOL) == a_space) || (lettertype(*PrevEOL) == unprintable)) && (PrevEOL > DocStartPtr))  PrevEOL--;
    while ((lettertype(*PrevEOL) != a_eol) && (PrevEOL > DocStartPtr))  PrevEOL--;
    
    
    // we don't add anything if the 'line' began with a comma
    
    if (((*PrevEOL != ',') && (*PrevEOL != ';') && (lettertype(*PrevEOL) == a_eol)) || (PrevEOL == DocStartPtr))    {
        
        if (PrevEOL == DocStartPtr)  char_num = 1;
        else  char_num = PrevEOL - DocStartPtr + 2;
        
        line_to_insert = FindLine(&NewlinePositions, PrevEOL - DocStartPtr + 2, 0) + 1;
        
        if_add = yz_true;
        if (NewlinePositions.elementnum > 0)   {
            if (*LL_ULong(&NewlinePositions, line_to_insert-1) == char_num)    {
                if_add = yz_false;
                line_to_insert--;
            }}
        if (NewlinePositions.elementnum >= line_to_insert)   {
            if (*LL_ULong(&NewlinePositions, line_to_insert) == char_num)    {
                if_add = yz_false;
            }}
        
        if (if_add == yz_true)    {
            linkedLretrn = FInsertElements(&NewlinePositions, line_to_insert, 1, uncleared);
            if (linkedLretrn != passed)  return memory_err;
            
            *LL_ULong(&NewlinePositions, line_to_insert) = char_num;
            *(LL_ULong(&NewlinePositions, line_to_insert) + 1) = FirstCmdElement;       }
        
        else  {
            if (*(LL_ULong(&NewlinePositions, line_to_insert) + 1) > FirstCmdElement)   {       // unlikely
                *(LL_ULong(&NewlinePositions, line_to_insert) + 1) = FirstCmdElement;
            }   }
    }
    
    return passed;
}



// ParseExpression() processes an expression from *scanPtr up to (and not including) endPtr.
// Expands the expression as Hobbish code in the CodeLLPtr list if CodeLLPtr is not zero.
// One call of this function expands the last lowest-OoO arithmetic operator to be found,
// and recursively calls itself for the arguments of that operator.
// For example, if given 4-2+1 it writes out the +, then calls itself to write out '4-2' and '1'.
// Brings 'StartPtr' up to the first character of the token/command that ends the expression;
// which is NOT necessarily the last character of the expression (if the ending token is several characters long)

parse_retrn ParseExpression(char **StartPtr, char *endPtr, Slong *lowOoO_cmd, Slong *best_cmd_ID, linkedlist *CodeLLPtr, Ulong *CodeElement)
{
    parse_retrn retrn;
    char *BestLeftPtr = NULL, *BestRightPtr, *scanPtr, *scanPtrCopy, *LeftPtr;
    char current_OoOlevel, best_OoOlevel = 0;
    Ulong CodeInsertElement, OldCodeInsertElement, NewCodeInsertElement, FirstExprElement, OldCodeElement;
    Slong current_cmd, current_flags, prev_cmd, prev_cmd_copy, best_prev_cmd = 0, loop_counter = 0, best_loop_pass = 0;
    Boolean parentheses_cmd = yz_false, prev_paren, best_prev_paren = 0, best_is_parentheses = yz_false, negate_flag = yz_false;
    const Slong dh_flags = def_flags + dg_hidden_member;
    const Slong dqh_flags = deq_flags + dg_hidden_member;


        // get up to the beginning of the expression
    
    if (lettertype(**StartPtr) == unprintable)
            next_char(StartPtr);
    scanPtr = *StartPtr;


        // First, find the symbol we're expanding around.  Loop over each operator and measure its order-of-operations level.
    
    prev_cmd = not_a_command;
    *lowOoO_cmd = not_a_command;

    scanPtrCopy = scanPtr;
    while ( ((scanPtr < endPtr) || (endPtr == NULL)) && (if_expression_terminator(*scanPtr) == yz_false) && (get_control_command(&scanPtrCopy) == not_a_Ccmd)) 
    {
        LeftPtr = scanPtr;
        loop_counter++;


            // if it's a parenthesis, that's not a function call, then skip to the end of that expression
        
        prev_cmd_copy = prev_cmd;
        prev_paren = parentheses_cmd;
        if ((*scanPtr == '(') && (prev_paren == yz_false) && (AllowedCMDJumpTable[type_arg][prev_cmd_copy] == &IllegalCmd))   {
            retrn = get_next_word(&scanPtr);
            if (retrn == passed)
                retrn = ParseExpression(&scanPtr, endPtr, &current_cmd, &current_flags, NULL, NULL);
            if ((retrn == passed) && (*scanPtr != ')'))   {
                errPtr = scanPtr;
                return right_parenthesis_expected_err;    }
            current_OoOlevel = OoO_no_args;
            parentheses_cmd = yz_true;         }


            // if its an ordinary operator, get its OoO level
        
        else   {        
            retrn = GetOperation(&scanPtr, endPtr, &current_cmd, &current_flags, NULL, NULL, prev_cmd, prev_paren, &CodeInsertElement);
            if (current_cmd == negatef)  current_OoOlevel = OoO_negate;
            else  current_OoOlevel = OoOlevel[current_cmd];
            parentheses_cmd = yz_false;        }

        if (retrn != passed)
            return retrn;


            // if the OoO level is lower than the current lowest, set that operator as the point to expand the expression around
        
        if ( ((OpDirection[(Ulong) current_OoOlevel] == l_to_r) && (current_OoOlevel >= best_OoOlevel))
                || ((OpDirection[(Ulong) current_OoOlevel] == r_to_l) && (current_OoOlevel >  best_OoOlevel))
                || (loop_counter == 1) )   {
        if (!((loop_counter > 1) && (current_cmd == negatef)))   {
            best_OoOlevel = current_OoOlevel;
            best_loop_pass = loop_counter;
            best_prev_cmd = prev_cmd;
            best_prev_paren = prev_paren;
            *best_cmd_ID = current_flags;
            best_is_parentheses = parentheses_cmd;
            BestLeftPtr = LeftPtr;             // points to the left side of the expression
            *lowOoO_cmd = current_cmd;
            if ((current_cmd == negatef) && (parentheses_cmd == yz_false))  {  current_cmd = mulf;  negate_flag = yz_true;  }
            else negate_flag = yz_false;
        }}


            // step to the next operator
        
        retrn = get_next_word(&scanPtr);
        if (retrn != passed)  return retrn;

        prev_cmd = current_cmd;
        scanPtrCopy = scanPtr;
    }

    if (*lowOoO_cmd == not_a_command)  {  *StartPtr = scanPtr;  return passed;  }    // we must not have read a single operation


        // if we're just skipping over code, exit here -- just get StartPtr to the end
    
    if (CodeLLPtr == NULL)  {
        *StartPtr = scanPtr;
        return passed;   }


        // if it's just a parenthetical expression '(...)', step inside the parentheses and rerun the parser
    
    if (best_is_parentheses == yz_true)   {
        BestRightPtr = BestLeftPtr;
        retrn = get_next_word(&BestRightPtr);     // below, don't need to bother to check closing ); it was already done in Pass 1 above
        if (retrn == passed)
            retrn = ParseExpression(&BestRightPtr, endPtr, &current_cmd, &current_flags, CodeLLPtr, CodeElement);      }
            
        // otherwise, re-read and translate the lowest-OoO operation
    
    else  {
        BestRightPtr = BestLeftPtr;
        retrn = GetOperation(&BestRightPtr, endPtr, &current_cmd, &current_flags, CodeLLPtr, CodeElement, best_prev_cmd,
                                                                                best_prev_paren, &CodeInsertElement);
    }
    if (retrn != passed)  return retrn;


        // if it's a negatef (not a real bytecode operator), replace w/ mulf -1 ...
    
    if (negate_flag == yz_true)   {     // if negate flag is set, don't worry about the left argument at all
        *lowOoO_cmd = mulf;             // treat the right-hand argument as we would for a regular multiplication
        retrn = WriteNum(CodeLLPtr, CodeElement, -1, yz_false);   }


        // Process left-hand argument of low-OoO operator, if one is expected/present.
        // Do this one before the right-hand argument, since that should be faster w/ my LL routines.

    else if ((left_args[*lowOoO_cmd] == no_arg) || (best_is_parentheses == yz_true))    {
        if (best_loop_pass != 1)  {
            errPtr = BestLeftPtr;            // if we have a left-hand argument and none was expected, throw an error
            return no_left_arg_allowed_err;      }    }

    else    {                      // expect a left-hand argument
        Slong left_cmd, left_flags;
        
        
            // For these next operators, put the new code in the middle of what GetOperation() compiled.
            // For example:  a.b would compile sti b, then sti sm a b
        
        if ((*lowOoO_cmd == user_function) || (*lowOoO_cmd == step_to_memberID) || (*lowOoO_cmd == step_to_index) || (*lowOoO_cmd == resize)
                    || (*lowOoO_cmd == step_to_indices) || (*lowOoO_cmd == insert_indices_up) || (*lowOoO_cmd == insert_indices_down))  {
            FirstExprElement = CodeInsertElement;
            OldCodeInsertElement = CodeInsertElement;
            NewCodeInsertElement = CodeInsertElement;
            retrn = ParseExpression(StartPtr, BestLeftPtr, &left_cmd, &left_flags, CodeLLPtr, &NewCodeInsertElement);
            if (CodeLLPtr != NULL)  (*CodeElement) += NewCodeInsertElement - OldCodeInsertElement;           }
            
            
            // If it's def_general, compile the argument at the end of the bytecode.
            // Prefix with a 'this.' if it is defining a member by name (so a.b :: ulong --> this.a.b :: ulong).

        else if (*lowOoO_cmd == def_general)  {
            FirstExprElement = *CodeElement;
            OldCodeElement = *CodeElement;
            retrn = ParseExpression(StartPtr, BestLeftPtr, &left_cmd, &left_flags, CodeLLPtr, CodeElement);
            if (left_cmd == not_a_command)  {  errPtr = BestLeftPtr;  return arg_expected_err;  }      // check this before adding the prefix
            
            /* if (((*LL_SLong(CodeLLPtr, OldCodeElement-1) & dg_add_members) != 0) && ((*LL_SLong(CodeLLPtr, OldCodeElement) == search_member) ||
                    ((*LL_SLong(CodeLLPtr, OldCodeElement) >= step_to_index) && (*LL_SLong(CodeLLPtr, OldCodeElement) <= step_to_all)) ))  {
                while ((*LL_SLong(CodeLLPtr, OldCodeElement) >= step_to_index) && (*LL_SLong(CodeLLPtr, OldCodeElement) <= step_to_all))  {
                    OldCodeElement++;      }
                if (*LL_SLong(CodeLLPtr, OldCodeElement) == search_member)     {     // add the 'this.' prefix
                    *LL_SLong(CodeLLPtr, OldCodeElement) = step_to_memberID;
                    OldCodeElement++;
                    retrn = InsertCodeWord(CodeLLPtr, &OldCodeElement, this_var);
                    if (retrn != passed)  return retrn;
                    (*CodeElement)++;
        }   }   */}


            // All other operators are compiled at the end of the bytecode.

        else    {
            FirstExprElement = *CodeElement;
            retrn = ParseExpression(StartPtr, BestLeftPtr, &left_cmd, &left_flags, CodeLLPtr, CodeElement);    }

        if (retrn != passed)  return retrn;
        if (left_cmd == not_a_command)  {  errPtr = BestLeftPtr;  return arg_expected_err;  }      // we expected a left argument; none exists
        if (left_cmd == negatef)  left_cmd = mulf;       // equivalent for our purpose here
        
        
            // If the LH argument is illegal left of our pivot operator, and if adding a hidden '$1 :: ' prefix will make it legal, then do so.
        
        if (AllowedCMDJumpTable[(Ulong) left_args[*lowOoO_cmd]][left_cmd] == &IllegalCmd)   {
            if ( ((left_cmd == code_block) || (left_cmd == append_code))
                        && (AllowedCMDJumpTable[(Ulong) left_args[*lowOoO_cmd]][def_general] != &IllegalCmd))   {
                retrn = InsertPrefix(CodeLLPtr, &FirstExprElement, dh_flags, yz_false);
                if (retrn != passed)  {  errPtr = BestLeftPtr;  return retrn;  }
                (*CodeElement) += 4;      }
            else    {
                errPtr = *StartPtr;
                return invalid_command_err;
    }   }   }


        // Process left-hand argument of low-OoO operator, if one is expected/present.
        // Do this first since that should be faster w/ my LL routines.

    if ((right_args[*lowOoO_cmd] == no_arg) || (best_is_parentheses == yz_true))   {   // if no argument to the right is expected
        if (best_loop_pass != loop_counter)  {
            errPtr = BestLeftPtr;
            return no_right_arg_allowed_err;   }   }
    
    else    {                // expect an argument to the right
        Slong right_cmd, right_flags;

        
            // parse the expression to the right
        
        FirstExprElement = *CodeElement;
        retrn = get_next_word(&BestRightPtr);
        if (retrn == passed)  retrn = ParseExpression(&BestRightPtr, endPtr, &right_cmd, &right_flags, CodeLLPtr, CodeElement);
        if (retrn != passed)  return retrn;


            // If no expression detected, throw an error (unless the operator is return).
            // A bare 'return' gets converted into 'return *'
        
        if (right_cmd == not_a_command)     {
            if (*lowOoO_cmd == func_return)     {
                right_cmd = no_var;
                retrn = InsertCodeWord(CodeLLPtr, CodeElement, no_var);
                if (retrn != passed)  {  errPtr = BestRightPtr;  return retrn;  }  }
            else
                {  errPtr = BestRightPtr;  return arg_expected_err;  }  }

        if (right_cmd == negatef)  right_cmd = mulf;     // equivalent
        
        
            // special case:  convert 'a = {}' into 'a = ($1 :: {})' with a hidden-define operator
        
        if ((*best_cmd_ID == equ_flags) && (right_cmd == code_block))    {    // for other def_generals, we don't need the extra def
            retrn = InsertPrefix(CodeLLPtr, &FirstExprElement, dh_flags, yz_false);
            if (retrn != passed)  {  errPtr = BestLeftPtr;  return retrn;  }
            (*CodeElement) += 4;     }
        
        
            // If we have an illegal right-hand argument, we can add the following prefixes:
            // - pivot operator of right-hand argument is {} or : -- add '$1 ::' with a hidden define, if possible
            // - current operator is 'return' -- add '$1 :=' (hidden deq) if possible
            // all other prefixes disallowed
        
        else if (AllowedCMDJumpTable[(Ulong) right_args[*lowOoO_cmd]][right_cmd] == &IllegalCmd)      {
        
            if ( ((right_cmd == code_block) || (right_cmd == append_code))    // add '$1 ::'
                        && (AllowedCMDJumpTable[(Ulong) right_args[*lowOoO_cmd]][def_general] != &IllegalCmd))   {
                retrn = InsertPrefix(CodeLLPtr, &FirstExprElement, dh_flags, yz_false);
                if (retrn != passed)  {  errPtr = BestLeftPtr;  return retrn;  }
                (*CodeElement) += 4;      }
                
            else if ((*lowOoO_cmd == func_return) && (AllowedCMDJumpTable[type_arg][right_cmd] != &IllegalCmd))    {
                retrn = InsertPrefix(CodeLLPtr, &FirstExprElement, dqh_flags, yz_false);        // add '$1 :='
                if (retrn != passed)  {  errPtr = BestLeftPtr;  return retrn;  }
                (*CodeElement) += 4;      }
            
            else   {    // throw an error
                errPtr = BestRightPtr;
                return invalid_command_err;
    }   }   }

    *StartPtr = scanPtr;
    if (negate_flag == yz_true)  *lowOoO_cmd = negatef;     // very cumbersome -- it was temporarily set to mulf to type-check its arguments

    return passed;
}


// GetOperation() reads in a single operator (and, :=, ^, @, [variable], etc.),
// and writes it out in bytecode at the position of CodeElement (if CodeLLPtr != 0).
// *scanPtr afterwards will point to the last character of the operator.

parse_retrn GetOperation(char **scanPtr, char *endPtr, Slong *the_cmd, Slong *cmd_identifier, linkedlist *CodeLLPtr,
            Ulong *CodeElement, Slong prev_cmd, Boolean prev_paren, Ulong *CodeInsertElement)
{
    char *forward_pos;
    int counter;
    parse_retrn retrn;
    Slong PE_flags, PrefixFlag, var_ID;
    Boolean if_float, no_left_arg_expected = yz_false;

    *cmd_identifier = 0;
    
    
        // if there was no previous command, or if it doesn't accept a right-hand argument,
        // then set a flag denoting that '*', '.', '-' are not to be construed as operators
        // having left-hand arguments (as these are ambiguous cases)
        
    if (prev_paren == yz_false)    {
        if (prev_cmd == not_a_command)  no_left_arg_expected = yz_true;
        else if (prev_cmd == negatef)  no_left_arg_expected = yz_true;
        else if (right_args[prev_cmd] != no_arg)  no_left_arg_expected = yz_true;     }


        // get ourselves to the start of the command
    
    *the_cmd = not_a_command;
    if (lettertype(**scanPtr) == a_null)  {
        errPtr = *scanPtr;
        return unexpected_eof_err;  }
    if (lettertype(**scanPtr) == unprintable)
        next_char(scanPtr);


        // Now we start parsing the operator.
        // First:  is it a fundamental Hobbish operation (e.g. '^', ':=', 'return')?
    
    for (counter = 0; counter < written_commands_num; counter++)   {
        forward_pos = *scanPtr;
        if (CompareString(&forward_pos, (char *) CommandNames[counter]) == yz_true)  {
            *the_cmd = CommandIDs[counter];
            *scanPtr = forward_pos;
            counter = written_commands_num;
    }   }



        // assume initially that '(' is a user function and '*' is a multiplication
    
    if ((*the_cmd == not_a_command) && (**scanPtr == '('))       // e.g. '(*' or something.  Treat it as a user function anyway
        *the_cmd = user_function;                                // to get a good errPtr; the error will be called later on
    else if ((*the_cmd == not_a_command) && (**scanPtr == '*'))  // can be tripped up by '*:' or something
        *the_cmd = mulf;


        // Is it a built-in Yazoo function?  (e.g. save(), compile())
    
    if (*the_cmd == not_a_command)
    for (counter = 0; counter < bi_commands_num; counter++)   {
        forward_pos = *scanPtr;
        if (CompareString(&forward_pos, (char *) BuiltInCommandNames[counter]) == yz_true)  {
            *the_cmd = built_in_function;
            *cmd_identifier = counter;
            *scanPtr = forward_pos;
            next_char(scanPtr);
            if (**scanPtr != '(')   {
                errPtr = *scanPtr;
                return Yazoo_function_err;  }
            counter = bi_commands_num;    }   }


        // Is it a register?
    
    if (*the_cmd == not_a_command)
    for (counter = 0; counter < internal_vars_num; counter++)   {
        forward_pos = *scanPtr;
        if (CompareString(&forward_pos, (char *) InternalVarNames[counter]) == yz_true)  {
            *the_cmd = internal_var;
            *cmd_identifier = counter;
            *scanPtr = forward_pos;
            counter = internal_vars_num;   }   }


        // Re-evaluate the operator if it's '-', a '.', or a '*' -- they might
        // signify the beginnig of a number or 'nothing'
    
    if (no_left_arg_expected == yz_true)    {
        if (*the_cmd == subf)   {
            forward_pos = *scanPtr;
            next_char(&forward_pos);
            if ((numeric_character(*forward_pos) == yz_true) && (*forward_pos != '-'))
                *the_cmd = not_a_command;   // process it as a number
            else
                *the_cmd = negatef;   }     // or as a negation
        else if (*the_cmd == step_to_memberID)
            *the_cmd = not_a_command;       // read in as a decimal point (part of a number)
        else if (*the_cmd == mulf)
            *the_cmd = no_var;     // NOT not_a_command
    }


        // If it's a step_to_var command, read in the name of the variable to step into.
    
    if (*the_cmd == step_to_memberID)   {
        next_char(scanPtr);
        retrn = ReadVarName(scanPtr, &var_ID);
        if (retrn != passed)  return retrn;    }


        // Is it a number?
    
    if ((*the_cmd == not_a_command) && (no_left_arg_expected == yz_true))
    if (numeric_character(**scanPtr) == yz_true)     {
        double temp_num;

        retrn = ReadNum(scanPtr, &temp_num, &if_float);   // if so, read it...
        if (retrn == underflow_err)  retrn = passed;
        if (retrn != passed)  return retrn;
        
        *the_cmd = constant_float;
        if (CodeLLPtr != NULL)   {
            retrn = WriteNum(CodeLLPtr, CodeElement, temp_num, if_float);    // and write it into the bytecode
            if (retrn != passed)  errPtr = *scanPtr;
            return retrn;        }   }                // .. and we're done


        // OK, if it's not a command, and it textual (1st char), then it must be a user variable.  Read it in.
        
    if ((*the_cmd == not_a_command) && (lettertype(**scanPtr) == a_letter) )
    {
        retrn = ReadVarName(scanPtr, &var_ID);
        if (retrn != passed)  return retrn;
        *the_cmd = search_member;
    }


        // Still haven't figured out what it is?  Then there's something wrong.
    
    if (*the_cmd == not_a_command)   {
        errPtr = *scanPtr;
        return unknown_command_err;   }


        // If it's one of the define operators, map the textual operator to the set of flags we should be using.
    
    if (*the_cmd > negatef)   {
        if (*the_cmd == define)  *cmd_identifier = def_flags;
        else if (*the_cmd == var_define)  *cmd_identifier = vdf_flags;
        else if (*the_cmd == member_define)  *cmd_identifier = mdf_flags;
        else if (*the_cmd == proxy_define)  *cmd_identifier = pdf_flags;
        else if (*the_cmd == def_equate_at)  *cmd_identifier = dqa_flags;
        else if (*the_cmd == equate_at)  *cmd_identifier = eqa_flags;
        else if (*the_cmd == equate)  *cmd_identifier = equ_flags;
        else if (*the_cmd == def_equate)  *cmd_identifier = deq_flags;         }
        
        
        // Now, write the command.  Numbers were handled above separately.
        // More complex things (strings, [], etc.) will be done below.
        
    if (CodeLLPtr != NULL)   {
        if (*the_cmd == negatef)
            retrn = InsertCodeWord(CodeLLPtr, CodeElement, mulf);
        else if (*the_cmd > negatef)   {
            retrn = InsertCodeWord(CodeLLPtr, CodeElement, def_general);
            if (retrn != passed)  {  errPtr = *scanPtr;  return retrn;  }
            retrn = InsertCodeWord(CodeLLPtr, CodeElement, *cmd_identifier);    }
        else
            retrn = InsertCodeWord(CodeLLPtr, CodeElement, *the_cmd);
        if (retrn != passed)  {  errPtr = *scanPtr;  return retrn;  }
        
        
            // For the step-to operators, mark the point where the left-hand argument of the operator will go.
            // (We write the following argument right here, so we don't want to add the LH argument to the bytecode top.)
            
        if ((*the_cmd == step_to_memberID) || (*the_cmd == step_to_index) || (*the_cmd == step_to_indices)
                        || (*the_cmd == resize) || (*the_cmd == insert_indices_up) || (*the_cmd == insert_indices_down))
            *CodeInsertElement = *CodeElement;
        else if (*the_cmd == user_function)
            *CodeInsertElement = *CodeElement;


            // Write any member ID that our operator might need.
        
        if ((*the_cmd == step_to_memberID) || (*the_cmd == search_member))  {
            retrn = InsertCodeWord(CodeLLPtr, CodeElement, var_ID);
            if (retrn != passed)  {  errPtr = *scanPtr;  return retrn;     }   }
        else if ((*the_cmd == built_in_function) || (*the_cmd == internal_var))   {
            retrn = InsertCodeWord(CodeLLPtr, CodeElement, *cmd_identifier);
            if (retrn != passed)  {  errPtr = *scanPtr;  return retrn;     }   }
    }
    if ((*the_cmd >= define) && (*the_cmd <= def_equate))  *the_cmd = def_general;


        // If it's a string, convert it to bytecode here.
    
    if (*the_cmd == constant_string)    {
        retrn = ReadString(scanPtr, CodeLLPtr, CodeElement);
        if (retrn != passed)  return retrn;   }


        // Operators having inlined-code arguments will recursively call ParseStatements().
    
    if ((*the_cmd == code_block) || (*the_cmd == user_function) || (*the_cmd == built_in_function))    {
        
        retrn = get_next_word(scanPtr);
        if (retrn != passed)  return retrn;
        
        
            // For user functions, write out '1' denoting the code to run.
            // For all functions, construct the argument list.
            
        if ((CodeLLPtr != NULL) && ((*the_cmd == user_function) || (*the_cmd == built_in_function)))    {
            PrefixFlag = def_flags;
            SetFlag(PrefixFlag, hidden_member, yz_true);
            if (*the_cmd == built_in_function)  {
                if (*cmd_identifier == 10)        // trap()'s define operator for its arguments doesn't run the constructor
                    SetFlag(PrefixFlag, run_constructor, yz_false);    }
            /*else     {
                retrn = InsertCodeWord(CodeLLPtr, CodeElement, 1);     // write out the '1' for user functions
                if (retrn != passed)  {  errPtr = *scanPtr;  return retrn;  }    } */

                // write out the prefix defining the argument list
            
            retrn = InsertPrefix(CodeLLPtr, CodeElement, PrefixFlag, yz_false);
            if (retrn == passed)  retrn = InsertCodeWord(CodeLLPtr, CodeElement, code_block);
            if (retrn != passed)  return retrn;     }
            
            
            // Compile the code contained in the argument list (if it's a function) or the braces.
            
        retrn = ParseStatements(scanPtr, CodeLLPtr, CodeElement, not_a_Ccmd, 0);
        if (retrn != passed)  return retrn;


            // Finally, check to make sure the right terminating character is present.
        
        if (*the_cmd == code_block)   {
            if (**scanPtr != '}')  {
                errPtr = *scanPtr;
                return right_brace_expected_err;   }   }
        else   {
            if (**scanPtr != ')')   {
                errPtr = *scanPtr;
                return right_parenthesis_expected_err;   }   }
            
            
            // Last step -- write out the '0' signaling an end to the code.
            
        if (CodeLLPtr != NULL)   {
            retrn = InsertCodeWord(CodeLLPtr, CodeElement, null_cmd);
            if (retrn != passed)  {
                errPtr = *scanPtr;
                return retrn;    }    }
    }


        // The index operators also require special treatment.  Read in the various terms in the brackets.
    
    if ((*the_cmd == step_to_index) || (*the_cmd == resize) || (*the_cmd == insert_indices_up) || (*the_cmd == insert_indices_down))    {
        Slong varID_cmd;
        Ulong tempCodeElement = 0;    // tempCodeElement is just after the step_to command
        char *BeginBracketPtr = *scanPtr;        // for error pointers


            // Get past the brackets and compile the first argument.
        
        if (CodeLLPtr != NULL)  tempCodeElement = *CodeElement;
        retrn = get_next_word(scanPtr);
        if (retrn == passed)  retrn = ParseExpression(scanPtr, endPtr, &varID_cmd, &PE_flags, CodeLLPtr, CodeElement);
        if (retrn != passed)  return retrn;
        if (varID_cmd == negatef)  varID_cmd = mulf;     // equivalent


            // make sure arg #1 is present and numeric
            
        if (varID_cmd == not_a_command)   {
            errPtr = *scanPtr;
            return var_or_num_expected_err;    }

        if (AllowedCMDJumpTable[data_arg][varID_cmd] == &IllegalCmd)    {
            errPtr = BeginBracketPtr;
            return var_or_num_expected_err;    }


            // The resize operator only takes a single argument.  All others can have a second; if that is present, compile it.
            
        if ((*the_cmd != resize) && (**scanPtr == ','))    {
        
            next_char(scanPtr);      // get to the beginning of the second argument
            if (lettertype(**scanPtr) == a_space)    {
                retrn = get_next_word(scanPtr);
                if (retrn != passed)  return retrn;   }


                // update the operator if it was a step-to-index (singular)

            if (*the_cmd == step_to_index)  {
                *the_cmd = step_to_indices;
                if (CodeLLPtr != NULL)  *LL_SLong(CodeLLPtr, tempCodeElement-1) = step_to_indices;      }
                
                
                // compile index #2
                
            retrn = ParseExpression(scanPtr, endPtr, &varID_cmd, &PE_flags, CodeLLPtr, CodeElement);
            if (retrn != passed)  return retrn;
            if (varID_cmd == negatef)  varID_cmd = mulf;


                // make sure the second index is present & potentially numeric
            
            if (varID_cmd == not_a_command)   {
                errPtr = *scanPtr;
                return var_or_num_expected_err;    }
            if (AllowedCMDJumpTable[data_arg][varID_cmd] == &IllegalCmd)   {
                errPtr = BeginBracketPtr;
                return var_or_num_expected_err;    }
        }
        
        
            // Insertion operators always take two indices; if only one was given then we need to copy it over for the second argument.
            // Just recompile the same text twice.
            
        else if ((CodeLLPtr != NULL) && (*the_cmd != step_to_index) && (*the_cmd != resize))   {
        
                // recompile
                
            *scanPtr = BeginBracketPtr;
            retrn = get_next_word(scanPtr);
            if (retrn == passed)  retrn = ParseExpression(scanPtr, endPtr, &varID_cmd, &PE_flags, CodeLLPtr, CodeElement);
            if (retrn != passed)  return retrn;
            if (varID_cmd == negatef)  varID_cmd = mulf;

                // and re-check (is this really necessary?)
            
            if (varID_cmd == not_a_command)   {
                errPtr = *scanPtr;
                return var_or_num_expected_err;    }
            if (AllowedCMDJumpTable[data_arg][varID_cmd] == &IllegalCmd)    {
                errPtr = BeginBracketPtr;
                return var_or_num_expected_err;   }
        }


            // make sure the ending bracket is in place
            
        if (**scanPtr != ']')  {
            errPtr = *scanPtr;
            return right_bracket_expected_err;      }   }
    
    return passed;
}


// InsertCodeWord() adds a bytecode word (unsigned long) to the linked list that contains the developing code.

parse_retrn InsertCodeWord(linkedlist *CodeLLPtr, Ulong *CodeElement, Slong CodeWord)
{
    LL_retrn linkedLretrn;

    linkedLretrn = AddToCLLP(CodeLLPtr, CodeElement, 1);
    if (linkedLretrn != passed)  return memory_err;

    *LL_SLong(CodeLLPtr, (*CodeElement)-1) = CodeWord;

    return passed;
}


// AddToCLLP() inserts elements into the code list, and adjusts the NewlinePositions array
// to correct for the changed positions of the code-list elements above the insertion point.

parse_retrn AddToCLLP(linkedlist *CodeLLPtr, Ulong *CodeElement, Ulong NewElementsNum)
{
    Ulong NP_counter;
    LL_retrn linkedLretrn;

    linkedLretrn = InsertElements(CodeLLPtr, *CodeElement, NewElementsNum, uncleared);
    if (linkedLretrn != passed)  return memory_err;
    (*CodeElement) += NewElementsNum;

    NP_counter = NewlinePositions.elementnum;
    while (*CodeElement < *(LL_ULong(&NewlinePositions, NP_counter) + 1))   {
        (*(LL_ULong(&NewlinePositions, NP_counter) + 1)) += NewElementsNum;
        NP_counter--;    }                       // can never reach the first element

    return passed;
}


// InsertPrefix() inserts a [new_name] := or :=@ (the command stored in first_cmd) 
// to go before some operator in order to make it legal in the context of the expression.

parse_retrn InsertPrefix(linkedlist *CodeLLPtr, Ulong *PrefixLocation, Slong DG_Flags, Boolean IfMultipleIndices)
{
    Slong NewFlags;
    parse_retrn retrn;
    
    
        // Tokens are by default unjammable, and have the update flag set.
    
    NewFlags = DG_Flags;
    SetFlag(NewFlags, unjammable, yz_true);
    NullNamesNum++;       // increment this -- we are assigning a new one below

    if (NewFlags != dg_bad_flags)   {
        SetFlag(NewFlags, update_members, yz_false);     // hidden parameters may change their type and we don't want an error (e.g. {args()})


            // write out the def-general operator

        retrn = InsertCodeWord(CodeLLPtr, PrefixLocation, def_general);
        if (retrn != passed)  return retrn;
        retrn = InsertCodeWord(CodeLLPtr, PrefixLocation, NewFlags);
        if (retrn != passed)  return retrn;    }


            // if it's a token of an array, add a step-to-all:  $1[*] :: ...

    if (IfMultipleIndices == yz_true)   {
        retrn = InsertCodeWord(CodeLLPtr, PrefixLocation, step_to_all);
        if (retrn != passed)  return retrn;    }
    
    
        // Write out the search-member operator and the new negative-ID member number.
        
    retrn = InsertCodeWord(CodeLLPtr, PrefixLocation, search_member);
    if (retrn != passed)  return retrn;
    retrn = InsertCodeWord(CodeLLPtr, PrefixLocation, 0);
    if (retrn != passed)  return retrn;
    *LL_SLong(CodeLLPtr, (*PrefixLocation)-1) = -NullNamesNum;
    
    return passed;
}


// WriteNum():  same as InsertCodeWord(), but takes care of the fact that the numeric types have different sizes.

parse_retrn WriteNum(linkedlist *CodeLLPtr, Ulong *CodeElement, double num_to_write, Boolean if_float)
{
    LL_retrn linkedLretrn;
    signed long long_num;

         // Write out the operator, then the number.
        
    if (if_float == yz_false)   {        // use a signed long integer, if we are able
        linkedLretrn = AddToCLLP(CodeLLPtr, CodeElement, 2);
        if (linkedLretrn != passed)  return memory_err;
        long_num = (signed long) num_to_write;
        *LL_SLong(CodeLLPtr, (*CodeElement)-2) = constant_Slong;
        *LL_SLong(CodeLLPtr, (*CodeElement)-1) = long_num;    }
        
    else    {               // or, if have to write out the full float..
        linkedLretrn = AddToCLLP(CodeLLPtr, CodeElement, doublesize_in_longs+1);
        if (linkedLretrn != passed)  return memory_err;
        *LL_SLong(CodeLLPtr, (*CodeElement)-doublesize_in_longs-1) = constant_float;
        SetElements(CodeLLPtr, (*CodeElement)-doublesize_in_longs, (*CodeElement)-1, &num_to_write);       }

    return passed;
}



// ********* Routines to read in primitive objects (numbers, etc.) ************


// ReadVarName() reads in a word from the code and compares it to the list of known words stored in AllUserNames.
// If the word already is stored, return the elementnum; otherwise, make room & return the elementnum.

parse_retrn ReadVarName(char **StartPtr, Slong *var_ID)
{
    Ulong counter;
    LL_retrn linkedLretrn;
    parse_retrn retrn;
    char *scanPtr;

    if (lettertype(**StartPtr) != a_letter)  {    // make sure it's a bona fide name
        errPtr = *StartPtr;
        return variable_expected_err;    }


        // compare to the list of names stored in AllUserNames
        
    for (counter = 1; counter <= AllUserNames.elementnum; counter++)   {
        scanPtr = *StartPtr;
        if (CompareString(&scanPtr, *(char **) LL_Ptr(&AllUserNames, counter)) == yz_true)  {
            *StartPtr = scanPtr;
            *var_ID = (Slong) counter;
            return passed;    }   }


        // If the variable wasn't found in the list, we have to add it.

    linkedLretrn = AddElements(&AllUserNames, 1, uncleared);
    if (linkedLretrn != passed)  {
        errPtr = *StartPtr;
        return memory_err;   }
    
    *var_ID = (Slong) AllUserNames.elementnum;      // the new element at the top of the list
    
    retrn = GetNewWord(*StartPtr, (char **) LL_Ptr(&AllUserNames, (Ulong) *var_ID));          // don't need to update scanPtr
    if (CompareString(StartPtr, *(char **) LL_Ptr(&AllUserNames, counter)) != yz_true)  {     // just have to get to the end of the word
        errPtr = *StartPtr;
        return unknown_err;    }    // not sure why I do this check but I guess it doesn't hurt
    return retrn;
}


// ReadString() compiles a string enclosed in quotes (or, if CodeLLPtr == NULL, just skips over it).
// Doesn't write the constant_string code word -- that is done by GetOperation().

parse_retrn ReadString(char **scanPtr, linkedlist *CodeLLPtr, Ulong *CodeElement)
{
    Ulong words_num, chars_num = 0;
    char *startPtr;
    char *start_string_Ptr, *charPtr;
    LL_retrn LLcode;

    next_string_char(scanPtr);   // get past the starting '"'
    
    
        // First, count the number of characters in the string.
    
    startPtr = *scanPtr;
    while ((**scanPtr != '"') && ((lettertype(**scanPtr) != a_eol) || (**scanPtr == ',') || (**scanPtr == ';')) && (lettertype(**scanPtr) != a_null))   {
        if (**scanPtr == '\\')  {      // deal with special characters
            next_string_char(scanPtr);
            if ((lettertype(**scanPtr) == a_number) || ((**scanPtr >= 'A') && (**scanPtr <= 'F')) || ((**scanPtr >= 'a') && (**scanPtr <= 'f')))   {
                next_string_char(scanPtr);
                next_string_char(scanPtr);    }
            else if ((**scanPtr == '\\') || (**scanPtr == 'n') || (**scanPtr == 't') || (**scanPtr == '"'))
                next_string_char(scanPtr);
            else    {
                errPtr = *scanPtr;
                return string_read_err;   }    }
        else    {
            next_string_char(scanPtr);    }
        chars_num++;    }
        
        
        // If we hit and end-of-line or end-of-file before the ending quote, throw an error.
        
    if (((lettertype(**scanPtr) == a_eol) && (**scanPtr != ',') && (**scanPtr != ';')) || (lettertype(**scanPtr) == a_null))  {
            errPtr = *scanPtr;
            return string_read_err;         }

    if (CodeLLPtr == NULL)  return passed;        // leave if we're not trying to read in any new elements


        // Get a buffer to store the string in temporarily.
        
    words_num = (Ulong) ceil(1.*chars_num/sizeof(Ulong));    // align along a long boundary
    start_string_Ptr = (char *) malloc((size_t) (words_num+1)*sizeof(Ulong));      // avoid size-0 malloc
    if (start_string_Ptr == NULL)   {
        errPtr = *scanPtr;
        return memory_err;     }
    charPtr = start_string_Ptr;
    *scanPtr = startPtr;


        // Now re-read the string, this time storing the characters in the buffer as we go.
    
    while ((**scanPtr != '"') && ((lettertype(**scanPtr) != a_eol) || (**scanPtr == ',') || (**scanPtr == ';')) && (lettertype(**scanPtr) != a_null))   {
        if (**scanPtr == '\\')  {
            next_string_char(scanPtr);
            
                // general special character:  \xy where x, y are hexadecimal
                
            if ((lettertype(**scanPtr) == a_number) || ((**scanPtr >= 'A') && (**scanPtr <= 'F')) || ((**scanPtr >= 'a') && (**scanPtr <= 'f')))   {

                if (lettertype(**scanPtr) == a_number)  *charPtr = **scanPtr - '0';
                else if ((**scanPtr >= 'A') && (**scanPtr <= 'F'))  *charPtr = **scanPtr - 'A' + 10;
                else if ((**scanPtr >= 'a') && (**scanPtr <= 'f'))  *charPtr = **scanPtr - 'a' + 10;
                else  {
                    errPtr = *scanPtr;
                    free((void *) start_string_Ptr);
                    return string_read_err;         }

                *charPtr *= 16;
                next_string_char(scanPtr);

                if (lettertype(**scanPtr) == a_number)  *charPtr += **scanPtr - '0';
                else if ((**scanPtr >= 'A') && (**scanPtr <= 'F'))  *charPtr += **scanPtr - 'A' + 10;
                else if ((**scanPtr >= 'a') && (**scanPtr <= 'f'))  *charPtr += **scanPtr - 'a' + 10;
                else  {
                    errPtr = *scanPtr;
                    free((void *) start_string_Ptr);
                    return string_read_err;     }     }

            else if (**scanPtr == '\\')     // the backslash
                    *charPtr = '\\';
            else if (**scanPtr == 'n')      // the end-of-line
                    *charPtr = '\n';
            else if (**scanPtr == 't')      // tab
                    *charPtr = '\t';
            else if (**scanPtr == '"')      // quote
                    *charPtr = '"';

            else    {               // we don't know what it is
                errPtr = *scanPtr;
                free((void *) start_string_Ptr);
                return string_read_err;     }   }

        else    {
            *charPtr = **scanPtr;    }       // if it's a normal character, just copy it over

        next_string_char(scanPtr);
        charPtr++;    }


        // Make room for the string in the code list.
    
    LLcode = AddToCLLP(CodeLLPtr, CodeElement, words_num+1);
    if (LLcode != passed)    {
        errPtr = *scanPtr;
        CodeElement -= words_num+1;
        free((void *) start_string_Ptr);
        return memory_err;     }


        // Now copy our buffer over to the code list.
    
    *LL_SLong(CodeLLPtr, (*CodeElement)-words_num-1) = chars_num;
    if (words_num > 0)   {
        SetElements(CodeLLPtr, (*CodeElement)-words_num, (*CodeElement)-1, (void *) start_string_Ptr);   }

    free((void *) start_string_Ptr);

    return passed;
}


// if_expression_terminator() returns true if the character terminates an expression; false otherwise.

Boolean if_expression_terminator(char scan_char)
{
    if ((scan_char == ')') || (scan_char == ']') || (scan_char == '}') || (lettertype(scan_char) == a_eol) || (lettertype(scan_char) == a_null))
        return yz_true;

    return yz_false;
}


// get_control_command() checks to see whether *scanPtr points to an if/while/etc; if so, returns the control command ID.
// Moves scanPtr over to the next word (not the end of the control word) if a match is found.

Ulong get_control_command(char **scanPtr)
{
    int counter;
    char *forward_pos;
    parse_retrn retrn;

    for (counter = 0; counter < control_commands_num; counter++)   {
        forward_pos = *scanPtr;
        if (CompareString(&forward_pos, (char *) ControlCommandNames[counter]) == yz_true)  {
            *scanPtr = forward_pos;
            retrn = get_next_word(scanPtr);
            if (retrn != passed)  return retrn;

            if (counter == endif_Ccmd_B)  return endif_Ccmd;       // these 3 lines take care of alternative spellings
            if (counter == endfor_Ccmd_B)  return endfor_Ccmd;     // (e.g. 'endif' vs. 'end if')
            if (counter == endwhile_Ccmd_B)  return endwhile_Ccmd;
            else  return counter;  }   }

    return not_a_Ccmd;
}


// CompareString() compares the text pointed at by *scanPtr with some template word.
// Note:  this routine alters scanPtr even if the comparison is false!

Boolean CompareString(char **scanPtr, char *stringTop)
{
    char Ltype = lettertype(*stringTop);
    char *stringPtr = stringTop;
    char *next_scanPtr;
    parse_retrn retrn;

    if (Ltype == a_null)  return yz_false;
    
    while (*stringPtr != 0)    {
    
            // deal w/ spaces -- a single space in the template can correspond to multiple spaces in the script (e.g. 'end   if')

        if ((lettertype(**scanPtr) == a_space) && (lettertype(*stringPtr) == a_space))   {
            retrn = get_next_word(scanPtr);
            if (retrn != passed)  return retrn;
            retrn = get_next_word(&stringPtr);
            if (retrn != passed)  return retrn;   }


            // for ordinary characters, compare in the straightforward way

        else   {
            if (*stringPtr != **scanPtr)  return yz_false;
            next_char(&stringPtr);
            if (*stringPtr != 0)  next_char(scanPtr);   }
    }


        // So far so good..  Make sure the script word stops where the template stops (e.g. 'endifxbz' should be illegal)

    next_scanPtr = *scanPtr;
    next_char(&next_scanPtr);
    if ((lettertype(*next_scanPtr) == Ltype) && (Ltype != a_symbol))  return yz_false;
    if ((Ltype == a_letter) && (lettertype(*next_scanPtr) == a_number))  return yz_false;

    return yz_true;
}

// GetNewWord() loads a word into our list of recognized names.
// Sticks a 0 character at the end, so one extra byte is allocated.

parse_retrn GetNewWord(char *beginning, char **stringPtr)
{
    Ulong letter_num = 0;
    char *scanPtr = beginning;
    char *temp_stringPtr;

    if (lettertype(*scanPtr) != a_letter)  return variable_expected_err;


        // First, count the number of characters in the name.
        
    do  {
        letter_num++;
        next_char(&scanPtr);
    }  while ((lettertype(*scanPtr) == a_letter) || (lettertype(*scanPtr) == a_number));

    
        // allocate storage, and copy the name over
    
    temp_stringPtr = (char *) malloc((size_t) letter_num+1);
    *stringPtr = temp_stringPtr;
    if (temp_stringPtr == NULL)  return memory_err;
        
    scanPtr = beginning;
    while ((lettertype(*scanPtr) == a_letter) || (lettertype(*scanPtr) == a_number))    {
        *temp_stringPtr = *scanPtr;
        temp_stringPtr++;
        next_char(&scanPtr);    }

    *temp_stringPtr = 0;
    return passed;
}


// ReadNum() reads in a number in decimal or hexadecimal.
// Maybe I should use scanf(), but this is convenient since it evolves scanPtr and sets if_float.

parse_retrn ReadNum(char **scanPtr, double *retrn_num, Boolean *if_float)
{
    signed long place_value, exponent, long_num;
    double float_to_add, scale_factor = 1., temp_float;
    Boolean negative = yz_false;
    Boolean exp_negative = yz_false;
    Boolean after_decimal = yz_false;
    Boolean read_a_num = yz_false;
    Boolean if_hexadecimal = yz_false;
    Ulong error_code = passed;       // for warnings, we can try to finish reading the number first
    Boolean divide_by_base = yz_false;
    char *lastChar = *scanPtr;
    
    
        // check to see if it's hexadecimal -- the number will be preceded by a '$'
    
    *retrn_num = 0;
    if (**scanPtr == '$')  {  if_hexadecimal = yz_true; lastChar = *scanPtr; next_char(scanPtr);  }
    
    if (if_hexadecimal == yz_false) place_value = 10;
    else place_value = 16;
    
    
        // We return a default 'type' -- signed long or double -- in case this number is defining a variable type (think a := #).
        // The type is a signed long unless a decimal point is present, or the number is out of integer range.
        
    *if_float = yz_false;
    
    
        // check for a leading minus sign
    
    if (**scanPtr == '-')  {
        negative = yz_true;
        lastChar = *scanPtr;
        next_char(scanPtr);    }
    
    
        // Loop over all the characters, updating our number in the floating-point register as we go.
    
    do {
        
            // A decimal point will fix the exponent from advancing with each digit, and set if_float.
        
        if (**scanPtr == '.')  {
            if (after_decimal == yz_true)   {
                errPtr = *scanPtr;
                return not_a_number_err;    }
            after_decimal = yz_true;
            scale_factor = 1./place_value;
            *if_float = yz_true;          }
        
        
            // read a digit
        
        else  {
            if (if_hexadecimal == yz_false)  {   // decimal
                if (lettertype(**scanPtr) != a_number)  {  errPtr = *scanPtr;   return not_a_number_err;  }
                float_to_add = (double) (**scanPtr - '0');  }
            else  {          // hexadecimal
                if (lettertype(**scanPtr) == a_number)
                    float_to_add = (double) (**scanPtr - '0');
                else  {
                    if ((**scanPtr >= 'A') && (**scanPtr <= 'F'))        // hexadigits can be either upper- or lower-case
                        float_to_add = (double) (**scanPtr - 'A' + 10);
                    else if ((**scanPtr >= 'a') && (**scanPtr <= 'f'))
                        float_to_add = (double) (**scanPtr - 'a' + 10);
                    else  {  errPtr = *scanPtr;   return not_a_number_err;  }
            }  }
            read_a_num = yz_true;
            
            
                // Add up the new digit, and adjust place values as necessary.
                
            if (after_decimal == yz_false)  {
                if (DBL_MAX/place_value < *retrn_num) {  errPtr = *scanPtr;   return overflow_err;  }
                *retrn_num *= place_value;  }
            else
                float_to_add *= scale_factor;
            
            if (DBL_MAX - float_to_add < *retrn_num)  {  errPtr = *scanPtr;   return overflow_err;  }
            *retrn_num += float_to_add;
            
            if (after_decimal == yz_true)  scale_factor /= place_value;     }
        
        lastChar = *scanPtr;        // get to the next digit
        next_char(scanPtr);      }
    while (((lettertype(**scanPtr) == a_number) || (lettertype(**scanPtr) == a_letter)
            || (**scanPtr == '.')) && (((if_hexadecimal == yz_false) && (**scanPtr != 'e') && (**scanPtr != 'E'))
                        || ((if_hexadecimal == yz_true) && (**scanPtr != 'h') && (**scanPtr != 'H'))) );
    
    if (read_a_num == yz_false)   {    // we have to have read at least one digit
        errPtr = *scanPtr;
        return not_a_number_err;   }
    
    
        // Now we check for an exponent for decimal numbers ('e' or 'E') or hexadecimal numbers (use 'h' or 'H').
        // Both exponent and hexponent are themselves base-10 integers.
    
    if (((if_hexadecimal == yz_false) && ((**scanPtr == 'e') || (**scanPtr == 'E')))
                || ((if_hexadecimal == yz_true) && ((**scanPtr == 'h') || (**scanPtr == 'H'))) )     {
        
            // get past the 'e' or 'h' and check for a minus sign
        
        lastChar = *scanPtr;
        next_char(scanPtr);
        exponent = 0;
        if (**scanPtr == '-')    {
            exp_negative = yz_true;
            lastChar = *scanPtr;
            next_char(scanPtr);    }
        else if (**scanPtr == '+')    {
            lastChar = *scanPtr;
            next_char(scanPtr);    }
        
        
            // Read in the digits (no decimal point allowed this time).
        
        do  {
            if (lettertype(**scanPtr) != a_number)    {  errPtr = *scanPtr;   return not_a_number_err;  }
            exponent *= 10;
            exponent += (long) ((**scanPtr)-'0');
            
            if ((exponent > YZ_max_exponent) && (exp_negative == yz_false))  {    // check over/underflow
                errPtr = *scanPtr;
                return overflow_err;     }
            else if ((-exponent < YZ_min_exponent) && (exp_negative == yz_true))
                {  *retrn_num = 0;  error_code = underflow_err;  }
            
            lastChar = *scanPtr;    // get to next char
            next_char(scanPtr);
        } while (lettertype(**scanPtr) == a_number);
        
        if (lettertype(**scanPtr) == a_letter)   {  errPtr = *scanPtr;   return not_a_number_err;  }
        if (exp_negative == yz_true) exponent = -exponent;
        
        
            // Shift the number by the 10^e-power (exponent) or 16^h-power (hexponent).
        
        if (exponent < 0)  {            // avoid underflow errors near double-precision limit
            exponent++;
            divide_by_base = yz_true;       }
        
        if (if_hexadecimal == yz_false)  temp_float = pow(10., exponent);
        else  temp_float = pow(16., exponent);
        
        if (temp_float < DBL_MIN) error_code = underflow_err;
        else if ((exp_negative == yz_false) && (DBL_MAX/temp_float < *retrn_num))  {
            errPtr = *scanPtr;
            return overflow_err;    }
        
        if (*retrn_num != 0.)    {
            *retrn_num *= temp_float;
            if (divide_by_base == yz_true)  {
                if (if_hexadecimal == yz_false)  *retrn_num /= 10.;
                else  *retrn_num /= 16.;        }
            if (*retrn_num < DBL_MIN) error_code = underflow_err;     }
    }
    
    
        // Fix the sign; test whether a double or slong is the more appropriate type.
        // Machines with 8-byte integers will find some calculations more exact with slongs --
        // those will not be read exactly using this routine at present.
    
    if (negative == yz_true)  *retrn_num *= -1.;
    long_num = (signed long) *retrn_num;
    if (*retrn_num != long_num)  *if_float = yz_true;
    
    *scanPtr = lastChar;    // to make sure we don't skip overshoot with our next get_next_word() call
    return error_code;      // any soft errors (warnings)
}


// FindLine() searches the NewlinePositions LL for the line containing
// either a given source character index or bytecode long-word index.

Ulong FindLine(linkedlist *theList, Ulong SourceIndex, Ulong HobbishIndex)
{
    Ulong top = theList->elementnum, bottom = 1, top_value, bottom_value, mid, mid_value, desired_value, offset;
    
    if (HobbishIndex == 0)  {  offset = 0;  desired_value = SourceIndex;  }
    else  {  offset = 1;  desired_value = HobbishIndex;  }

    if (top == 0)  return 0;       // we constrain top and bottom as we search
    top_value = *(LL_ULong(theList, top) + offset);
    bottom_value = *(LL_ULong(theList, bottom) + offset);
    
    if (top_value <= desired_value)  return top;
    
    
        // Search by bisection -- both character indices and bytecode indices(?) should be in order from low-to-high
        
    while (top > bottom+1)   {
        mid = bottom + (desired_value-bottom_value)*(top-bottom)/(top_value-bottom_value);        
        if (mid <= bottom)  mid = bottom+1;      // so we don't resample top or bottom
        else if (mid >= top)  mid = top-1;
        
        mid_value = *(LL_ULong(theList, mid) + offset);

        if (mid_value < desired_value)  {
            bottom = mid;
            bottom_value = mid_value;     }
        else if (mid_value > desired_value)  {
            top = mid;
            top_value = mid_value;     }
        else  return mid;
    }
    
    return bottom;
}



//   ********* Low-level Parsing routines **********


// get_start_of_line() scans to the beginning of the present line (or next line if there is only whitespace/comments)

parse_retrn get_start_of_line(char **scanPtr, linkedlist *CodeLLPtr, Ulong *CodeElement)
{
    parse_retrn retrn;

    if (**scanPtr == '|')  next_char(scanPtr);         // next_char skips a comment \* block
    
    while ((lettertype(**scanPtr) == a_eol) || (lettertype(**scanPtr) == a_space))  {
        if ((**scanPtr == ';') && (CodeLLPtr != NULL))  InsertCodeWord(CodeLLPtr, CodeElement, code_marker);
        next_char(scanPtr);
        if (lettertype(**scanPtr) == a_space)   {
            retrn = get_next_word(scanPtr);
            if (retrn != passed)  return retrn;  }   }

    if (**scanPtr == '&')  get_next_word(scanPtr);

    return passed;
}


// get_next_word() skips to end of current word/number, if *scanPtr is on one, and brings *scanPtr up to next word/symbol/etc.

parse_retrn get_next_word(char **scanPtr)
{
    Boolean line_skipped = yz_false;


        // first, get past current word/number

    if ((lettertype(**scanPtr) == a_symbol) && (**scanPtr != '&'))
        next_char(scanPtr);
    else while ((lettertype(**scanPtr) == a_letter) || (lettertype(**scanPtr) == a_number))
        next_char(scanPtr);


        // second, skip all the spaces

    while ((lettertype(**scanPtr) == a_space) || (lettertype(**scanPtr) == unprintable))
        next_char(scanPtr);


        // skip over as many blank lines as we need to

    while (**scanPtr == '&')    {
        next_char(scanPtr);
        while ((lettertype(**scanPtr) == a_space) || ((lettertype(**scanPtr) == a_eol) && (**scanPtr != ',') && (**scanPtr != ';')))
            {  if (lettertype(**scanPtr) == a_eol) line_skipped = yz_true;  next_char(scanPtr);  }
        if (line_skipped == yz_false)   {
            errPtr = *scanPtr;
            return eol_expected_err;   }    }

    return passed;
}


// next_char() moves *scanPtr up to the next printable character.
//  Same as next_string_char, but will skip over comment blocks.

void next_char(char **scanPtr)
{
    if (**scanPtr != '|')  next_string_char(scanPtr);


        // a comment; we hit a '|'

    while (**scanPtr == '|')  {
        
        next_string_char(scanPtr);
        
        if (**scanPtr == '*')    {      // case 1:  a comment block  |* ... *| -- skip to end
            Boolean one_star = yz_false;
            
            next_string_char(scanPtr);
            while ((lettertype(**scanPtr) != a_null) && ((**scanPtr != '|') || (one_star == yz_false)))    {
                while ((lettertype(**scanPtr) != a_null) && (**scanPtr != '*'))  next_string_char(scanPtr);
                if (lettertype(**scanPtr) != a_null)  {  next_string_char(scanPtr);  one_star = yz_true;  }
            }
            if (**scanPtr == '|') next_string_char(scanPtr);     // this both closes |* *| or step past the to-end-of-line | mark
        }
        
        else    {               // case 2:  a single-line comment -- skip past EOL
        while ((lettertype(**scanPtr) != a_null) && ((**scanPtr == ',') || (**scanPtr == ';') || (lettertype(**scanPtr) != a_eol)))  {
            next_string_char(scanPtr);
        }}
    }

    return;
}


// next_string_char() skips to the next character; ignores unprintable ones.
// Used for parsing strings when we don't have to worry about comments.

void next_string_char(char **scanPtr)
{
    if (lettertype(**scanPtr) != a_null) do  {
        (*scanPtr)++;
    } while (lettertype(**scanPtr) == unprintable);

    return;
}


// numeric_character() returns true or false according to whether theChar could be part of a number or not.
// Used for determining whether to read a character sequence as a number by looking only at its first letter.

Boolean numeric_character(char theChar)
{
    if ((lettertype(theChar) == a_number) || (theChar == '$') || (theChar == '.') || (theChar == '-'))
        return yz_true;

    return yz_false;
}

// bracket_character() returns true or false according to whether theChar is an opening or closing parenthesis, bracket, or brace.

Boolean bracket_character(char theChar)
{
    if ((theChar == '(') || (theChar == ')') || (theChar == '[') || (theChar == ']') || (theChar == '{') || (theChar == '}'))
        return yz_true;

    return yz_false;
}


// lettertype() returns the 'type' of the character beneath scanPtr
// letter_types are defined in cmpile.h.

letter_type lettertype(char theChar)
{
    if ((theChar == ' ') || (theChar == '\t'))  return a_space;
    if ((theChar == '\n') || (theChar == 13) || (theChar == ',') || (theChar == ';'))  return a_eol;
    if (theChar == '\0')  return a_null;
    if ((theChar >= '0') && (theChar <= '9'))  return a_number;
    if ( ((theChar <= 'z')&&(theChar >= 'a')) || ((theChar <= 'Z')&&(theChar >= 'A')) || (theChar == '_'))
        return a_letter;
    if ((theChar <= 31) || (theChar == 127))  return unprintable;
    return a_symbol;
}
