/*
 *  cmpile.h
 *
 *  Yazoo
 *  Copyright (C) 2005 Brian C. Ross
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of the License, or any later version.

 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef Compile_h
#define Compile_h

#include "yazooh.h"


// letter_types are defined here

#define letter_type char
#define a_space 0
#define a_eol 1
#define a_null 2
#define a_letter 3
#define a_number 4
#define a_symbol 5
#define unprintable 6


// Compile-time error codes

#define parse_retrn unsigned char

#define memory_err 1  /* mem_err for LLs is also a 1, so it is consistent  */

#define unexpected_eof_err 2
#define eol_expected_err 3
#define eof_expected_err 4

#define unknown_command_err 5
#define invalid_command_err 6
#define arg_expected_err 7
#define no_left_arg_allowed_err 8
#define no_right_arg_allowed_err 9

#define var_or_num_expected_err 10
#define variable_expected_err 11
#define right_bracket_expected_err 12
#define right_brace_expected_err 13
#define right_parenthesis_expected_err 14

#define overflow_err 15
#define underflow_err 16
#define not_a_number_err 17
#define string_read_err 18
#define Yazoo_function_err 19

#define must_follow_if_err 20
#define conditional_expected_err 21
#define endif_expected_err 22
#define endfor_expected_err 23
#define endwhile_expected_err 24
#define until_expected_err 25
#define in_expected_err 26
#define left_bracket_expected_err 27
#define comma_expected_err 28
#define illegal_control_command_err 29
#define unexpected_control_terminator_err 30
#define not_a_statement_err 31

#define unknown_err 32



// The control command IDs

#define if_Ccmd 0
#define elseif_Ccmd 1
#define else_Ccmd 2
#define endif_Ccmd 3
#define endif_Ccmd_B 4
#define while_Ccmd 5
#define endwhile_Ccmd 6
#define endwhile_Ccmd_B 7
#define do_Ccmd 8
#define until_Ccmd 9
#define for_Ccmd 10
#define in_Ccmd 11
#define step_Ccmd 12
#define endfor_Ccmd 13
#define endfor_Ccmd_B 14

#define not_a_Ccmd (control_commands_num+1)



// The various order-of-operations levels (lower ones are evaluated first)

#define OoO_nolevel 100
#define OoO_no_args 0
#define OoO_stepvar 3
#define OoO_pow 4
#define OoO_negate 5
#define OoO_muldiv 6
#define OoO_addsub 7
#define OoO_block 8
#define OoO_append 9
#define OoO_subcode 10
#define OoO_define 12
#define OoO_cond_test 13
#define OoO_lnot 14
#define OoO_lbool 15
#define OoO_command 16


// Order-of-operations for same-OoO operations in series

#define l_to_r 1    /* (a+b)+c  */
#define r_to_l 2    /* a=(b=c)  */



// Pseudo-commands used by the compiler

#define negatef (commands_num+2)
#define define (commands_num+3)
#define var_define (commands_num+4)
#define member_define (commands_num+5)
#define proxy_define (commands_num+6)
#define def_equate_at (commands_num+7)
#define equate_at (commands_num+8)
#define equate (commands_num+9)
#define def_equate (commands_num+10)



// Numbers corresponding to the def-general flags; flags word = appropriate sum of these numbers

#define dg_proxy (1<<8)
#define dg_unjammable (1<<7)
#define dg_hidden_member (1<<6)
#define dg_run_constructor (1<<5)
#define dg_relink_target (1<<4)
#define dg_new_target (1<<3)
#define dg_add_members (1<<2)
#define dg_update_members (1<<1)
#define dg_post_equate 1
#define dg_bad_flags 255


// The flags words for the various def-general operators

#define def_flags (dg_add_members+dg_update_members+dg_new_target+dg_run_constructor)
#define vdf_flags (dg_add_members+dg_new_target+dg_run_constructor)
#define mdf_flags (dg_update_members+dg_add_members)
#define pdf_flags (dg_add_members+dg_update_members+dg_new_target+dg_run_constructor+dg_proxy)
#define dqa_flags (dg_add_members+dg_update_members+dg_relink_target)
#define eqa_flags (dg_relink_target)
#define equ_flags (dg_post_equate)
#define deq_flags (dg_add_members+dg_update_members+dg_new_target+dg_post_equate+dg_run_constructor)




// Misc definitions

#define YZ_max_exponent 308
#define YZ_min_exponent -324




// Global struct definition

typedef struct {
    char *_errPtr, *_DocStartPtr;
    int _doublesize_in_longs;
    linkedlist _AllUserNames, _NewlinePositions;
    Ulong _NullNamesNum;
} yz_compile_global_struct;

extern yz_compile_global_struct yz_compile_globals;

#define errPtr yz_compile_globals._errPtr
#define DocStartPtr yz_compile_globals._DocStartPtr
#define doublesize_in_longs yz_compile_globals._doublesize_in_longs
#define AllUserNames yz_compile_globals._AllUserNames
#define NewlinePositions yz_compile_globals._NewlinePositions
#define NullNamesNum yz_compile_globals._NullNamesNum



// External prototypes

extern parse_retrn InitCompile(Ulong);
extern void FinishCompile();
extern parse_retrn Compile(char **, linkedlist *);
extern parse_retrn ParseStatements(char **, linkedlist *, Ulong *, Ulong, Ulong);
extern parse_retrn AddLine(char *, Ulong);
extern parse_retrn ParseExpression(char **, char *, Slong *, Slong *, linkedlist *, Ulong *);
extern parse_retrn ReadVarName(char **, Slong *);
extern parse_retrn ReadString(char **, linkedlist *, Ulong *);
extern parse_retrn InsertCodeWord(linkedlist *, Ulong *, Slong);
extern parse_retrn AddToCLLP(linkedlist *, Ulong *, Ulong);
extern parse_retrn InsertPrefix(linkedlist *, Ulong *, Slong, Boolean);
extern parse_retrn WriteNum(linkedlist *, Ulong *, double, Boolean);
extern char OoOValue(char c);
extern parse_retrn GetOperation(char **, char *, Slong *, Slong *, linkedlist *, Ulong *, Slong, Boolean, Ulong *);
extern Boolean if_expression_terminator(char);
extern Ulong get_control_command(char **);
extern Boolean CompareString(char **, char *);
extern parse_retrn GetNewWord(char *, char **);
extern parse_retrn ReadNum(char **, double *, Boolean *);
extern Ulong FindLine(linkedlist *, Ulong, Ulong);
extern parse_retrn get_start_of_line(char **, linkedlist *, Ulong *);
extern parse_retrn get_next_word(char **);
extern void next_char(char **);
extern void next_string_char(char **);
extern Boolean numeric_character(char);
extern Boolean bracket_character(char);
extern letter_type lettertype(char);


#endif
