/*
 *  intrpt.c(pp) - low-level routines for interpreting bytecode
 *
 *  Yazoo
 *  Copyright (C) 2005 Brian C. Ross
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of the License, or any later version.

 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <math.h>
#include "yazooh.h"
#include "intrpt.h"
#include "hobbsh.h"
#include "cmpile.h"
#include "extrnl.h"
#include "userfn.h"
#include "yzmain.h"
#include <float.h>
#include <limits.h>



// * * * * *  Interpreter globals  * * * * *

yz_interpret_global_struct yz_interpret_globals;     // see intrpt.h


    // next: the amount of extra space allocated in linked lists
    // these LLs are used for variable data, member lists, code lists, ... -- almost everything
    // LLFreeSpace is set in InitYazoo(); the default is 100%, which the user can change (0-255)

const unsigned char LLFreeSpace = 100;



// * * * * *  Low-level routines for copying and comparing data  * * * * *


// Next 4 routines:  called by def-general (equate, etc.)
// These copy data between two type-matched bundles.

void CopyBundleData(bundle *SourceBundle, Ulong SourceOffset, bundle *DestBundle, Ulong DestOffset, Ulong IndexWidth)
{   CCTemplate(SourceBundle, SourceOffset, DestBundle, DestOffset, IndexWidth, &CopyBundleData, &Copy_Data, &Copy_String, &Copy_Numeric);  }

void Copy_Data(linkedlist *SourceLL, Ulong SourceIndex, linkedlist *DestLL, Ulong DestIndex, Ulong NumberToCopy)
{  FCopyElements(SourceLL, SourceIndex, DestLL, DestIndex, NumberToCopy);  }

void Copy_String(linkedlist *SourceStringLL, linkedlist *DestStringLL)
{
    LL_retrn linkedLretrn;
    
    if (SourceStringLL->memory == NULL)  {
        if (DestStringLL->memory != NULL)  FDeleteLinkedList(DestStringLL);    }
    else  {
        if ((DestStringLL->memory != NULL) && (DestStringLL->elementnum != SourceStringLL->elementnum))  FDeleteLinkedList(DestStringLL);
        if (DestStringLL->memory == NULL)  {
            linkedLretrn = FNewLinkedList(DestStringLL, SourceStringLL->elementnum, sizeof(char), 0, uncleared);
            if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }      }
        FCopyElements(SourceStringLL, 1, DestStringLL, 1, SourceStringLL->elementnum);    }
}

void Copy_Numeric(linkedlist *SourceLL, Ulong SourceIndex, unsigned char SourceType, linkedlist *DestLL, Ulong DestIndex, unsigned char DestType)
{
    LoadDPRegister(FElement(SourceLL, SourceIndex), SourceType);
    SaveDPRegister(FElement(DestLL, DestIndex), DestType);
    if (err_code == out_of_range_err)  {  err_code = passed;  SetWarning(out_of_range_err, PCCodePtr-1);  }
}


// Next 4 routines:  called by the comparison operator (==)
// Compare data between two type-matched bundles to test for equality.

void CompareBundleData(bundle *SourceBundle, Ulong SourceOffset, bundle *DestBundle, Ulong DestOffset, Ulong IndexWidth)
{   CCTemplate(SourceBundle, SourceOffset, DestBundle, DestOffset, IndexWidth, &CompareBundleData, &Compare_Data, &Compare_String, &Compare_Numeric);  }

void Compare_Data(linkedlist *SourceLL, Ulong SourceIndex, linkedlist *DestLL, Ulong DestIndex, Ulong NumberToCompare)
{
    if (GL_Conditional == yz_false)  return;
    if (SourceLL->elementsize != DestLL->elementsize)  GL_Conditional = yz_false;
    else if (FCompareElements(SourceLL, SourceIndex, DestLL, DestIndex, NumberToCompare) == notequal)  GL_Conditional = yz_false;
}

void Compare_String(linkedlist *SourceStringLL, linkedlist *DestStringLL)
{
    if (GL_Conditional == yz_false)  return;
    if (((SourceStringLL->memory == NULL) && (DestStringLL->memory != NULL)) || ((SourceStringLL->memory != NULL) && (DestStringLL->memory == NULL)))
                    {  GL_Conditional = yz_false;  return;  }
    if (SourceStringLL->elementnum != DestStringLL->elementnum)  {  GL_Conditional = yz_false;  return;  }

    if (FCompareElements(SourceStringLL, 1, DestStringLL, 1, SourceStringLL->elementnum) == notequal)  GL_Conditional = yz_false;
}

void Compare_Numeric(linkedlist *SourceLL, Ulong SourceIndex, unsigned char SourceType, linkedlist *DestLL, Ulong DestIndex, unsigned char DestType)
{
    double FirstNumber;
    
    if (GL_Conditional == yz_false)  return;
    
    LoadDPRegister(FElement(SourceLL, SourceIndex), SourceType);
    FirstNumber = DPRegister;
    LoadDPRegister(FElement(DestLL, DestIndex), DestType);
    if (FirstNumber != DPRegister)  GL_Conditional = yz_false;
}


// CCTemplate() explores two bundle trees simultaneously, in order to perform a given operation between them.
// This is used for copying and comparing data.
// This routine skips hidden members.

void CCTemplate(bundle *SourceBundle, Ulong SourceOffset, bundle *DestBundle, Ulong DestOffset, Ulong IndexWidth,
        void(*CC_Composite)(bundle *, Ulong, bundle *, Ulong, Ulong),
        void(*CC_Data)(linkedlist *, Ulong, linkedlist *, Ulong, Ulong),
        void(*CC_String)(linkedlist *, linkedlist *),
        void(*CC_Numeric)(linkedlist *, Ulong, unsigned char, linkedlist *, Ulong, unsigned char))
{
    bundle *NextSourceBundle, *NextDestBundle;
    member *loopSourceMember, *loopDestMember;
    linkedlist *SourceLL, *DestLL;
    Ulong NextSourceOffset, NextDestOffset, NextIndices, SourceMemberCounter, DestMemberCounter, IndexCounter;
    unsigned char SourceType, DestType;
    Boolean IfProxy;
    
    if (IfBusy(SourceBundle, busy_source_copy_flag) != yz_false)  {  SetError(self_reference_err, PCCodePtr-1);  return;  }
    if (IfBusy(DestBundle, busy_dest_copy_flag) != yz_false)  {  SetError(self_reference_err, PCCodePtr-1);  return;  }
    SetBusy(SourceBundle, busy_source_copy_flag);
    SetBusy(DestBundle, busy_dest_copy_flag);

    if (SourceBundle->variable_ptr->type == composite_type)  {
        if (DestBundle->variable_ptr->type != composite_type)  SetError(type_mismatch_err, PCCodePtr-1);
        if (err_code != passed)  {  
            ClearBusy(SourceBundle, busy_source_copy_flag);
            ClearBusy(DestBundle, busy_dest_copy_flag);
            return;      }

        SourceMemberCounter = 0;
        DestMemberCounter = 0;
        FindNextThickMember(SourceBundle->variable_ptr, &loopSourceMember, &SourceMemberCounter);
        FindNextThickMember(DestBundle->variable_ptr, &loopDestMember, &DestMemberCounter);
        
        while (SourceMemberCounter <= SourceBundle->variable_ptr->mem.members.elementnum)     {
            if (DestMemberCounter > DestBundle->variable_ptr->mem.members.elementnum)  SetError(type_mismatch_err, PCCodePtr-1);
            else if (loopSourceMember->indices != loopDestMember->indices)  SetError(type_mismatch_err, PCCodePtr-1);
            if (err_code != passed)  {  
                ClearBusy(SourceBundle, busy_source_copy_flag);
                ClearBusy(DestBundle, busy_dest_copy_flag);
                return;        }

            IfProxy = yz_false;
            if (loopSourceMember->member_ptr != NULL)
                if (loopSourceMember->member_ptr->variable_ptr->type == proxy_type)  IfProxy = yz_true;
            if (loopDestMember->member_ptr != NULL)
                if (loopDestMember->member_ptr->variable_ptr->type == proxy_type)  IfProxy = yz_true;
            if (IfProxy == yz_true)      {          // then copy each offset individually

                for (IndexCounter = 0; IndexCounter < IndexWidth*loopSourceMember->indices; IndexCounter++)   {
                    NextSourceBundle = loopSourceMember->member_ptr;
                    NextDestBundle = loopDestMember->member_ptr;
                    NextSourceOffset = SourceOffset;
                    NextDestOffset = DestOffset;
                    NextIndices = 1;            // a dummy variable
            
                    StepBundle(&NextSourceBundle, &NextSourceOffset, &NextIndices, loopSourceMember, IndexCounter, 1);
                    StepBundle(&NextDestBundle, &NextDestOffset, &NextIndices, loopDestMember, IndexCounter, 1);
                    if (err_code == passed)  CC_Composite(NextSourceBundle, NextSourceOffset, NextDestBundle, NextDestOffset, 1);
                    if (err_code != passed)  {  
                        if ((NextSourceBundle == NULL) && (NextDestBundle == NULL))  err_code = passed;
                        else   {
                            ClearBusy(SourceBundle, busy_source_copy_flag);
                            ClearBusy(DestBundle, busy_dest_copy_flag);
                            return;
                    }   }
            }   }
            
            else    {        // we can just do them all at once, which is preferable since it is much faster
                NextSourceBundle = loopSourceMember->member_ptr;
                NextDestBundle = loopDestMember->member_ptr;
                NextSourceOffset = SourceOffset;
                NextDestOffset = DestOffset;
                
                StepBundle(&NextSourceBundle, &NextSourceOffset, &NextIndices, loopSourceMember, 0, 1);
                StepBundle(&NextDestBundle, &NextDestOffset, &NextIndices, loopDestMember, 0, 1);
                if (err_code == passed)
                    CC_Composite(NextSourceBundle, NextSourceOffset, NextDestBundle, NextDestOffset, IndexWidth*loopSourceMember->indices);
                if (err_code != passed)  {  
                    if ((NextSourceBundle == NULL) && (NextDestBundle == NULL))  err_code = passed;
                    else    {
                        ClearBusy(SourceBundle, busy_source_copy_flag);
                        ClearBusy(DestBundle, busy_dest_copy_flag);
                        return;   }     }
            }

            FindNextThickMember(SourceBundle->variable_ptr, &loopSourceMember, &SourceMemberCounter);
            FindNextThickMember(DestBundle->variable_ptr, &loopDestMember, &DestMemberCounter);     }

        if (DestMemberCounter <= DestBundle->variable_ptr->mem.members.elementnum)  SetError(type_mismatch_err, PCCodePtr-1);    }
    
    else    {          // it's a primitive variable
        if ((SourceBundle->variable_ptr->type == DestBundle->variable_ptr->type) && (SourceBundle->variable_ptr->type != string_type))    {

                // if type-matched, can copy/compare all at once -- fastest option

            if (SourceBundle->variable_ptr->mem.data.elementsize != DestBundle->variable_ptr->mem.data.elementsize)
                        SetError(type_mismatch_err, PCCodePtr-1);          // could happen if they are both block_types

            else  CC_Data(&(SourceBundle->variable_ptr->mem.data), SourceOffset+1, &(DestBundle->variable_ptr->mem.data), DestOffset+1, IndexWidth);   }

        else if (DestBundle->variable_ptr->type == composite_type)  SetError(type_mismatch_err, PCCodePtr-1);
        
        else   {
        
                // if not type-matched (or if it's strings) then we have to copy/compare each element individually
        
            SourceLL = &(SourceBundle->variable_ptr->mem.data);
            DestLL = &(DestBundle->variable_ptr->mem.data);
            SourceType = SourceBundle->variable_ptr->type;
            DestType = DestBundle->variable_ptr->type;
            if ((SourceType == block_type) || (DestType == block_type))  SetError(type_mismatch_err, PCCodePtr-1);

            else if (SourceType == string_type)  {
                if (DestType != string_type)  SetError(type_mismatch_err, PCCodePtr-1);
                else for (IndexCounter = 1; IndexCounter <= IndexWidth; IndexCounter++)  {
                    CC_String((linkedlist *) FElement(SourceLL, SourceOffset+IndexCounter), (linkedlist *) FElement(DestLL, DestOffset+IndexCounter));
            }   }
            else   {
                if (DestType == string_type)  SetError(type_mismatch_err, PCCodePtr-1);
                else for (IndexCounter = 1; IndexCounter <= IndexWidth; IndexCounter++)  {
                    CC_Numeric(SourceLL, SourceOffset+IndexCounter, SourceType, DestLL, DestOffset+IndexCounter, DestType);
                    if (err_code != passed)  break;
        }   }   }
    }

    ClearBusy(SourceBundle, busy_source_copy_flag);
    ClearBusy(DestBundle, busy_dest_copy_flag);
}


// Returns the next non-hidden member of a given variable

void FindNextThickMember(variable *theVariable, member **theMember, Ulong *MemberCounter)
{
    do  {
        (*MemberCounter)++;
        if (*MemberCounter > theVariable->mem.members.elementnum)  return;
        *theMember = (member *) FElement(&(theVariable->mem.members), *MemberCounter);
    } while ((*theMember)->if_hidden == yz_true);
}



// * * * * *  Functions operating on variables, including all members thereof  * * * * *


// Next 3 routines:  used by size(), as well as forced-equate for 'type-checking'.
// Here we calculate the size in bytes of a bundle (including its members, ...)

void SizeBundle(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *DataSize, void *StringSize)
{  SRWTemplate(theBundle, BundleOffset, IndexWidth, DataSize, StringSize, yz_false, &SizeBundle, &Size_Data, &Size_String);  }

void Size_Data(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *DataSize, void *Dummy)
{
    Ulong SizeToAdd = theBundle->variable_ptr->mem.data.elementsize*IndexWidth;

    if (ExtCallMode == full_call)  SizeToAdd += Align(SizeToAdd);
    *(Ulong *) DataSize += SizeToAdd;
}

void Size_String(bundle *theBundle, Ulong StringIndex, void *DataSize, void *StringSize)
{
    linkedlist *StringLL = (linkedlist *) FElement(&(theBundle->variable_ptr->mem.data), StringIndex);

    if (*(signed long *) StringSize == no_string)  *(signed long *) StringSize = 0;
    if (StringLL->memory == NULL)  return;
    *(signed long *) StringSize += StringLL->elementnum;
}


// Next 3 routines:  also used by size().
// These differ from the last in that storage shared between multiple members only gets counted once.

void StorageSizeBundle(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *DataSize, void *StringSize)
{  SRWTemplate(theBundle, BundleOffset, IndexWidth, DataSize, StringSize, yz_false, &StorageSizeBundle, &StorageSize_Data, &StorageSize_String);  }

void StorageSize_Data(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *DataSize, void *Dummy)
{
    Ulong bottom = BundleOffset, cB;
    signed long top;
    bundle *loopBundle;
    pinned_LL *BundlesPLL = &(theBundle->variable_ptr->bundles);

    if (IfBusy(theBundle, busy_SS_flag) != yz_false)  return;

    while (bottom < BundleOffset+IndexWidth)     {    // add up free memory blocks
        top = BundleOffset+IndexWidth;
        for (cB = 1; cB <= BundlesPLL->data.elementnum; cB++)       {
        if (*LL_ULong(&(BundlesPLL->references), cB) != 0)      {    // partially-overlapping bundles change the top/bottom of the block
            loopBundle = (bundle *) FElement(&(BundlesPLL->data), cB);
            if (IfBusy(loopBundle, busy_SS_flag) != yz_false)   {
                if ((loopBundle->bundle_offset <= bottom) && (loopBundle->bundle_offset+loopBundle->instances >= bottom))
                    bottom = loopBundle->bundle_offset+loopBundle->instances+1;
                if ((loopBundle->bundle_offset <= top) && (loopBundle->bundle_offset+loopBundle->instances >= top))
                    top = loopBundle->bundle_offset-1;
        }}  }
    
        if ((bottom <= top) && (top >= 0))
            *(Ulong *) DataSize += (top - bottom) * theBundle->variable_ptr->mem.data.elementsize;
        bottom = top+1;
    }
    
    SetBusy(theBundle, busy_SRW_flag);
}

void StorageSize_String(bundle *theBundle, Ulong StringIndex, void *DataSize, void *StringSize)
{
    Ulong cB;
    bundle *loopBundle;
    linkedlist *StringLL = (linkedlist *) FElement(&(theBundle->variable_ptr->mem.data), StringIndex);
    pinned_LL *BundlesPLL = &(theBundle->variable_ptr->bundles);

    if (StringLL->memory == NULL)  return;

    for (cB = 1; cB <= BundlesPLL->data.elementnum; cB++)       {
    if (*LL_ULong(&(BundlesPLL->references), cB) != 0)      {
        loopBundle = (bundle *) FElement(&(BundlesPLL->data), cB);
        if (IfBusy(loopBundle, busy_SS_flag) != yz_false)   {     // only add if it's not overlapping w/ another flagged bundle
            if ((loopBundle->bundle_offset <= StringIndex) && (loopBundle->bundle_offset+loopBundle->instances >= StringIndex))
                return;
    }}  }

    if (*(signed long *) StringSize == no_string)  *(signed long *) StringSize = 0;
    *(signed long *) StringSize += StringLL->elementnum;
}


// Next 3 routines:  used by forced-equate.
// These copy data from a source buffer into a bundle.

void WriteBundle(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *BufferPtr, void *StringSize)        // writes the variables
{  SRWTemplate(theBundle, BundleOffset, IndexWidth, BufferPtr, StringSize, yz_true, &WriteBundle, &Write_Data, &Write_String);  }

void Write_Data(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *BufferPtr, void *Dummy)
{
    linkedlist *theLL = &(theBundle->variable_ptr->mem.data);
    
    FSetElements(theLL, BundleOffset+1, BundleOffset+IndexWidth, *(void **) BufferPtr);
    *(char **) BufferPtr += theLL->elementsize*IndexWidth;
    if (ExtCallMode == full_call)  *(char **) BufferPtr += Align(theLL->elementsize*IndexWidth);
}

void Write_String(bundle *theBundle, Ulong StringIndex, void *BufferPtr, void *StringSize)
{
    linkedlist *StringLL = (linkedlist *) FElement(&(theBundle->variable_ptr->mem.data), StringIndex);
    LL_retrn linkedLretrn;

    if (StringLL->memory != NULL)  FDeleteLinkedList(StringLL);
    if (*(signed long *) StringSize == 0)  return;
    linkedLretrn = FNewLinkedList(StringLL, (Ulong) *(signed long *) StringSize, sizeof(char), 0, uncleared);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }

    FSetElements(StringLL, 1, StringLL->elementnum, *(void **) BufferPtr);
    *(char **) BufferPtr += *(signed long *) StringSize;
    *(signed long *) StringSize = 0;
}


// Next 3 routines:  used by forced-equate.
// These copy data from a bundle into a target memory buffer.

void ReadBundle(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *BufferPtr, void *StringSize)
{  SRWTemplate(theBundle, BundleOffset, IndexWidth, BufferPtr, StringSize, yz_true, &ReadBundle, &Read_Data, &Read_String);  }

void Read_Data(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *BufferPtr, void *Dummy)
{
    linkedlist *theLL = &(theBundle->variable_ptr->mem.data);

    FGetElements(theLL, BundleOffset+1, BundleOffset+IndexWidth, *(void **) BufferPtr);
    *(char **) BufferPtr += theLL->elementsize*IndexWidth;
    if (ExtCallMode == full_call)  *(char **) BufferPtr += Align(theLL->elementsize*IndexWidth);
}

void Read_String(bundle *theBundle, Ulong StringIndex, void *BufferPtr, void *StringSize)
{
    linkedlist *StringLL = (linkedlist *) FElement(&(theBundle->variable_ptr->mem.data), StringIndex);
    
    if (StringLL->memory == NULL)  return;
    FGetElements(StringLL, 1, StringLL->elementnum, *(void **) BufferPtr);
    *(char **) BufferPtr += StringLL->elementnum;
}


// Next 2 routines, along with ReadString() (above):  used by print_string().
// These copy data from a bundle into a buffer, writing out any numeric data in ASCII format (rather than binary)

void ReadBundleString(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *StringPtr, void *StringSize)
{  SRWTemplate(theBundle, BundleOffset, IndexWidth, StringPtr, StringSize, yz_true, &ReadBundleString, &Read_DataString, &Read_String);  }

void Read_DataString(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *DestString, void *Dummy)
{
    Ulong counter, character_num;

    if (theBundle->variable_ptr->type == block_type)  {
        Read_Data(theBundle, BundleOffset, IndexWidth, DestString, Dummy);
        return;       }
    for (counter = 1; counter <= IndexWidth; counter++)  {
        LoadDPRegister(FElement(&(theBundle->variable_ptr->mem.data), BundleOffset+counter), theBundle->variable_ptr->type);
        PrintNumber(*(char **) DestString, DPRegister, &character_num, theBundle->variable_ptr->type);
        *(char **) DestString += character_num;          }
}


// Next 2 routines:  invoked by print_string()
// They calculate the size in bytes of a string that will be the 'printout' of a bundle
// (i.e. with numbers written out in ASCII).

void SizeBundleString(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *DataSize, void *StringSize)
{  SRWTemplate(theBundle, BundleOffset, IndexWidth, DataSize, StringSize, yz_false, &SizeBundleString, &Size_DataString, &Size_String);  }

void Size_DataString(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *DataSize, void *Dummy)
{
    Ulong counter, character_num;

    if (theBundle->variable_ptr->type == block_type)  {
        *(Ulong *) DataSize += theBundle->variable_ptr->mem.data.elementsize*IndexWidth;
        return;         }

    for (counter = 1; counter <= IndexWidth; counter++)  {
        LoadDPRegister(FElement(&(theBundle->variable_ptr->mem.data), BundleOffset+counter), theBundle->variable_ptr->type);
        PrintNumber(NULL, DPRegister, &character_num, theBundle->variable_ptr->type);
        *(Ulong *) DataSize += character_num;            }
}


// Next 3 routines:  called by read_string()
// Copies data from the source string into the target bundle.
// Numbers are parsed and converted from ASCII to binary in the process.

void WriteBundleString(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *StringPtr, void *OverflowMarker)
{  SRWTemplate(theBundle, BundleOffset, IndexWidth, StringPtr, OverflowMarker, yz_true, &WriteBundleString, &Write_DataString, &Write_StringToString);  }

void Write_DataString(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *SourceString, void *OverflowMarker)
{
    Ulong counter;
    char OneLetterType;
    parse_retrn retrn;
    letter_type CharType = lettertype(**(char **) SourceString);
    Boolean dummy_if_float;
    
    
        // white space in the source string is skipped over
        // (lettertype(), a_space, etc. are defined in cmpile.c(pp)/h)
    
    while ((CharType == a_space) || (CharType == a_eol) || (CharType == unprintable))
        {  (*(char **) SourceString)++;  CharType = lettertype(**(char **) SourceString);  }
    
    
        // block types in the target bundle are written directly from the next n bytes of the string
    
    if (theBundle->variable_ptr->type == block_type)  {
        if (*((char **) SourceString) + theBundle->variable_ptr->mem.data.elementsize*IndexWidth > (char *) OverflowMarker)
            {  SetError(type_mismatch_err, PCCodePtr-1);  return;  }
        FGetElements(&(theBundle->variable_ptr->mem.data), BundleOffset+1, BundleOffset+IndexWidth, *(void **) SourceString);
        *(char **) SourceString += theBundle->variable_ptr->mem.data.elementsize*IndexWidth;            }
        
        
        // numeric types in the target bundle are parsed using ReadNum() (defined in cmpile.c(pp))
    
    else for (counter = 1; counter <= IndexWidth; counter++)  {
        retrn = ReadNum((char **) SourceString, &DPRegister, &dummy_if_float);
        if (retrn == underflow_err)  SetWarning(underflow_warning, PCCodePtr-1);
        else if (retrn != passed)  {  SetWarning(read_number_err, PCCodePtr-1);  return;  }
        do {  (*(char **) SourceString)++; OneLetterType = lettertype(**(char **) SourceString);  }
        while ((OneLetterType != a_symbol) && (OneLetterType != a_number) && (OneLetterType != a_letter) &&
                                                    (*(char **) SourceString < (char *) OverflowMarker));
        SaveDPRegister(FElement(&(theBundle->variable_ptr->mem.data), BundleOffset+counter), theBundle->variable_ptr->type);
        if (err_code == out_of_range_err)  {  err_code = passed;  SetWarning(out_of_range_err, PCCodePtr-1);     }       }
}

void Write_StringToString(bundle *theBundle, Ulong StringIndex, void *SourcePtr, void *OverflowMarker)
{
    linkedlist *NewString = (linkedlist *) FElement(&(theBundle->variable_ptr->mem.data), StringIndex);
    Ulong NewStringSize = 0;
    LL_retrn linkedLretrn;
    letter_type CharType = CharType = lettertype(**(char **) SourcePtr);


        // the leading spaces are skipped even for a string-to-string copy
        
    while ((CharType == a_space) || (CharType == a_eol) || (CharType == unprintable))
        {  (*(char **) SourcePtr)++;  CharType = lettertype(**(char **) SourcePtr);  }
    while ((CharType == a_letter) || (CharType == a_number) || (CharType == a_symbol))
        {  NewStringSize++;  CharType = lettertype(*((*(char **) SourcePtr)+NewStringSize));  }

    if (NewString->memory != NULL)  FDeleteLinkedList(NewString);
    linkedLretrn = FNewLinkedList(NewString, NewStringSize, sizeof(char), 0, uncleared);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }

    FSetElements(NewString, 1, NewString->elementnum, *(void **) SourcePtr);
    *(char **) SourcePtr += NewStringSize;
    if (NewStringSize == 0)  SetWarning(type_mismatch_err, PCCodePtr-1);
    if ((**(char **) SourcePtr == 0) && (*(void **) SourcePtr < OverflowMarker))  *(char **) SourcePtr += 1;
}


// Next 3 routines:  invoked by call()
// These count the number of seperate arguments that will go into the 'argv' array.

void CountDataLists(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *BundleCount, void *StringCount)
{  SRWTemplate(theBundle, BundleOffset, IndexWidth, BundleCount, StringCount, yz_false, &CountDataLists, &Count_DataBundle, &Count_StringBundle);  }

void Count_DataBundle(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *BundleCount, void *Dummy)
{  (*(Ulong *) BundleCount)++;  }

void Count_StringBundle(bundle *theBundle, Ulong StringIndex, void *BundleCount, void *StringCount)
{  (*(signed long *) StringCount)++;  }


// Next 3 routines:  invoked by call()
// These fill the argv array of a call() function with the addresses of the data lists.

void Fill_argv_Handle(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *PtrHandle, void *InfoHandle)
{  SRWTemplate(theBundle, BundleOffset, IndexWidth, PtrHandle, InfoHandle, yz_false, &Fill_argv_Handle, &PointTo_DataBundle, &PointTo_String);  }

void PointTo_DataBundle(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *PtrHandle, void *InfoHandle)
{
    LL_retrn linkedLretrn;

    linkedLretrn = FDefragmentLinkedList(&(theBundle->variable_ptr->mem.data));
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }

    if (theBundle->variable_ptr->mem.data.elementnum == 0)  **(char ***) PtrHandle = NULL;
    else  **(char ***) PtrHandle = (char *) FElement(&(theBundle->variable_ptr->mem.data), BundleOffset+1);
    (*(arg_info **) InfoHandle)->arg_type = theBundle->variable_ptr->type;
    (*(arg_info **) InfoHandle)->arg_size = theBundle->variable_ptr->mem.data.elementsize;
    (*(arg_info **) InfoHandle)->arg_indices = IndexWidth;
    (*(char ***) PtrHandle)++;
    (*(arg_info **) InfoHandle)++;
}

void PointTo_String(bundle *theBundle, Ulong StringIndex, void *PtrHandle, void *InfoHandle)      // unused
{
    LL_retrn linkedLretrn;
    linkedlist *theStringLL = (linkedlist *) FElement(&(theBundle->variable_ptr->mem.data), StringIndex);

    linkedLretrn = FDefragmentLinkedList(theStringLL);
    if (linkedLretrn != passed)  {  SetError(out_of_mem_err, PCCodePtr-1);  return;  }

    **(char ***) PtrHandle = (char *) FElement(theStringLL, 1);
    (*(arg_info **) InfoHandle)->arg_type = 9;
    (*(arg_info **) InfoHandle)->arg_size = theStringLL->elementnum;
    (*(arg_info **) InfoHandle)->arg_indices = 1;
    (*(char ***) PtrHandle)++;
    (*(arg_info **) InfoHandle)++;
}


// SRWTemplate() is used for scanning a single bundle tree (in comparison to CCTemplate() which scans two at once).
// Invoked by size(), feq, read_string(), print_string(), call().
// This routine skips hidden members; so for example size() will not register these.

void SRWTemplate(bundle *theBundle, Ulong BundleOffset, Ulong IndexWidth, void *GlobalPtr, void *SecondaryPtr, Boolean IfIndexOrder,
        void(*SLS_Composite)(bundle *, Ulong, Ulong, void *, void *),
        void(*SLS_Data)(bundle *, Ulong, Ulong, void *, void *),
        void(*SLS_String)(bundle *, Ulong, void *, void *))
{
    bundle *NextBundle;
    member *loopMember;
    Ulong NextOffset, NextIndices, MemberCounter, IndexCounter;
    
    if (IfBusy(theBundle, busy_SRW_flag) != yz_false)  {  SetError(self_reference_err, PCCodePtr-1);  return;  }
    SetBusy(theBundle, busy_SRW_flag);

    if (theBundle->variable_ptr->type == composite_type)   {
        
        for (MemberCounter = 1; MemberCounter <= theBundle->variable_ptr->mem.members.elementnum; MemberCounter++)  {
            loopMember = (member *) FElement(&(theBundle->variable_ptr->mem.members), MemberCounter);
            if ((loopMember->member_ptr != NULL) && (loopMember->if_hidden == yz_false))      {
                if ((loopMember->member_ptr->variable_ptr->type == proxy_type) || (IfIndexOrder != yz_false))  {
                    for (IndexCounter = 0; IndexCounter < IndexWidth*loopMember->indices; IndexCounter++)  {
                        NextBundle = loopMember->member_ptr;
                        NextOffset = BundleOffset;
                        NextIndices = 1;            // a dummy variable
                
                        StepBundle(&NextBundle, &NextOffset, &NextIndices, loopMember, IndexCounter, 1);
                        if (err_code == void_proxy_err)  err_code = passed;
                        else {
                            if (err_code == passed)  SLS_Composite(NextBundle, NextOffset, 1, GlobalPtr, SecondaryPtr);
                            if (err_code != passed)  {  ClearBusy(theBundle, busy_SRW_flag);  return;  }
                }   }   }
                
                else  {
                    NextBundle = loopMember->member_ptr;
                    NextOffset = BundleOffset;
            
                    StepBundle(&NextBundle, &NextOffset, &NextIndices, loopMember, 0, 1);
                    if (err_code == passed)  SLS_Composite(NextBundle, NextOffset, IndexWidth*loopMember->indices, GlobalPtr, SecondaryPtr);
                    if (err_code != passed)  {  ClearBusy(theBundle, busy_SRW_flag);  return;  }
            }   }
    }   }
    
    else  {        // it's a primitive variable
        if ((theBundle->variable_ptr->type == string_type) && (ExtCallMode == no_call))  {
            for (IndexCounter = 1; IndexCounter <= IndexWidth; IndexCounter++)  {
                SLS_String(theBundle, IndexCounter+BundleOffset, GlobalPtr, SecondaryPtr);
                if (err_code != passed)  break;       }   }
        else  SLS_Data(theBundle, BundleOffset, IndexWidth, GlobalPtr, SecondaryPtr);            }

    ClearBusy(theBundle, busy_SRW_flag);
}


// PrintNumber() prints a number out in characters, and returns a count of the number of characters.
// This is used by print_string(), in the Size_DataString() and Read_DataString() routines above.
// If DestString == NULL, we are just getting the size of the string, not writing anything.
// Otherwise, we write DestString in the confidence that print_string() has given it the required storage.

void PrintNumber(char *DestString, double NumberToPrint, Ulong *CharacterCount, unsigned char NumberType)
{
    double WorkDouble = NumberToPrint, ten_power, MaxPrecision;
    signed long digit, exponent = 0;
    const signed long HighExponent = 6, LowExponent = -2;
    Ulong exp_characters, MaxDigits = 0;
    signed int OneDigit;
    Ulong underflow_factor = 0;
    

        // take care of +- infinity

    if ((NumberToPrint > DBL_MAX) || (NumberToPrint < -DBL_MAX))  {
        if (NumberToPrint < 0)  {
            *CharacterCount = 4;
            if (DestString != NULL)  *DestString = '-';   }
        else  *CharacterCount = 3;
        
        if (DestString != NULL)  {
            *(DestString+(*CharacterCount)-3) = 'i';
            *(DestString+(*CharacterCount)-2) = 'n';
            *(DestString+(*CharacterCount)-1) = 'f';      }

        return;         }
    
    
        // deal with the not-a-number possibility
    
    if (!((NumberToPrint >= 0) || (NumberToPrint <= 0)))  {
        *CharacterCount = 3;

        if (DestString != NULL)  {
            *DestString = 'n';
            *(DestString+1) = 'a';
            *(DestString+2) = 'n';         }
        
        return;         }
    
    
    if (fabs(WorkDouble) < DBL_MIN)  {
        underflow_factor = 20;
        WorkDouble *= 1.e20;        }
    
    
        // handle a conventional number, with a possible exponent
    
    if (WorkDouble == 0)  {  *CharacterCount = 1;  if (DestString != NULL)  *DestString = '0';  return;  }
    if (NumberType == single_type)  MaxDigits = FLT_DIG;
    else if (NumberType == double_type)  MaxDigits = DBL_DIG;
    if ((GL_MaxDigits != 0) && (GL_MaxDigits < MaxDigits))  MaxDigits = GL_MaxDigits;


        // write out a leading '-'

    *CharacterCount = 0;
    if (WorkDouble < 0)  {  *CharacterCount = 1;  if (DestString != NULL)  *DestString = '-';  WorkDouble *= -1;  }
    
    
        // compute the exponent and number of decimal places
    
    ten_power = 1;
    if (WorkDouble >= 10)  while (WorkDouble/ten_power >= 10)  {  ten_power *= 10;  exponent++;  }
    else if (WorkDouble < 1)  do  {  ten_power /= 10.;  exponent--;  }  while (ten_power/WorkDouble > 1);   // WD > 0 -- checked above

    if (NumberType < single_type)  MaxPrecision = .5;
    else  MaxPrecision = ten_power*5*pow(0.1, MaxDigits);       // the latter to prevent truncation errors (1 -> 0.999...)
    WorkDouble += MaxPrecision*0.5;

    if (ten_power*10 <= WorkDouble)  {  ten_power *= 10;  exponent++;  }   // from round-off errors; we didn't add MaxP when computing t_p, exp
    digit = exponent;
    
    if (((exponent <= HighExponent) && (exponent >= LowExponent)) || (NumberType < single_type))  {
        if (exponent < 1)  {  digit = 0;  ten_power = 1;  }
        exponent = 0;         }


        // write out the digits
    
    while (((WorkDouble > MaxPrecision) || ((exponent == 0) && (ten_power > 0.9))) && (ten_power > MaxPrecision))  {
        if (digit-exponent == -1)  {
            (*CharacterCount)++;
            if (DestString != NULL)  *(DestString+*CharacterCount-1) = '.';         }

        (*CharacterCount)++;
        
        OneDigit = (char) floor(WorkDouble/ten_power);
        if (OneDigit < 0)  OneDigit = 0;       // to deal w/ numerical issues at the edge of float range
        else if (OneDigit > 9)  {  OneDigit = 9;  WorkDouble = ten_power;  }
        if (DestString != NULL)  *(char *) (DestString + *CharacterCount - 1) = OneDigit+'0';
        WorkDouble -= OneDigit*ten_power;

        digit--;  ten_power /= 10;         }
        
        
        // write out the exponent
        
    if (exponent != 0)  {
        (*CharacterCount)++;
        if (DestString != NULL)  *(DestString+*CharacterCount-1) = 'e';
        exponent -= underflow_factor;
        if (DestString == NULL)  PrintNumber(NULL, exponent, &exp_characters, Slong_type);
        else  PrintNumber(DestString+*CharacterCount, exponent, &exp_characters, Slong_type);
        (*CharacterCount) += exp_characters;     }

    if (*CharacterCount == 0)  {  *CharacterCount = 1;  if (DestString != NULL)  *DestString = '0';  return;  }
}



// * * * * *  Routines for walking up and down the member/fibril trees  * * * * *


// The next two routines search up the fibril path for a variable member with a name of SoughtVarID.
// SearchMember_DragFibril() carries a (new, if necessary) fibril along with it; SearchMember() doesn't bother.
// Returns the (fibril &) bundle (in SearchCoordinate._Fibril/Bundle) & member number.
// Note that the bundle (& fibril) returned are those of the stem variable; user can step from there using DefStepToIndex or just STI.
// Assumes we are starting from a composite variable.

void SearchMember_DragFibril(signed long SoughtVarID, member **StemMember, Ulong *StemMemberNumber)
{
    fibril *OldSearchFibril;
    run_retrn retrn;

    RefFibril(PCCoordinate._Fibril);
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate._Fibril = PCCoordinate._Fibril;
    SearchCoordinate._Offset = PCCoordinate._Offset;
    SearchCoordinate._IndexWidth = PCCoordinate._IndexWidth;
    
    do  {
        SearchCoordinate._Bundle = SearchCoordinate._Fibril->bundle_ptr;
        if (SearchCoordinate._Bundle->variable_ptr->type != proxy_type)  {
            retrn = FindMemberID(SearchCoordinate._Bundle, SoughtVarID, StemMember, StemMemberNumber);
            if (retrn == passed)  {
                return;    }
            else if (retrn != member_not_found_err)  {
                SetError(retrn, PCCodePtr);
                return;    }   }
                
        if (SearchCoordinate._Fibril->stem == NULL)  {
            break;         }
        
        if (SearchCoordinate._Fibril->if_absolute_index == yz_false)  {      // step back into a non-proxy variable
            if (SearchCoordinate._Fibril->stem_indices == 0)  SearchCoordinate._Offset = 0;
            else   {
                if (SearchCoordinate._Fibril->stem_indices == 0)  SearchCoordinate._Offset = SearchCoordinate._Fibril->bundle_ptr->bundle_offset;
                else  SearchCoordinate._Offset = (Ulong) floor(
                    (SearchCoordinate._Offset - SearchCoordinate._Fibril->bundle_ptr->bundle_offset) / SearchCoordinate._Fibril->stem_indices);  }
            if (SearchCoordinate._Fibril->stem_indices > 1)
                SearchCoordinate._IndexWidth /= SearchCoordinate._Fibril->stem_indices;         }
        else                       // step back past a proxy
            SearchCoordinate._Offset = SearchCoordinate._Fibril->stem_offset;
            
        OldSearchFibril = SearchCoordinate._Fibril;
        SearchCoordinate._Fibril = SearchCoordinate._Fibril->stem;
        RefFibril(SearchCoordinate._Fibril);
        DerefFibril(&OldSearchFibril);       // should not completely disappear, since there's a code beneath it.
    }  while (yz_true);

    RefFibril(PCCoordinate._Fibril);
    DerefFibril(&(SearchCoordinate._Fibril));
    SearchCoordinate = PCCoordinate;
    SetError(member_not_found_err, PCCodePtr);
    return;
}


// Same as last routine, but doesn't ensure a fibril path (which is not needed in, e.g., the right-hand side of equate)

void SearchMember(signed long SoughtVarID, member **StemMember, Ulong *StemMemberNumber)
{
    fibril *SearchFibril;      // use this in place of SearchCoordinate._Fibril so we don't get confused with referencing
    run_retrn retrn;

    DerefFibril(&(SearchCoordinate._Fibril));
    SearchFibril = PCCoordinate._Fibril;    // don't bother referencing it, since we won't use it except to climb the bundle tree
    SearchCoordinate._Offset = PCCoordinate._Offset;
    SearchCoordinate._IndexWidth = PCCoordinate._IndexWidth;
    *StemMember = NULL;
    
    do  {
        SearchCoordinate._Bundle = SearchFibril->bundle_ptr;
        if (SearchCoordinate._Bundle->variable_ptr->type != proxy_type)  {
            retrn = FindMemberID(SearchCoordinate._Bundle, SoughtVarID, StemMember, StemMemberNumber);
            if (retrn == passed)   {
                return;    }
            else if (retrn != member_not_found_err)  {
                SetError(retrn, PCCodePtr);
                return;    }   }
                
        if (SearchFibril->stem == NULL)  break;
        if (SearchFibril->if_absolute_index == yz_false)  {     // step back into a non-proxy variable
            if (SearchFibril->stem_indices == 0)  {
                SearchCoordinate._Offset = 0;
                SearchCoordinate._IndexWidth = SearchFibril->stem->bundle_ptr->instances;   }    // <--  because theoretically it
            else   {                                                                             // only happens (?) when we define a size-0 array
                if (SearchFibril->stem_indices == 0)  SearchCoordinate._Offset = SearchFibril->bundle_ptr->bundle_offset;
                else  SearchCoordinate._Offset = (Ulong) floor(
                        (SearchCoordinate._Offset - SearchFibril->bundle_ptr->bundle_offset) / SearchFibril->stem_indices);   }
            if (SearchFibril->stem_indices > 1)  {
                if (SearchFibril->stem_indices > SearchCoordinate._IndexWidth)  SearchCoordinate._IndexWidth = 1;    // cannot step to partial
                else  SearchCoordinate._IndexWidth /= SearchFibril->stem_indices;                 }                  // member if SC.IW > 1
            }
        else  {          // step back past a proxy
            SearchCoordinate._Offset = SearchFibril->stem_offset;
            SearchCoordinate._IndexWidth = SearchFibril->stem_indices;       }
            
        SearchFibril = SearchFibril->stem;
    }  while (yz_true);

    SearchCoordinate = PCCoordinate;
    SearchCoordinate._Fibril = NULL;
    SetError(member_not_found_err, PCCodePtr);
    return;
}


// Next two routines:  walk SearchBundle from one bundle (variable) to one of its members
// The two functions are the same, except that the former updates PCCoordinate._Fibril as well
// (by creating a new fibril in the destination bundle if necessary).
// StepSearchBundleAndFibril() is used when defining variables.

void StepSearchBundleAndFibril(member *StartingMember, Ulong MemberNumber, Ulong EntryOffset, Ulong IndexWidth)
{
    Ulong StemOffset;
    fibril *OldSearchFibril = SearchCoordinate._Fibril;
    Boolean found_fibril = yz_false;
    run_retrn retrn;
    
    StepToMember(&(SearchCoordinate._Bundle), &(SearchCoordinate._Offset), StartingMember, EntryOffset);
    if (err_code != passed)  return;
    SearchCoordinate._IndexWidth *= IndexWidth;
    
    retrn = DrawFibril(SearchCoordinate._Bundle, OldSearchFibril, &(SearchCoordinate._Fibril), 0, StartingMember->indices, yz_false,
                    OldSearchFibril->source_code, MemberNumber, NULL);
    if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
    
    RefFibril(SearchCoordinate._Fibril);
    DerefFibril(&OldSearchFibril);          // same as dereferencing SC_F before stepping, but less dangerous
    
    while (SearchCoordinate._Bundle->variable_ptr->type == proxy_type)  {
        bundle **ProxyMember;
        
        StemOffset = SearchCoordinate._Offset;
        OldSearchFibril = SearchCoordinate._Fibril;
        found_fibril = yz_false;
        
        if (SearchCoordinate._IndexWidth != 1)  {  SetError(step_multiple_proxies_err, PCCodePtr-1);  return;  }
        
        ProxyMember = (bundle **) FElement(&(SearchCoordinate._Bundle->variable_ptr->mem.data), SearchCoordinate._Offset+1);
        if (*ProxyMember == NULL)   {
            if ((GL_Define.DefineOn == yz_true) && (GL_Define.FinalStep == yz_false))  {
                AddVB(ProxyMember, GL_Path.StemMember->indices, composite_type, yz_true);
                GL_Path.StemMember->member_ptr = GL_Path.StemMember->member_ptr;      }
            else  {  SearchCoordinate._Bundle = NULL;  SetError(void_proxy_err, PCCodePtr-1);  return;  }     }
        
        SearchCoordinate._Bundle = *ProxyMember;
        SearchCoordinate._Offset = SearchCoordinate._Bundle->bundle_offset;
        
        retrn = DrawFibril(SearchCoordinate._Bundle, OldSearchFibril, &(SearchCoordinate._Fibril), SearchCoordinate._Offset, 1, yz_true,
                            OldSearchFibril->source_code, 0, NULL);
        if (retrn != passed)  {  SetError(retrn, PCCodePtr-1);  return;  }
        
        RefFibril(SearchCoordinate._Fibril);
        DerefFibril(&OldSearchFibril);
    }
    
    return;
}


// Same as StepSearchBundleAndFibril(), except that this one doesn't pull a new fibril path along.

void StepSearchBundle(member *StartingMember, Ulong EntryOffset, Ulong IndexWidth)
{
    StepBundle(&(SearchCoordinate._Bundle), &(SearchCoordinate._Offset), &(SearchCoordinate._IndexWidth), StartingMember, EntryOffset, IndexWidth);
}


// This is a general routine for stepping a bundle/variable pointer to one of its members, passing over proxy variables.

void StepBundle(bundle **BundleToStep, Ulong *OffsetToStep, Ulong *IndicesToStep, member *StartingMember, Ulong EntryOffset, Ulong IndexWidth)
{
    bundle *NextBundle;
    
    StepToMember(BundleToStep, OffsetToStep, StartingMember, EntryOffset);
    if (err_code != passed)  return;
    *IndicesToStep *= IndexWidth;
    
    while ((*BundleToStep)->variable_ptr->type == proxy_type)  {
        NextBundle = *(bundle **) FElement(&((*BundleToStep)->variable_ptr->mem.data), (*OffsetToStep)+1);
        if (NextBundle == NULL)  {  SetError(void_proxy_err, PCCodePtr-1);  return;  }
        if (NextBundle->jam_status == unjammed)  {  SetError(unjammed_proxy_err, PCCodePtr-1);  return;  }
        *BundleToStep = NextBundle;
        
        if (*IndicesToStep != 1)  {  SetError(step_multiple_proxies_err, PCCodePtr-1);  return;  }
        *OffsetToStep = (*BundleToStep)->bundle_offset;      }
    
    return;
}


// Steps a bundle/variable pointer to one of its member bundles/variables, even if that member is a proxy variable.

void StepToMember(bundle **BundleToStep, Ulong *OffsetToStep, member *StartingMember, Ulong EntryOffset)
{
    (*BundleToStep) = StartingMember->member_ptr;
    if (StartingMember->member_ptr == NULL)  {  SetError(void_member_err, PCCodePtr-1);  return;  };
    if (StartingMember->member_ptr->jam_status == unjammed)  {  *BundleToStep = NULL;  SetError(unjammed_member_err, PCCodePtr-1);  return;  }
    *OffsetToStep = (*OffsetToStep)*StartingMember->indices + (*BundleToStep)->bundle_offset + EntryOffset;
}


// FindMemberID() returns the member number (not index!) of the member with the sought ID (name).
// Returns 0 if no match found in the current bundle, or if we're for some reason looking for an ID of zero
// (which is not allowed since the entries in the names LL begin at one) (also since ID 0 is reserved for implicit variable definitions).

run_retrn FindMemberID(bundle *theBundle, signed long SoughtMemberID, member **SoughtMember, Ulong *SoughtMemberNumber)
{
    Ulong counter;
    linkedlist *MembersList;

    if (theBundle->variable_ptr->type != composite_type)  return member_not_found_err;
    MembersList = &(theBundle->variable_ptr->mem.members);

    for (counter = 1; counter <= MembersList->elementnum; counter++)   {
        *SoughtMember = (member *) FElement(MembersList, counter);
        if ((*SoughtMember)->member_ID == SoughtMemberID)   {
            *SoughtMemberNumber = counter;
            return passed;
    }   }

    *SoughtMember = NULL;
    return member_not_found_err;
}


// FindMemberIndex() returns the member number corresponding to the given index [n] requested by the user
// Remember that some members (e.g. arrays) have multiple indices, whereas hidden members are skipped entirely from the index count.
// Returns, in addition, the member * and the 'entry offset', an offset from the first index of the member which can be 0.

run_retrn FindMemberIndex(variable *theVariable, Ulong SoughtMemberIndex, member **SoughtMember, Ulong *LL_EntryNum, Ulong *EntryOffset)
{
    Ulong IndexCounter;
    Ulong FirstArray = theVariable->first_array;
    Ulong NumEntries;


        // process this quickly if the requested index is before the first non-1-index-width member
        // the latter condition (SMI > 0) ensures that we return a member only if one already exists,
        // so {a[0]::..} doesn't make two members when constructor is run twice
        
    if ((SoughtMemberIndex < FirstArray) && (SoughtMemberIndex > 0))  {
        *SoughtMember = (member *) FElement(&(theVariable->mem.members), SoughtMemberIndex);
        *LL_EntryNum = SoughtMemberIndex;
        *EntryOffset = 0;
        return passed;         }


        // OK, there are array-style members, so we have to count indices
        
    NumEntries = theVariable->mem.members.elementnum;
    IndexCounter = FirstArray;

    for (*LL_EntryNum = FirstArray; *LL_EntryNum <= NumEntries; (*LL_EntryNum)++)  {
        *SoughtMember = (member *) FElement(&(theVariable->mem.members), *LL_EntryNum);
        if ((*SoughtMember)->if_hidden == yz_false)  {
            if (IndexCounter+(*SoughtMember)->indices > SoughtMemberIndex)  {
                *EntryOffset = SoughtMemberIndex - IndexCounter;
                return passed;         }
            IndexCounter += (*SoughtMember)->indices;      }
    }

    *EntryOffset = IndexCounter;     // kind of a kludge -- a way of passing back the first undefined index
    return invalid_index_err;
}



// * * * * *  Error and warning routines  * * * * *


// SetError() sets the error code and error pointer.

void SetError(Ulong ErrorCode, Slong *ErrorIndex)
{
    err_code = ErrorCode;
    
    if (err_script.code_ptr != NULL)  Deref_CodeRef(&err_script);
    err_script = PCCodeRef;
    err_script.anchor = NULL;
    Ref_CodeRef(&err_script);

    err_index = ErrorIndex - err_script.code_ptr + 1;
}


// SetWarning() sets the warning code and warning pointer.  Warnings don't stop execution.

void SetWarning(Ulong WarningCode, Slong *WarningPtr)
{
    warning_code = WarningCode;
    
    if (warning_script.code_ptr != NULL)  Deref_CodeRef(&warning_script);
    warning_script = PCCodeRef;
    warning_script.anchor = NULL;
    Ref_CodeRef(&warning_script);

    warning_index = WarningPtr - warning_script.code_ptr + 1;
}



// * * * * *  Memory routines  * * * * *


// InitYazoo() initializes the interpreter when Yazoo boots up.
// Allocates memory for things like VariableList, the PCStack, Zero, etc.

run_retrn InitYazoo(Slong *ZeroCode, Ulong ZeroCodeLength)
{
    bundle BlankBundle;
    variable *ZeroBase;
    Ulong ZeroCodeIndex;
    Slong *ZeroCodeEntryPtr;
    run_retrn retrn;
    
    SearchCoordinate._Bundle = NULL;     // these can confuse AddMemory if not initialized to NULL first
    PCCoordinate._Bundle = NULL;
    SearchCoordinate._Fibril = GL_Path.StemCoordinate._Fibril = NULL;     // these would have to be referenced otherwise
    
    PCCoordinate._Bundle = &BlankBundle;            // to prevent an error on startup when AddMemory checks PCCoordinate._Bundle->variable_ptr
    PCCoordinate._Bundle->variable_ptr = NULL;

    retrn = NewPLL(&VariableList, 0, sizeof(variable), LLFreeSpace);        // make space for variables & codes
    if (retrn != passed)  return retrn;
    retrn = NewPLL(&MasterCodeList, 0, sizeof(Slong **), LLFreeSpace);
    if (retrn != passed)  return retrn;
    retrn = NewStack(&PCStack, sizeof(bundle_coordinate)+sizeof(Ulong), 10, LLFreeSpace);           // 10?
    if (retrn != passed)  return retrn;

    retrn = AddCompositeVariable(&ZeroBase);
    if (retrn != passed)  return retrn;
    retrn = AddBundle(ZeroBase, 1, 0, &Zero, yz_true);
    if (retrn != passed)  return retrn;
//    retrn = AddFibril(Zero, NULL, &ZeroSuspensor, 0, 1, yz_true, 1, 0, NULL);
    retrn = DrawFibril(Zero, NULL, &ZeroSuspensor, 0, 1, yz_true, 1, 0, NULL);
    if (retrn != passed)  return retrn;

    retrn = AddCode(ZeroCode, &ZeroCodeEntryPtr, &ZeroCodeIndex, ZeroCodeLength);
    if (retrn != passed)  return retrn;
    retrn = AddCodeRef(&(ZeroBase->code_list), ZeroSuspensor, ZeroCodeEntryPtr, ZeroCodeIndex);
    if (retrn != passed)  return retrn;
    PCCodeRef = *(code_ref *) FElement(&(ZeroBase->code_list), 1);

    RefBundle(Zero);
    RefFibril(ZeroSuspensor);

    PCCodePtr = ZeroCodeEntryPtr;        // should immediately call CheckHobbish()

    retrn = FNewLinkedList(&OneCodeLL, 1, sizeof(code_ref), 0, uncleared);      // used in hobbsh.c to pass codes from code blocks
    if (retrn != passed)  return retrn;
    OneCodeRef = (code_ref *) FElement(&OneCodeLL, 1);

    err_code = passed;
    err_index = 1;              // for now
    warning_code = passed;
    err_script.code_ptr = NULL;
    warning_script.code_ptr = NULL;
    
    GL_RecursionCounter = 0;
    ArgsCoordinate._Bundle = NULL;
    ReturnCoordinate._Bundle = NULL;
    
    return passed;
}


// Deletes some global variables when Yazoo exits.  Maybe unnecessary.

void CleanUp()
{
    DeletePLL(&VariableList);
    DeletePLL(&MasterCodeList);
    DeleteStack(&PCStack);
    DeleteLinkedList(&OneCodeLL);

    return;
}


// Adds a composite variable:  so, sets up 'mem' as the members list and the fibril_list, and sets the type to composite_type.

run_retrn AddCompositeVariable(variable **TheNewVar)
{
    LL_retrn linkedLretrn;
    run_retrn retrn;

    retrn = AddVariable(TheNewVar);
    if (retrn != passed)  return retrn;
    (*TheNewVar)->type = composite_type;

    linkedLretrn = FNewLinkedList(&((*TheNewVar)->mem.members), 0, sizeof(member), LLFreeSpace, uncleared);
    if (linkedLretrn != passed)  return out_of_mem_err;

    return passed;
}


// Creates a primitive variable:  construes 'mem' as data and allocates appropriate memory.

run_retrn AddPrimitiveVariable(variable **TheNewVar, Ulong VariableType, Ulong DataSize)
{
    LL_retrn linkedLretrn;
    run_retrn retrn;

    retrn = AddVariable(TheNewVar);
    if (retrn != passed)  return retrn;
    (*TheNewVar)->type = VariableType;
                                                  // we have 0 elements until the variable gets linked to a member
    linkedLretrn = FNewLinkedList(&((*TheNewVar)->mem.data), 0, DataSize, LLFreeSpace, uncleared);
    if (linkedLretrn != passed)  return out_of_mem_err;

    return passed;
}


// AddVariable() adds a new variable to the variable tree, and sets a number of the fields in the variable record.
// This routine is called by both AddPrimitiveVariable() and AddCompositeVariable().
// It not allocate memory for members or data -- that is done by each of the two routines manually.
// The type is also not set; business is set to zero.  Assumes zero initial instances of the variable.
// The new variable must be referenced by a bundle immediately or it will likely be overwritten.

run_retrn AddVariable(variable **TheNewVar)
{
    LL_retrn linkedLretrn;
    run_retrn retrn;
    Ulong NewVarIndex;

    retrn = AddPLLElement(&VariableList, (void **) TheNewVar, &NewVarIndex);
    if (retrn != passed)  return retrn;

    (*TheNewVar)->PLL_index = NewVarIndex;
    (*TheNewVar)->references = LL_ULong(&(VariableList.references), NewVarIndex);
    (*(*TheNewVar)->references) = 0;         // the bundles are all the references we want

    linkedLretrn = FNewLinkedList(&((*TheNewVar)->code_list), 0, sizeof(code_ref), LLFreeSpace, uncleared);
    if (linkedLretrn != passed)  return out_of_mem_err;

    retrn = NewPLL(&((*TheNewVar)->bundles), 0, sizeof(bundle), LLFreeSpace);
    if (retrn != passed)  return retrn;
    retrn = NewPLL(&((*TheNewVar)->fibril_list), 3, sizeof(fibril), LLFreeSpace);        // why 3?
    if (retrn != passed)  return retrn;

    (*TheNewVar)->instances = 0;
    (*TheNewVar)->first_array = 1;
    (*TheNewVar)->business = 0;

    return passed;
}


// References a variable (signifying that it is being used)

void RefVariable(variable *theVariable)
{
    RefPLLPtr(theVariable->references);

    return;
}


// Dereferences a variable, and deletes it if refs == 0.

void DerefVariable(variable *theVariable)
{
    Ulong counter, CodeCounter;
    member *loopMember;
    linkedlist *loopList;

    DerefPLLPtr(&VariableList, theVariable->PLL_index, theVariable->references);
    if (*(theVariable->references) != 0)  return;

    if ((theVariable->type < composite_type) || (theVariable->type == proxy_type))      {
        if (theVariable->type == string_type)       {
        for (counter = 1; counter <= theVariable->mem.data.elementnum; counter++)  {
            loopList = (linkedlist *) FElement(&(theVariable->mem.data), counter);
            if (loopList->memory != NULL)  FDeleteLinkedList(loopList);
        }}
        FDeleteLinkedList(&(theVariable->mem.data));            }
    else   {
        for (counter = 1; counter <= theVariable->mem.members.elementnum; counter++)   {
            loopMember = (member *) FElement(&(theVariable->mem.members), counter);
            for (CodeCounter = 1; CodeCounter <= loopMember->code_list.elementnum; CodeCounter++)   {
                Deref_CodeRef((code_ref *) FElement(&(loopMember->code_list), CodeCounter));   }
            FDeleteLinkedList(&(loopMember->code_list));
            DerefBundle(&(loopMember->member_ptr));       }
        FDeleteLinkedList(&(theVariable->mem.members));   }      // they've already been unlinked

    DeletePLL(&(theVariable->bundles));         // All the bundles must already have been deleted, since variables are referenced by the bundles.
    DeletePLL(&(theVariable->fibril_list));     // All fibrils are already gone, since all the bundles are gone.

    for (counter = 1; counter <= theVariable->code_list.elementnum; counter++)
        Deref_CodeRef((code_ref *) FElement(&(theVariable->code_list), counter));

    FDeleteLinkedList(&(theVariable->code_list));

    return;
}


// AddBundle() creates new instances of the variable, and allocates the associated storage space.
// Does not reference the bundle (and the host variable does not get referenced until the bundle is first referenced).
// If BundleOffset is greater than any possible offset (given by HostVariable->instances), this routine creates new space in the variable;
// otherwise, it is construed as an alias at BundleOffset.

run_retrn AddBundle(variable *HostVariable, Ulong AddedInstances, Ulong BundleOffset, bundle **NewBundle, Boolean IfCanJam)
{
    Ulong NewIndex;
    LL_retrn linkedLretrn;
    run_retrn retrn;

    if (AddedInstances > LONG_MAX)  return index_argument_err;   // so there are no errors calling AddMemory(slong AddedInstances)
    
    linkedLretrn = AddPLLElement(&(HostVariable->bundles), (void **) NewBundle, &NewIndex);
    if (linkedLretrn != passed)  return out_of_mem_err;

    (*NewBundle)->PLL_index = NewIndex;
    (*NewBundle)->references = (Ulong *) FElement(&(HostVariable->bundles.references), NewIndex);
    (*NewBundle)->variable_ptr = HostVariable;
    (*NewBundle)->bundle_offset = BundleOffset;
    (*NewBundle)->business = yz_false;
    (*NewBundle)->jam_status = IfCanJam;
    (*NewBundle)->remote_fibril = NULL;         // don't reference (to avoid loops)
    
    *((*NewBundle)->references) = 0;

    if (BundleOffset == HostVariable->instances)   {    // Do this first, since otherwise AddMemory will adjust our new (correct) offset
        (*NewBundle)->instances = 0;                     // until AddMemory()
        retrn = AddMemory(*NewBundle, BundleOffset, AddedInstances);          // no need to check overlap
        UnflagVariables(HostVariable, busy_add_flag);
        return retrn;       }
    else  (*NewBundle)->instances = AddedInstances;

    return passed;
}


// References a bundle, and its host variable as well if this is the first reference.

void RefBundle(bundle *theBundle)
{
    if (theBundle == NULL)  return;
    
    RefPLLPtr(theBundle->references);
    
    if (*(theBundle->references) == 1)   {
        RefVariable(theBundle->variable_ptr);
    }
        
    return;
}


// Dereferences a bundle, and deletes it if refs == 0.

void DerefBundle(bundle **theBundle)
{
    Ulong counter;  //, BottomDelete, TopDelete;

    if (*theBundle == NULL)  return;
    
    DerefPLLPtr(&((*theBundle)->variable_ptr->bundles), (*theBundle)->PLL_index, (*theBundle)->references);
    if (*((*theBundle)->references) != 0)  {  *theBundle = NULL;  return;  }

    if ((*theBundle)->variable_ptr->type == proxy_type)   {   // can there be overlapping bundles?
        for (counter = 1; counter <= (*theBundle)->instances; counter++)   {
            DerefBundle((bundle **) FElement(&((*theBundle)->variable_ptr->mem.data), (*theBundle)->bundle_offset+counter));
    }   }
    
    //DerefFibril(&((*theBundle)->remote_fibril));
    DerefVariable((*theBundle)->variable_ptr);
    *theBundle = NULL;
    
    return;
}


// CombBundles() removes any bundles/variables/etc. that cannot be accessed.
// Inaccessible memory can occur if A references B and B references A, but neither are connected
// to the memory tree where PCCoordinate, or any PCCoordinates on the PCStack, reside.
// This is the rigorous, manual garbage-collection routine invoked by SpringCleaning().

void CombBundles()
{
    Ulong VariableCounter, MemberCounter, ProxyCounter, bundle_coordinateCounter, FibrilCounter, CodeCounter, CodeTop;
    variable *loopVariable;
    bundle **loopProxy;
    member *loopMember;
    Ulong MembersTop;


        // First flag all accessible regions of memory
    
    CombBranch(PCCoordinate._Bundle);
    for (bundle_coordinateCounter = 1; bundle_coordinateCounter <= PCStack.top; bundle_coordinateCounter++)
        CombBranch(((bundle_coordinate *) FElement(&(PCStack.data), bundle_coordinateCounter))->_Bundle);
        
        
        // Now scan for variables that have their flags cleared
    
    for (VariableCounter = 1; VariableCounter <= VariableList.data.elementnum; VariableCounter++)   {
    if (*LL_ULong(&(VariableList.references), VariableCounter) != 0)   {
        loopVariable = (variable *) FElement(&(VariableList.data), VariableCounter);

        if (IfBusy(loopVariable, busy_comb_flag) == yz_false)  {
        
            RefVariable(loopVariable);            // make sure the variable won't be deleted while we're still dismembering it

            CodeTop = loopVariable->code_list.elementnum;     // remove its code, since the anchor references the variable indirectly
            for (CodeCounter = 1; CodeCounter <= CodeTop; CodeCounter++)   {
                Deref_CodeRef((code_ref *) FElement(&(loopVariable->code_list), CodeCounter));   }
            FDeleteElements(&(loopVariable->code_list), 1, CodeTop);
            
            if (loopVariable->type == composite_type)  {      // composite?  then remove its members' code lists and bundle *s
                MembersTop = loopVariable->mem.members.elementnum;      // have it stored separately for FDeleteElements()
                for (MemberCounter = 1; MemberCounter <= MembersTop; MemberCounter++)  {
                    loopMember = (member *) FElement(&(loopVariable->mem.members), MemberCounter);
                    for (CodeCounter = 1; CodeCounter <= loopMember->code_list.elementnum; CodeCounter++)  {
                        Deref_CodeRef((code_ref *) FElement(&(loopMember->code_list), CodeCounter));   }
                    if (loopMember->member_ptr != NULL)  UnlinkBundle(loopVariable, &(loopMember->member_ptr), 0, MemberCounter);
					FDeleteLinkedList(&(loopMember->code_list));
                }
                FDeleteElements(&(loopVariable->mem.members), 1, MembersTop);      // ?
            }
            else if (loopVariable->type == proxy_type)  {
                MembersTop = loopVariable->mem.data.elementnum;
                for (ProxyCounter = 1; ProxyCounter <= MembersTop; ProxyCounter++)  {
                    loopProxy = (bundle **) FElement(&(loopVariable->mem.data), ProxyCounter);
//                    if (*loopProxy != NULL)  UnlinkBundle(loopVariable, loopProxy, ProxyCounter-1, 0);
            }   }

            for (FibrilCounter = 1; FibrilCounter <= loopVariable->fibril_list.data.elementnum; FibrilCounter++)  {
            if (*LL_ULong(&(loopVariable->fibril_list.references), FibrilCounter) != 0)  {
                DerefFibril(&(((fibril *) FElement(&(loopVariable->fibril_list.data), FibrilCounter))->stem));
            }}
            DerefVariable(loopVariable);      // the mortal blow
        }
    }}
    
    
        // remove the accessibility flags
    
    UnflagBundlesAndStems(PCCoordinate._Bundle, busy_comb_flag);
    for (bundle_coordinateCounter = 1; bundle_coordinateCounter <= PCStack.top; bundle_coordinateCounter++)  // and also the others
        UnflagBundlesAndStems(((bundle_coordinate *) FElement(&(PCStack.data), bundle_coordinateCounter))->_Bundle, busy_comb_flag);

    return;
}


// CombBranch() recursively scans the bundle tree to find all accessible bundles & variables, and set their business comb flags.
// This is called by CombVariables().

void CombBranch(bundle *theBundle)
{
    Ulong counter;
    member *loopMember;
    fibril *loopFibril;

    if (theBundle == NULL)  return;
    if (IfBusy(theBundle, busy_comb_flag) != yz_false)  return;
    SetBusy(theBundle, busy_comb_flag);
    SetBusy(theBundle->variable_ptr, busy_comb_flag);
    
    if (theBundle->variable_ptr->type == composite_type)  {
        for (counter = 1; counter <= theBundle->variable_ptr->mem.members.elementnum; counter++)  {
            loopMember = (member *) FElement(&(theBundle->variable_ptr->mem.members), counter);
            CombBranch(loopMember->member_ptr);
    }   }
    else if (theBundle->variable_ptr->type == proxy_type)  {
        for (counter = theBundle->bundle_offset+1; counter <= theBundle->bundle_offset+theBundle->instances; counter++)  {
            CombBranch(*(bundle **) FElement(&(theBundle->variable_ptr->mem.data), counter));
    }   }

    for (counter = 1; counter <= theBundle->variable_ptr->fibril_list.data.elementnum; counter++)  {
    if (*LL_ULong(&(theBundle->variable_ptr->fibril_list.references), counter) != 0)  {
        loopFibril = (fibril *) FElement(&(theBundle->variable_ptr->fibril_list.data), counter);
        if (loopFibril->stem != NULL)  CombBranch(loopFibril->stem->bundle_ptr);      // no need to do loopFibril->bundle_ptr
    }}

    for (counter = 1; counter <= theBundle->variable_ptr->code_list.elementnum; counter++)  {
        loopFibril = ((code_ref *) FElement(&(theBundle->variable_ptr->code_list), counter))->anchor;
        if (loopFibril != NULL)  CombBranch(loopFibril->bundle_ptr);         // anchors are NULL in err/w_script, thus usually in R_*_script
    }

    return;
}


// UnlinkBundle() cuts all 'matching' fibrils in UnlinkedBundle whose stems lie in StemVariable, then removes the UnlinkedBundle pointer.
// Matching fibrils either have StemOffset referring to the member we're unlinking, or for a proxy member Route = the proxy being unlinked.
// Derefs both the fibrils and UnlinkedBundle.

void UnlinkBundle(variable *StemVariable, bundle **UnlinkedBundle, Ulong StemOffset, Ulong StemMemberNumber)
{
    pinned_LL *UnlinkedFibrilList = &((*UnlinkedBundle)->variable_ptr->fibril_list);
    bundle *HoldBundlePtr;
    fibril *loopFibril, *ClippedStem;
    Ulong counter;
    
    for (counter = 1; counter <= UnlinkedFibrilList->data.elementnum; counter++)  {
    if (*LL_ULong(&(UnlinkedFibrilList->references), counter) != 0)  {
        loopFibril = (fibril *) FElement(&(UnlinkedFibrilList->data), counter);
        if (loopFibril->stem != NULL)  {
        if (loopFibril->stem->bundle_ptr->variable_ptr == StemVariable)  {  // &&
//                (((loopFibril->stem_offset == StemOffset) && (loopFibril->if_absolute_index == yz_true)) ||
//                        (loopFibril->stem_member == StemMemberNumber)))  {
            ClippedStem = loopFibril->stem;
            loopFibril->stem = NULL;
            DerefFibril(&ClippedStem);
    }}  }}
    
    HoldBundlePtr = *UnlinkedBundle;
    *UnlinkedBundle = NULL;
    DerefBundle(&HoldBundlePtr);        // We just deleted a link, so in any event dereference UnlinkedBundle.    
    
    return;
}


// AddMember()  creates a new member for a variable, at position NewMemberNumber in the member linked-list and having NewIndices.
// The member_ID, type and constant flag are set separately, since it's awkward to pass the appropriate variables.
// A bundle in the target variable is created automatically if DestBundle == NULL --> new space is being created (i.e. done with a :=).
// Or, new space is created if DestIndex > DestVariable->instances.

run_retrn AddMember(bundle *CurrentBundle, Ulong *NewMemberNumber, Ulong NewIndices, Ulong IndexWidth, Boolean if_proxy,
                                                                                    member **NewMember, Boolean IfHidden)
{
    variable *NewProxyVariable, *HostVariable = CurrentBundle->variable_ptr;
    linkedlist *memberLL = &(HostVariable->mem.members);
    bundle *NewProxyBundle;
//    member *loopMember;
    Ulong counter;
    run_retrn retrn;
    LL_retrn linkedLretrn;
    
    
        // first bump up the stem_member fields of all fibrils whose stems are in this variable & stem_members are above the insertion point 
        
/*    for (counter = NewMemberNumber; counter <= HostVariable->mem.members.elementnum; counter++)  {
        loopMember = (member *) FElement(&(HostVariable->mem.members), counter);
        if (loopMember->member_ptr != NULL)  {
            AdjustFibrilStemMembers(&(loopMember->member_ptr->variable_ptr->fibril_list), HostVariable, counter, 1);
    }   }*/
    
    
        // add the new member and write its fields
    
    *NewMemberNumber = memberLL->elementnum + 1;
    linkedLretrn = FInsertElements(memberLL, *NewMemberNumber, 1, uncleared);
    if (linkedLretrn != passed)  return out_of_mem_err;
    
    *NewMember = (member *) FElement(memberLL, *NewMemberNumber);
    (*NewMember)->indices = NewIndices;
    (*NewMember)->member_ptr = NULL;
    (*NewMember)->if_hidden = IfHidden;
    
    linkedLretrn = FNewLinkedList(&((*NewMember)->code_list), 0, sizeof(code_ref), LLFreeSpace, uncleared);
    if (linkedLretrn != passed)  return out_of_mem_err;
    
    
        // fix the first_array field
        
    if (HostVariable->first_array >= *NewMemberNumber)  {
        if ((NewIndices != 1) || (IfHidden == yz_true))  HostVariable->first_array = *NewMemberNumber;
        else  HostVariable->first_array++;            }
    
    if (if_proxy == yz_true)  {
        retrn = AddPrimitiveVariable(&NewProxyVariable, proxy_type, sizeof(bundle *));
        if (retrn != passed)  return retrn;
        NewProxyVariable->type = proxy_type;
        
        retrn = AddBundle(NewProxyVariable, (*NewMember)->indices*IndexWidth, 0, &NewProxyBundle, yz_true);
        if (retrn != passed)  return retrn;
        (*NewMember)->member_ptr = NewProxyBundle;
        RefBundle(NewProxyBundle);
    
        for (counter = 1; counter <= (*NewMember)->indices*IndexWidth; counter++)  {     // starts at 1 for bundle_offset = 0
            *(bundle **) FElement(&(NewProxyVariable->mem.data), NewProxyBundle->bundle_offset+counter) = NULL;
    }   }
    
    return passed;
}


// RemoveMember() deletes a member from the members LL of theVariable.
// Readjusts fibril stem_members, first_array.
// Assumes the member has already been unlinked.

void RemoveMember(variable *theVariable, Ulong DoomedMemberNumber)
{
    Ulong counter;
    member *DoomedMember, *loopMember;

        // lower the stem_member fields of all fibrils whose stems are in this variable & stem_members are above the insertion point 

/*    for (counter = DoomedMemberNumber+1; counter <= theVariable->mem.members.elementnum; counter++)  {
        loopMember = (member *) FElement(&(theVariable->mem.members), counter);
        if (loopMember->member_ptr != NULL)  {
            AdjustFibrilStemMembers(&(loopMember->member_ptr->variable_ptr->fibril_list), theVariable, counter, -1);
    }   }*/

    DoomedMember = (member *) FElement(&(theVariable->mem.members), DoomedMemberNumber);


        // fix the first_array field
        
    if (theVariable->first_array == DoomedMemberNumber)  {
        theVariable->first_array = theVariable->mem.members.elementnum;        // = # of elements + 1 after we're done deleting
        for (counter = DoomedMemberNumber+1; counter <= theVariable->mem.members.elementnum; counter++)  {
            loopMember = (member *) FElement(&(theVariable->mem.members), counter);
            if ((loopMember->indices != 1) || (loopMember->if_hidden == yz_true))  {
                theVariable->first_array = counter-1;          // will be the correct value after the deletion
                break;
    }   }   }
    else if (theVariable->first_array > DoomedMemberNumber)  (theVariable->first_array)--;

        // delete the member & associated codes & references
    
    if (DoomedMember->member_ptr != NULL)  DerefBundle(&(DoomedMember->member_ptr));
    
    for (counter = 1; counter <= DoomedMember->code_list.elementnum; counter++)  {
        Deref_CodeRef((code_ref *) FElement(&(DoomedMember->code_list), counter));   }
    FDeleteLinkedList(&(DoomedMember->code_list));
    
    FDeleteElement(&(theVariable->mem.members), DoomedMemberNumber);

    return;     // we ignored return values from FDeleteElement -- an error is possible, but seems implausible to me.
}


// AdjustFibrilStemMembers() changes the stem_member field of any fibril leading to member >= StemMemberNumber of StemVariable.
// This is needed if StemVariable acquired/lost members that change the member numbers around.

/*void AdjustFibrilStemMembers(pinned_LL *theFibrils, variable *StemVariable, Ulong StemMemberNumber, signed long StemMemberChange)
{
    fibril *loopFibril;
    Ulong counter;
    
    for (counter = 1; counter <= theFibrils->data.elementnum; counter++)  {
    if (*LL_ULong(&(theFibrils->references), counter) != 0)  {
        loopFibril = (fibril *) FElement(&(theFibrils->data), counter);
        if (loopFibril->stem != NULL)  {
        if ((loopFibril->stem->bundle_ptr->variable_ptr == StemVariable) && (loopFibril->stem_member == StemMemberNumber))  {
            loopFibril->stem_member += StemMemberChange;
    }}  }}
}*/


// Adds a fibril to an existing bundle.
// Writes the PLL_index, references, stem and bundle_ptr fields; does not reference the fibril initially.

run_retrn AddFibril(bundle *DestBundle, fibril *NewStemFib, fibril **NewFibril, Ulong Offset, Ulong Indices, Boolean IfAbsIndex,
                    Ulong SourceCode, Ulong StemMember, bundle *Jamb)
{
    Ulong NewFibrilIndex;
    run_retrn retrn;

    retrn = AddPLLElement(&(DestBundle->variable_ptr->fibril_list), (void **) NewFibril, &NewFibrilIndex);
    if (retrn != passed)  return retrn;

    (*NewFibril)->PLL_index = NewFibrilIndex;
    (*NewFibril)->references = LL_ULong(&(DestBundle->variable_ptr->fibril_list.references), NewFibrilIndex);
    (*NewFibril)->stem_offset = Offset;
    (*NewFibril)->stem_indices = Indices;
    (*NewFibril)->if_absolute_index = IfAbsIndex;
    (*NewFibril)->stem = NewStemFib;
    (*NewFibril)->bundle_ptr = DestBundle;
    (*NewFibril)->source_code = SourceCode;
//    (*NewFibril)->stem_member = StemMember;
    (*NewFibril)->jamb = Jamb;

    *((*NewFibril)->references) = 0;
    return passed;
}


// References a fibril; if first ref. then it also references its code, bundle & stem.

void RefFibril(fibril *theFibril)
{
    if (theFibril == NULL)  return;
    
    RefPLLPtr(theFibril->references);
    
    if (*(theFibril->references) == 1)  {
        RefFibril(theFibril->stem);
        RefPLLElement(&MasterCodeList, theFibril->source_code);
        RefBundle(theFibril->bundle_ptr);
        RefBundle(theFibril->jamb);     }
        
    return;
}


// Derefs the fibril; if refs == 0 it deletes the fibril & dereferences the associated code, bundle & stem.

void DerefFibril(fibril **theFibril)
{
    if (*theFibril == NULL)  return;
    
    DerefPLLPtr(&((*theFibril)->bundle_ptr->variable_ptr->fibril_list), (*theFibril)->PLL_index, (*theFibril)->references);
    if (*((*theFibril)->references) != 0)  {  *theFibril = NULL;  return;  }

            // if we're being deleted, we aren't the anchor for any other fibrils

    DerefFibril(&((*theFibril)->stem));
    DerefBundle(&((*theFibril)->bundle_ptr));
    DerefPLLElement(&MasterCodeList, (*theFibril)->source_code);
    DerefBundle(&((*theFibril)->jamb));

    *theFibril = NULL;
    
    return;
}


// DrawFibril() finds a fibril leading from the given stem to DestBundle, or creates one if none exists already.
// It is used when we splice fibrils with a <<, or when transform() creates a suspensor (stem == NULL).

run_retrn DrawFibril(bundle *DestBundle, fibril *NewStemFib, fibril **NewFibril,
                    Ulong Offset, Ulong Indices, Boolean IfAbsIndex, Ulong SourceCode, Ulong StemMemberNumber, bundle *Jamb)
{
    fibril *loopFibril;
    Ulong counter;
    
    *NewFibril = NULL;
    for (counter = 1; counter <= DestBundle->variable_ptr->fibril_list.data.elementnum; counter++)  {
    if (*LL_ULong(&(DestBundle->variable_ptr->fibril_list.references), counter) != 0)  {
        loopFibril = (fibril *) FElement(&(DestBundle->variable_ptr->fibril_list.data), counter);
        if ((loopFibril->stem == NewStemFib) && (loopFibril->stem_offset == Offset) && (loopFibril->stem_indices == Indices)
//                && (loopFibril->if_absolute_index == IfAbsIndex) && (loopFibril->stem_member == StemMemberNumber) && (loopFibril->jamb == Jamb) )  {
                && (loopFibril->if_absolute_index == IfAbsIndex) && (loopFibril->bundle_ptr == DestBundle) && (loopFibril->jamb == Jamb) )  {
            *NewFibril = loopFibril;           // then use this one instead of making a new fibril
            return passed;
    }}  }

    return(AddFibril(DestBundle, NewStemFib, NewFibril, Offset, Indices, IfAbsIndex, SourceCode, StemMemberNumber, Jamb));
}


// AddCode() adds an entry to the global MasterCodeList, copying the code from CodeSource.
// The reason to copy the code is that Yazoo will have a personal copy regardless
// of what happens to CodeSource (which may be stored in a user variable).

run_retrn AddCode(Slong *CodeSource, Slong **NewCodePtr, Ulong *NewCodeElement, Ulong CodeLength)
{
    void *NewPLLEntry;
    Ulong counter;
    LL_retrn linkedLretrn;

    linkedLretrn = AddPLLElement(&MasterCodeList, &NewPLLEntry, NewCodeElement);
    if (linkedLretrn != passed)  return out_of_mem_err;


        // the code copy is stored in a buffer, not a linked list, since it can't be resized
        
    *NewCodePtr = (Slong *) malloc((CodeLength+1)*sizeof(Ulong));
    if (*NewCodePtr == NULL)  return out_of_mem_err;
    *(Slong **) NewPLLEntry = *NewCodePtr;

    for (counter = 0; counter < CodeLength; counter++)
        *((*NewCodePtr)+counter) = *(CodeSource+counter);

    *((*NewCodePtr)+CodeLength) = 0;

    return passed;
}


// References an entry in the global code list.

void RefCode(Ulong *RefsPtr)
{  RefPLLPtr(RefsPtr);  return;  }


// De-references an entry in the global code list; frees its memory if refs == 0.

void DerefCode(Ulong *RefsPtr, Ulong RefsID)
{
    DerefPLLPtr(&MasterCodeList, RefsID, RefsPtr);
    if (RefsPtr == NULL)  free(*(void **) FElement(&(MasterCodeList.data), RefsID));
}


// AddCodeRef() adds an entry into the target code list.  References it once (this assumes that the variable is referenced).

run_retrn AddCodeRef(linkedlist *DestCodeLL, fibril *NewAnchor, Slong *CodePtr, Ulong CodeID)
{
    code_ref *NewCodeRef;
    LL_retrn linkedLretrn;

    linkedLretrn = AddElements(DestCodeLL, 1, uncleared);
    if (linkedLretrn != passed)  return out_of_mem_err;

    NewCodeRef = (code_ref *) FElement(DestCodeLL, DestCodeLL->elementnum);
    NewCodeRef->PLL_index = CodeID;
    NewCodeRef->references = (Ulong *) FElement(&(MasterCodeList.references), CodeID);
    NewCodeRef->code_ptr = CodePtr;
    NewCodeRef->anchor = NewAnchor;

    Ref_CodeRef(NewCodeRef);

    return passed;
}


// References a code_ref, and its anchor.
// By convention, we always reference/dereference the anchor with each ref/deref of the code_ref,
// probably because each invokation of Ref_CodeRef corresponds to the making of a full new copy of that code ref.

void Ref_CodeRef(code_ref *theCodeRef)
{
    RefCode(theCodeRef->references);
    RefFibril(theCodeRef->anchor);     // only if a variable code_ref

    return;
}


// Dereferences a code_ref.

void Deref_CodeRef(code_ref *theCodeRef)
{
    DerefCode(theCodeRef->references, theCodeRef->PLL_index);
    DerefFibril(&(theCodeRef->anchor));    // will be nonzero if from a variable; zero if from a member code_ref

    return;
}


// Makes sure no bundles occupies the same set of indices that will be resized is CurrentBundle.
// This is called by ResizeIndices() (in hobbsh.c(pp)), five times sequentially, incrementing the Mode parameter.
// Mode = cbo_set_flags (1) sets the busy_resize_flags; cbo_check (2) returns error upon finding an overlap;
// cbo_unjam (3) unjams any unjammable bundles in the way; cbo_set_no_add (4) does something strange -- prevents
// primitive bundles identical to CurrentBundle from being resized in AddMemory() below w/o causing an error (why?);
// cbo_unset_flags (5) clears the busy_resize_flags.

run_retrn CheckBundleOverlap(bundle *CurrentBundle, Ulong CurrentOffset, Ulong IndexWidth, char Mode)
{
    Ulong NextOffset, NextIndexWidth, counter, c2;
    member *loopMember;
    pinned_LL *BundlesPLL;
    bundle *loopBundle;
    run_retrn retrn;

    if (CurrentBundle->variable_ptr->type == composite_type)  {
        if (IfBusy(CurrentBundle, busy_overlap_flag) != yz_false)  return passed;
        SetBusy(CurrentBundle, busy_overlap_flag);

        for (counter = 1; counter <= CurrentBundle->variable_ptr->mem.members.elementnum; counter++)  {
            Ulong loop_top, step_indices;
            
            loopMember = (member *) FElement(&(CurrentBundle->variable_ptr->mem.members), counter);
            
            if (loopMember->member_ptr->variable_ptr->type != proxy_type)  {  loop_top = 1;  step_indices = loopMember->indices;  }
            else  {  loop_top = loopMember->indices;  step_indices = 1;  }
            
            for (c2 = 0; c2 < loop_top; c2++)       {
                loopBundle = CurrentBundle;
                NextOffset = 0;
                NextIndexWidth = IndexWidth;
                StepBundle(&loopBundle, &NextOffset, &NextIndexWidth, loopMember, c2, step_indices);
                if (err_code == passed)  {
                    retrn = CheckBundleOverlap(loopBundle, NextOffset, NextIndexWidth, Mode);
                    if (retrn != passed)  return retrn;      }
                else  err_code = passed;            // the member wasn't allocated
        }   }
        return passed;
    }

    if (Mode == cbo_set_flags)  {
        SetBusy(CurrentBundle, busy_resize_flag);
        return passed;       }
    else if (Mode == cbo_unset_flags)  {
        ClearBusy(CurrentBundle, busy_resize_flag);
        return passed;       }

    BundlesPLL = &(CurrentBundle->variable_ptr->bundles);
    for (counter = 1; counter <= BundlesPLL->data.elementnum; counter++)  {
    if ((counter != CurrentBundle->PLL_index) && (*LL_ULong(&(BundlesPLL->references), counter) != 0))  {
        loopBundle = (bundle *) FElement(&(BundlesPLL->data), counter);
        if ((loopBundle->bundle_offset <= CurrentOffset + IndexWidth - 1) &&
                    (loopBundle->bundle_offset + loopBundle->instances - 1 >= CurrentOffset))  {
            if ((IfBusy(loopBundle, busy_resize_flag) != yz_false) && (CurrentBundle->bundle_offset == loopBundle->bundle_offset)
                     && (CurrentBundle->instances == loopBundle->instances) && (loopBundle->jam_status != unjammed))  {
                if ((CurrentBundle->PLL_index < loopBundle->PLL_index) && (Mode == cbo_set_no_add))      {
                    SetBusy(loopBundle, busy_no_resize_flag);   }   }     // otherwise, bof is only set in composite variables
            else  {
                if ((Mode == cbo_check) && (loopBundle->jam_status == can_jam))  {
                    return overlapping_bundle_err;         }
                else if ((Mode == cbo_unjam) && (loopBundle->jam_status == cannot_jam))  {
                    loopBundle->jam_status = unjammed;
                    ClearBusy(loopBundle, busy_resize_flag);  }      }
        }
    }}
    
    return passed;
}


// AddMemory(), sort of a misnomer, either adds or deletes memory as a result of a resize, or the creation of a new bundle.
// Adjusts the offsets of the bundles and indexes of PC path accordingly.
// Calls itself recursively for the members of the bundle's variable if it is composite.
// Outermost invocation:  use offset for InsertionOffset, and # of elements of the member pointing to CurrentBundle to add for NewIndices.

run_retrn AddMemory(bundle *theBundle, Ulong InsertionOffset, signed long NewIndices)
{
    Ulong counter;
    member *loopMember;
    bundle *loopBundle, **loopProxy;
    bundle_coordinate *loopbundle_coordinate;
    variable *theVariable = theBundle->variable_ptr;
    linkedlist *loopStringLL;
    run_retrn retrn;
    LL_retrn linkedLretrn;

    if (NewIndices == 0)  return passed;

    theVariable->instances += NewIndices;


        // adjust the offset/index_width of any bundle that could conceivably be affected

    for (counter = 1; counter <= theVariable->bundles.data.elementnum; counter++)  {
    loopBundle = (bundle *) FElement(&(theVariable->bundles.data), counter);
    if ((loopBundle == theBundle) || (*LL_ULong(&(theVariable->bundles.references), counter) != 0))  {    // 1st cond:  if theBundle is not yet ref'd
        AdjustOffset_And_IW(loopBundle, theBundle, &(loopBundle->bundle_offset), &(loopBundle->instances), InsertionOffset, NewIndices);
        if ((loopBundle->bundle_offset >= InsertionOffset) && (loopBundle->remote_fibril != NULL))  {
            loopBundle->remote_fibril->stem_offset += NewIndices;
    }}  }

    if (PCCoordinate._Bundle->variable_ptr == theVariable)
        AdjustOffset_And_IW(PCCoordinate._Bundle, theBundle, &(PCCoordinate._Offset), &(PCCoordinate._IndexWidth), InsertionOffset, NewIndices);
    if (SearchCoordinate._Bundle != NULL)  {
    if (SearchCoordinate._Bundle->variable_ptr == theVariable)  {
        AdjustOffset_And_IW(SearchCoordinate._Bundle, theBundle, &(SearchCoordinate._Offset), &(SearchCoordinate._IndexWidth), InsertionOffset, NewIndices);
    }}
    if (GL_Object.Target._Bundle != NULL)  {
    if (GL_Object.Target._Bundle->variable_ptr == theVariable)  {
        AdjustOffset_And_IW(GL_Object.Target._Bundle, theBundle, &(GL_Object.Target._Offset), &(GL_Object.Target._IndexWidth),
                                                            InsertionOffset, NewIndices);
    }}

    for (counter = 1; counter <= PCStack.top; counter++)  {
        loopbundle_coordinate = (bundle_coordinate *) FElement(&(PCStack.data), counter);
        if (loopbundle_coordinate->_Bundle != NULL)  {
        if (loopbundle_coordinate->_Bundle->variable_ptr == theVariable)     {
            AdjustOffset_And_IW(loopbundle_coordinate->_Bundle, theBundle, &(loopbundle_coordinate->_Offset), &(loopbundle_coordinate->_IndexWidth),
                                        InsertionOffset, NewIndices);
    }   }}


        // recursively call AddMemory() for each member, if it is a composite variable
    
    if (theVariable->type == composite_type)  {
        if (IfBusy(theVariable, busy_add_flag) != yz_false)  return passed;   // I use a more elaborate scheme for preventing double-allocation of
        SetBusy(theVariable, busy_add_flag);                                  // memory for primitive vars -- which is better?

        for (counter = 1; counter <= theVariable->mem.members.elementnum; counter++)  {
            loopMember = (member *) FElement(&(theVariable->mem.members), counter);
            if ((loopMember->member_ptr != NULL) && (loopMember->member_ptr->jam_status != unjammed))  {
                if (((double) NewIndices)*loopMember->indices > LONG_MAX)  return index_argument_err;
                retrn = AddMemory(loopMember->member_ptr, InsertionOffset*(loopMember->indices)+loopMember->member_ptr->bundle_offset,
                                                                                        NewIndices*loopMember->indices);
                if (retrn != passed)  return retrn;
        }   }

        if (GL_Path.StemCoordinate._Bundle != NULL)  {
        if (GL_Path.StemCoordinate._Bundle->variable_ptr == theVariable)  {
            AdjustOffset_And_IW(GL_Path.StemCoordinate._Bundle, theBundle, &(GL_Path.StemCoordinate._Offset), &(GL_Path.StemCoordinate._IndexWidth),
                                                                                InsertionOffset, NewIndices);
            GL_Path.StemMember = NULL;       // we can't tell if it was added to or pushed up
        }}

        return passed;      }

    if (theVariable->type == proxy_type)  {
        AdjustProxyFibrilOffsets(theVariable, InsertionOffset+1, NewIndices);
        if (NewIndices < 0)  {
        for (counter = InsertionOffset+1; counter <= InsertionOffset+(-NewIndices); counter++)  {
            loopProxy = (bundle **) FElement(&(theVariable->mem.data), counter);
            if (*loopProxy != NULL)  DerefBundle(loopProxy);
    }   }}
    
    
        // if a primitive variable, add the memory directly to the linked lists
        
    if (IfBusy(theBundle, busy_no_resize_flag) != yz_false)  {
        ClearBusy(theBundle, busy_no_resize_flag);      }    // I don't understand this step (see comment in CheckBundleOverlap()).
    else  {
        if (NewIndices > 0)  {
            linkedLretrn = FInsertElements(&(theVariable->mem.data), InsertionOffset+1, NewIndices, cleared);
            if (linkedLretrn != passed)  return out_of_mem_err;
            if (theVariable->type == string_type)  {
            for (counter = InsertionOffset+1; counter <= InsertionOffset+NewIndices; counter++)  {
                ((linkedlist *) FElement(&(theVariable->mem.data), counter))->memory = NULL;
        }   }}
        else  {
            if (theVariable->type == string_type)  {
                for (counter = InsertionOffset+1; counter <= InsertionOffset+(-NewIndices); counter++)  {
                    loopStringLL = (linkedlist *) FElement(&(theVariable->mem.data), counter);
                    if (loopStringLL->memory != NULL)  FDeleteLinkedList(loopStringLL);
            }   }
            FDeleteElements(&(theVariable->mem.data), InsertionOffset+1, InsertionOffset+(-NewIndices));
    }   }

    return passed;
}


// AO&IW() updates the offset and/or index width of a bundle after addition/removal of elements of memory in that bundle's variable.
// The offset is affected if the new/deleted memory is below the current bundle; the index width if affected if it is in the bundle.

void AdjustOffset_And_IW(bundle *theBundle, bundle *ResizedBundle, Ulong *Offset, Ulong *IndexWidth, Ulong InsertionOffset, signed long NewIndices)
{
    if ((theBundle == ResizedBundle) && (*Offset <= InsertionOffset))  *IndexWidth += NewIndices;     // 2nd condition for PC/SearchCoordinates
    else if (*Offset >= InsertionOffset)  *Offset += NewIndices;      // no possibility of partial overlap with deletion set for NewIndices < 0
}


// APFO() adjusts the stem offsets of proxy fibrils whose stem bundle has experienced an offset change.

void AdjustProxyFibrilOffsets(variable *ProxyVariable, Ulong BottomIndexToAdjust, signed long OffsetChange)
{
    bundle *loopBundle;
    fibril *loopFibril;
    Ulong loopProxyIndex, loopFibrilIndex;
    
    for (loopProxyIndex = BottomIndexToAdjust; loopProxyIndex <= ProxyVariable->mem.data.elementnum; loopProxyIndex++)  {
        loopBundle = *(bundle **) FElement(&(ProxyVariable->mem.data), loopProxyIndex);
        if (loopBundle != NULL)  {
            for (loopFibrilIndex = 1; loopFibrilIndex <= loopBundle->variable_ptr->fibril_list.data.elementnum; loopFibrilIndex++)  {
                loopFibril = (fibril *) FElement(&(loopBundle->variable_ptr->fibril_list.data), loopFibrilIndex);
                if (*LL_ULong(&(loopBundle->variable_ptr->fibril_list.references), loopFibrilIndex) != 0)       {
                if (loopFibril->stem != NULL)            {
                if ((loopFibril->stem->bundle_ptr->variable_ptr == ProxyVariable) && (loopFibril->stem_offset == loopProxyIndex)
                                                                        && (loopFibril->bundle_ptr == loopBundle))      {
                    loopFibril->stem_offset += OffsetChange;
    }   }  }   }}}

    return;
}


// UnflagVariables() removes the busy flag on theVariable, its members' variables, etc.
// This needs to run after each outer AddMemory() call, in order for AM() to work again.

void UnflagVariables(variable *theVariable, unsigned char theFlag)
{
    member *loopMember;
    Ulong counter;

    if (IfBusy(theVariable, theFlag) == yz_false)  return;
    ClearBusy(theVariable, theFlag);

    if (theVariable->type == composite_type)  {
        for (counter = 1; counter <= theVariable->mem.members.elementnum; counter++)  {
            loopMember = (member *) FElement(&(theVariable->mem.members), counter);
            if (loopMember->member_ptr != NULL)  UnflagVariables(loopMember->member_ptr->variable_ptr, theFlag);  }}

    // Don't unflag proxy bundles; Add/DelMemory and CheckBundleOverlap() don't step into proxies.

    return;
}


// UnflagBundles() removes the busy flag on CurrentBundle, its variable's members' bundles, etc.
// This needs to run after each outer CheckBundleOverlap() call, in order for CBO() to work again.

void UnflagBundles(bundle *CurrentBundle, unsigned char theFlag)
{
    bundle *NextBundle;
    Ulong DummyOffset, DummyIndexWidth, counter, c2;
    Ulong HoldErrCode, HoldErrIndex = 0;
    code_ref HoldErrScript;
    
    if (CurrentBundle == NULL)  return;       // it's an undefined proxy
    if (IfBusy(CurrentBundle, theFlag) == yz_false)  return;
    ClearBusy(CurrentBundle, theFlag);

    if (err_code != passed)  {         // since we still try to unflag bundles even if there's been an error -- preserve that error.
        HoldErrCode = err_code;  HoldErrIndex = err_index;  HoldErrScript = err_script;  err_script.code_ptr = NULL;
        err_code = passed;        }
    else  HoldErrCode = passed;

    if (CurrentBundle->variable_ptr->type == composite_type)  {
        for (counter = 1; counter <= CurrentBundle->variable_ptr->mem.members.elementnum; counter++)  {
            Ulong loop_top;
            member *loopMember = (member *) FElement(&(CurrentBundle->variable_ptr->mem.members), counter);
            
            if (loopMember->member_ptr->variable_ptr->type != proxy_type)   loop_top = 1;
            else  loop_top = loopMember->indices;
            
            for (c2 = 0; c2 < loop_top; c2++)       {
                NextBundle = CurrentBundle;
                DummyOffset = 0;
                DummyIndexWidth = 1;
                StepBundle(&NextBundle, &DummyOffset, &DummyIndexWidth, loopMember, c2, 1);
                if (err_code == passed)  UnflagBundles(NextBundle, theFlag);
                else  err_code = passed;        }
    }   }
    
    if (HoldErrCode != passed)  {
        if (err_script.code_ptr != NULL)  Deref_CodeRef(&err_script);
        err_code = HoldErrCode;  err_index = HoldErrIndex;  err_script = HoldErrScript;        }

    return;
}


// UnflagBundlesAndStems() differs from UnflagBundles() in that it removes the busy flag from proxy bundles and
// from bundles that are connected to stems of CurrentBundle's variable's fibrils.
// This needs to run after each CombBundles() (used by SpringCleaning()), or else there may be memory leaks.

void UnflagBundlesAndStems(bundle *CurrentBundle, unsigned char theFlag)
{
    fibril *loopFibril;
    Ulong counter;
    Ulong HoldErrCode, HoldErrIndex = 0;
    code_ref HoldErrScript;

    if (CurrentBundle == NULL)  return;        // it's an undefined member or proxy
    if (IfBusy(CurrentBundle, theFlag) == yz_false)  return;

    ClearBusy(CurrentBundle, theFlag);
    ClearBusy(CurrentBundle->variable_ptr, theFlag);

    if (err_code != passed)  {       // since we still try to unflag bundles even if there's been an error -- preserve that error.
        HoldErrCode = err_code;  HoldErrIndex = err_index;  HoldErrScript = err_script;  err_script.code_ptr = NULL;
        err_code = passed;      }
    else  HoldErrCode = passed;

    if (CurrentBundle->variable_ptr->type == composite_type)  {
        for (counter = 1; counter <= CurrentBundle->variable_ptr->mem.members.elementnum; counter++)  {
            UnflagBundlesAndStems(((member *) FElement(&(CurrentBundle->variable_ptr->mem.members), counter))->member_ptr, theFlag);  }}
    else if (CurrentBundle->variable_ptr->type == proxy_type)  {
        for (counter = 1; counter <= CurrentBundle->instances; counter++)  {
            UnflagBundlesAndStems(*(bundle **) FElement(&(CurrentBundle->variable_ptr->mem.data), counter+CurrentBundle->bundle_offset), theFlag);  }}

    for (counter = 1; counter <= CurrentBundle->variable_ptr->fibril_list.data.elementnum; counter++)  {
    if (*LL_ULong(&(CurrentBundle->variable_ptr->fibril_list.references), counter) != 0)  {
        loopFibril = (fibril *) FElement(&(CurrentBundle->variable_ptr->fibril_list.data), counter);
        if (loopFibril->stem != NULL)  {
            UnflagBundlesAndStems(loopFibril->stem->bundle_ptr, theFlag);
    }}}

    if (HoldErrCode != passed)  {
        if (err_script.code_ptr != NULL)  Deref_CodeRef(&err_script);
        err_code = HoldErrCode;  err_index = HoldErrIndex;  err_script = HoldErrScript;     }

    return;
}


// Align() returns the smallest integer that is both >= the_long and is a multiple of size(Ulong) (which is 4).
// Used to ensure memory alignment in various parts of Yazoo.

Ulong Align(Ulong the_long)
{
    if (the_long % sizeof(Ulong) != 0)
        return sizeof(Ulong) - (the_long % sizeof(Ulong));

    return 0;
}
