/*
 *  intrpt.h
 *
 *  Yazoo
 *  Copyright (C) 2005 Brian C. Ross
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of the License, or any later version.

 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef Interpret_h
#define Interpret_h

#include <math.h>
#include "lnklst.h"
#include "yazooh.h"


// The type used for passing runtime errors

typedef unsigned char run_retrn;


// Runtime error codes

#define out_of_mem_err 1

#define illegal_command_err 2
#define code_overflow_err 3
#define inaccessible_code_warning 4
#define bad_jump_err 5

#define out_of_range_err 6
#define divide_by_zero_warning 7

#define member_not_found_err 8
#define no_member_err 9
#define invalid_index_err 10
#define void_member_err 11
#define void_proxy_err 12
#define step_multiple_members_err 13
#define step_multiple_proxies_err 14

#define underflow_warning 15
#define type_mismatch_err 16
#define argument_out_of_range_err 17
#define self_reference_err 18
#define illegal_target_err 19
#define undefined_member_err 20
#define target_deleted_err 21
#define incomplete_member_err 22
#define incomplete_variable_err 23
#define invalid_block_size_err 24
#define non_numeric_err 25
#define not_a_variable_err 26
#define unequal_data_size_err 27
#define read_number_err 28
#define no_target_err 29      /* not used */
#define mismatched_indices_err 30
#define overlapping_bundle_err 31
#define unjammed_member_err 32
#define unjammed_proxy_err 33

#define not_composite_err 34
#define undefined_string_err 35
#define multiple_indices_not_allowed_err 36
#define wrong_argument_count_err 37
#define string_expected_err 38
#define index_argument_err 39
#define argument_err 40
#define nonexistent_register_err 41
#define nonexistent_built_in_function_err 42
#define user_function_not_found_err 43

#define recursion_depth_err 44

#define IO_error 45

#define return_flag 46
#define finished_signal 47


// Flags used in the "business" field of variables & bundles

#define busy_SS_flag (1<<0)
#define busy_source_copy_flag (1<<1)   /* bundles */
#define busy_dest_copy_flag (1<<2)     /* bundles */
#define busy_comb_flag (1<<3)          /* v & b */
#define busy_no_resize_flag (1<<3)     /* bundles */
#define busy_add_flag (1<<4)           /* variables */
#define busy_overlap_flag (1<<5)       /* bundles */
#define busy_SRW_flag (1<<6)           /* bundles */
#define busy_resize_flag (1<<7)        /* bundles */
#define busy_print_flag (1<<7)         /* bundles */
#define IfBusy(a,b) ((a->business) & (b))
#define SetBusy(a,b) ((a->business) = (unsigned char) ((a->business) | (unsigned char) (b)))
#define ClearBusy(a,b) ((a->business) = (unsigned char) ((a->business) & (~(unsigned char) (b))))


// jam status

#define cannot_jam yz_false
#define can_jam yz_true
#define unjammed 2


// flags used by CheckBundleOverlap()

#define cbo_set_flags 1
#define cbo_check 2
#define cbo_unjam 3
#define cbo_set_no_add 4
#define cbo_unset_flags 5


// members of Zero

#define _yz_registers 1
#define _workspace1 2



// Structures forming Yazoo's runtime memory


// variable:  contains either physical storage (primitive variable), or a list of members (composite)

typedef struct {
    Ulong PLL_index;            // element_number in variable PLL
    Ulong *references;          // repeated here for speed; # of bundles + variable_ptrs

    union {
        linkedlist members;
        linkedlist data;
    } mem;

    pinned_LL bundles;          // one fibril list for each instance of this variable
    pinned_LL fibril_list;

    linkedlist code_list;       // of type code_ref
    unsigned char type;         // for primitive types
    unsigned char business;

    Ulong first_array;          // for efficiency in searching var tree -- the first member which has > 1 elements
    Ulong instances;            // how many independent instances of this type of variable are allocated (not counting =@)
                                // for primitive & proxy types, = data.elementnum
} variable;


// bundle:  represents a region of a variable accessible by a member in (usually) another variable

typedef struct {
    Ulong PLL_index;            // element_number in bundle PLL
    Ulong *references;          // repeated here for speed; # of fibrils + member refs

    variable *variable_ptr;     // supporting this bundle
    Ulong bundle_offset;        // the offset for this bundle in the LL or portion of subsequent LLs used by this var
                                // bundle_offset starts at zero, so add one for the FElement().
    Ulong instances;            // CUMULATIVE -- #_of of immediate predecessor times #_of of the one before, ...

    char jam_status;            // if this bundle can jam the variable to prevent resizes
    unsigned char business;     // to use when checking to see if a bundle is part of an 'island'

    struct fib_temp *remote_fibril;      // fibril that uses this bundle as a jamb to keep data from being deleted, even though it is not a member
} bundle;                       // (for updating fibril's stem_offset in case of bundle resize).  NOT referenced to avoid loops -- causes no problems.


// member:  points from one variable to another (bundle)

typedef struct {
    bundle *member_ptr;
    
    signed long member_ID;      // the name; 0 if cannot be found by step_to_memberID (i.e. no name in Yazoo code)
    Ulong indices;              // non-cumulative; e.g., for a[5, 10] := int, it would be 6 for the member starting at the 5th index

    linkedlist code_list;       // parallels the code_list in the targeted variable, but may be less restrictive
    unsigned char type;         // the requirements for this member -- may be met and exceeded in the variable
    Boolean if_hidden;          // hidden_member members don't have an index associated with them
} member;


// fibril:  a step on a search path

typedef struct fib_temp {
    Ulong PLL_index;            // element_number in fibril PLL
    Ulong *references;          // repeated here for speed; # of times used as stem or data anchor, or fibril anchor

    Boolean if_absolute_index;  // (only) if true, use first item in each union
    Ulong stem_offset;          // the total offset when the PC stepped here before creating the code below
    Ulong stem_indices;         // the number of the indices of the stem-member that leads here
//    Ulong stem_member;          // the member # (not index #) that leads here from the stem
    bundle *jamb;               // to prevent a variable from resizing and deleting a necessary offset (used by remote fibrils)

    struct fib_temp *stem;      // the fibril that this came from
    bundle *bundle_ptr;         // points to THIS bundle, which stores offset
    Ulong source_code;          // the global code PLL ID that was being run when this fibril was created -- not the code_ref entry
} fibril;


// code_ref:  an element in a variable or member code list

typedef struct {
    Ulong PLL_index;            // element_number in code PLL
    Ulong *references;          // repeated here for speed; # of times used by a variable
    Slong *code_ptr;
    fibril *anchor;             // the first step on the search path from this code
} code_ref;


// bundle_coordinate:  specifies a location in Yazoo's memory (examples include the 'program counter', SearchCoordinate, etc.)

typedef struct {
    bundle *_Bundle;
    fibril *_Fibril;
    Ulong _Offset;          // the total offset at that particular path element
    Ulong _IndexWidth;
} bundle_coordinate;



// Globals

typedef struct {
    pinned_LL _MasterCodeList;   // pointers to blocks of Hobbish that store code
    pinned_LL _VariableList;     // master list containing all variables
    stack _PCStack;              // for storing the old PCCoordinate, SearchCoordinate during function calls
} yz_interpret_global_struct;

extern yz_interpret_global_struct yz_interpret_globals;

#define MasterCodeList yz_interpret_globals._MasterCodeList
#define VariableList yz_interpret_globals._VariableList
#define PCStack yz_interpret_globals._PCStack




// Prototypes

extern const unsigned char LLFreeSpace;

extern void CopyBundleData(bundle *, Ulong, bundle *, Ulong, Ulong);
extern void Copy_Data(linkedlist *, Ulong, linkedlist *, Ulong, Ulong);
extern void Copy_String(linkedlist *, linkedlist *);
extern void Copy_Numeric(linkedlist *, Ulong, unsigned char, linkedlist *, Ulong, unsigned char);
extern void CompareBundleData(bundle *, Ulong, bundle *, Ulong, Ulong);
extern void Compare_Data(linkedlist *, Ulong, linkedlist *, Ulong, Ulong);
extern void Compare_String(linkedlist *, linkedlist *);
extern void Compare_Numeric(linkedlist *, Ulong, unsigned char, linkedlist *, Ulong, unsigned char);
extern void CCTemplate(bundle *, Ulong, bundle *, Ulong, Ulong,
    void(*CC_Composite)(bundle *, Ulong, bundle *, Ulong, Ulong),
    void(*CC_Data)(linkedlist *, Ulong, linkedlist *, Ulong, Ulong),
    void(*CC_String)(linkedlist *, linkedlist *),
    void(*CC_Numeric)(linkedlist *, Ulong, unsigned char, linkedlist *, Ulong, unsigned char));
extern void FindNextThickMember(variable *, member **, Ulong *);

extern void SizeBundle(bundle *, Ulong, Ulong, void *, void *);
extern void Size_Data(bundle *, Ulong, Ulong, void *, void *Dummy);
extern void Size_String(bundle *, Ulong, void *, void *);
extern void StorageSizeBundle(bundle *, Ulong, Ulong, void *, void *);
extern void StorageSize_Data(bundle *, Ulong, Ulong, void *, void *);
extern void StorageSize_String(bundle *, Ulong, void *, void *);
extern void WriteBundle(bundle *, Ulong, Ulong, void *, void *);
extern void Write_Data(bundle *, Ulong, Ulong, void *, void *Dummy);
extern void Write_String(bundle *, Ulong, void *, void *);
extern void ReadBundle(bundle *, Ulong, Ulong, void *, void *);
extern void Read_Data(bundle *, Ulong, Ulong, void *, void *Dummy);
extern void Read_String(bundle *, Ulong, void *, void *);
extern void ReadBundleString(bundle *, Ulong, Ulong, void *, void *);
extern void Read_DataString(bundle *, Ulong, Ulong IndexWidth, void *, void *Dummy);
extern void SizeBundleString(bundle *, Ulong, Ulong, void *, void *);
extern void Size_DataString(bundle *, Ulong, Ulong, void *, void *Dummy);
extern void WriteBundleString(bundle *, Ulong, Ulong, void *, void *);
extern void Write_DataString(bundle *, Ulong, Ulong, void *, void *Dummy);
extern void Write_StringToString(bundle *, Ulong, void *, void *);
extern void CountDataLists(bundle *, Ulong, Ulong, void *, void *);
extern void Count_DataBundle(bundle *, Ulong, Ulong, void *, void *Dummy);
extern void Count_StringBundle(bundle *, Ulong, void *, void *);
extern void Fill_argv_Handle(bundle *, Ulong, Ulong, void *, void *);
extern void PointTo_DataBundle(bundle *, Ulong, Ulong, void *, void *Dummy);
extern void PointTo_String(bundle *, Ulong, void *, void *);
extern void SRWTemplate(bundle *, Ulong, Ulong, void *, void *, Boolean,
    void(*SLS_Composite)(bundle *, Ulong, Ulong, void *, void *),
    void(*SLS_Data)(bundle *, Ulong, Ulong, void *, void *),
    void(*SLS_String)(bundle *, Ulong, void *, void *));
extern void PrintNumber(char *, double, Ulong *, unsigned char);

extern void SearchMember_DragFibril(signed long, member **, Ulong *);
extern void SearchMember(signed long, member **, Ulong *);
extern void StepSearchBundleAndFibril(member *, Ulong, Ulong, Ulong);
extern void StepSearchBundle(member *, Ulong, Ulong);
extern void StepBundle(bundle **, Ulong *, Ulong *, member *, Ulong, Ulong);
extern void StepToMember(bundle **, Ulong *, member *, Ulong);
extern run_retrn FindMemberID(bundle *, signed long, member **, Ulong *);
extern run_retrn FindMemberIndex(variable *, Ulong, member **, Ulong *, Ulong *);

extern void SetError(Ulong, Slong *);
extern void SetWarning(Ulong, Slong *);

extern run_retrn InitYazoo(Slong *, Ulong);
extern void CleanUp();
extern run_retrn AddCompositeVariable(variable **);
extern run_retrn AddPrimitiveVariable(variable **, Ulong, Ulong);
extern run_retrn AddVariable(variable **);
extern void RefVariable(variable *);
extern void DerefVariable(variable *);
extern run_retrn AddBundle(variable *, Ulong, Ulong, bundle **, Boolean);
extern void RefBundle(bundle *);
extern void DerefBundle(bundle **);
extern void CombBundles();
extern void CombBranch(bundle *);
extern void UnlinkBundle(variable *, bundle **, Ulong, Ulong);
extern run_retrn AddMember(bundle *, Ulong *, Ulong, Ulong, Boolean, member **, Boolean);
extern void RemoveMember(variable *, Ulong);
//extern void AdjustFibrilStemMembers(pinned_LL *, variable *, Ulong, signed long);
extern run_retrn AddFibril(bundle *, fibril *, fibril **, Ulong, Ulong, Boolean, Ulong, Ulong, bundle *);
extern void RefFibril(fibril *);
extern void DerefFibril(fibril **);
extern run_retrn DrawFibril(bundle *, fibril *, fibril **, Ulong, Ulong, Boolean, Ulong, Ulong, bundle *);
extern run_retrn AddCode(Slong *, Slong **, Ulong *, Ulong);
extern void RefCode(Ulong *);
extern void DerefCode(Ulong *, Ulong);
extern run_retrn AddCodeRef(linkedlist *, fibril *, Slong *, Ulong);
extern void Ref_CodeRef(code_ref *);
extern void Deref_CodeRef(code_ref *);
extern run_retrn CheckBundleOverlap(bundle *, Ulong, Ulong, char);
extern run_retrn AddMemory(bundle *, Ulong, signed long);
extern void AdjustOffset_And_IW(bundle *, bundle *, Ulong *, Ulong *, Ulong, signed long);
extern void AdjustProxyFibrilOffsets(variable *, Ulong, signed long);
extern void UnflagVariables(variable *, unsigned char);
extern void UnflagBundles(bundle *, unsigned char);
extern void UnflagBundlesAndStems(bundle *, unsigned char);

extern Ulong Align(Ulong);

#endif
