/*
 *  lnklst.c(pp) - linked list routines
 *
 *  Yazoo
 *  Copyright (C) 2005 Brian C. Ross
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of the License, or any later version.

 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */


// Linked list routines for dynamic memory allocation
// These store lists of variables, of any number of bytes each, in linked lists.
// All routines return "passed" or an error code (or a handle/pointer in cases of NewLinkedList/Element).
// Memory structure:  linkedlist (followed by optional header) -> (points to) sublist1 -> sublist2 -> etc.
// This c(pp) file is compiled twice -- once with LL_FastFlag set and once with it cleared.


#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "lnklst.h"




// **************** Private routines ****************



#ifndef LL_FastFlag         // only want to define these once


// LL_NewSubList() makes a new sublist, to be linked to the queue of a linked sublists.

sublistHeader *LL_NewSubList(unsigned long NumberOfElements, unsigned long MaximumElements, unsigned long ElementSize, Boolean if_cleared)
{
    sublistHeader *temp;

    if (if_cleared == yz_true)
        temp = (sublistHeader *) calloc((size_t) MaximumElements*ElementSize + sublistHeaderSize, (size_t) 1);
    else 
        temp = (sublistHeader *) malloc((size_t) MaximumElements*ElementSize + sublistHeaderSize);

    if (temp != NULL)  {
		temp->num_elements = NumberOfElements;
		temp->max_elements = MaximumElements;
		temp->next_sublist = NULL;     }

    return temp;
}


// LL_FindSubList() returns a pointer to the sublist with the given element in it.

sublistHeader * LL_FindSubList(linkedlist * theList, unsigned long TheElement, unsigned long *bottom_retrn)
{
    sublistHeader * CurrentSublist;
    unsigned long bottom, top;
    
    CurrentSublist = theList->memory;
    top = CurrentSublist->num_elements;
    bottom = 1;

    while (TheElement > top)
    {
        CurrentSublist = CurrentSublist->next_sublist;
        bottom = top+1;
        top += CurrentSublist->num_elements;
    }

    *bottom_retrn = bottom;
    return CurrentSublist;
}

#endif



// **************** Public routines ****************

// [F]NewLinkedList() constructs an empty linked list with a given amount of space; returns pointer to LL if successful.
// Passes requested number of elements and size of each (and whether you want all the memory set to zero automatically).
// Returns a pointer in your theList.memory.
// Spare room is the ratio of spare memory in the list to NumberOfElements, as a percentage from 0-255%.
// The linked list routines try to maintain this level of extra space when the list is dynamically resized.

#ifndef LL_FastFlag
LL_retrn NewLinkedList(linkedlist * theList, unsigned long NumberOfElements, unsigned short ElementSize, unsigned char SpareRoom, Boolean if_cleared)
#else
LL_retrn FNewLinkedList(linkedlist * theList, unsigned long NumberOfElements, unsigned short ElementSize, unsigned char SpareRoom, Boolean if_cleared)
#endif
{
    sublistHeader *temp;
    unsigned long max_els;

    max_els = NumberOfElements+((NumberOfElements*SpareRoom)/100);
    if (max_els < 1) max_els = 1;

    temp = LL_NewSubList(NumberOfElements, max_els, ElementSize, if_cleared);      // allocate one sublist
    theList->memory = temp;
    if (temp == NULL)  return mem_err;

    theList->elementsize = ElementSize;
    theList->elementnum = NumberOfElements;
    theList->spare_room = SpareRoom;

    return passed;
}


// [F]DeleteLinkedList() deletes an entire linked list (including all its sub-lists), and frees the associated memory.

#ifndef LL_FastFlag
LL_retrn DeleteLinkedList(linkedlist *theList)
#else
void FDeleteLinkedList(linkedlist *theList)
#endif
{
    sublistHeader *CurrentSubList, *NextSubList;

    NextSubList = theList->memory;
    while (NextSubList != NULL)
    {
        CurrentSubList = NextSubList;
        NextSubList = CurrentSubList->next_sublist;
        free((void *) CurrentSubList);
    }

    theList->memory = NULL;

#ifndef LL_FastFlag
    return passed;
#else
    return;
#endif
}


// InsertElements() makes room for new elements in the middle of a linked list.
// First new element goes at [InsertionPoint], so all new elements are inserted before the old element at InsertionPoint.

#ifndef LL_FastFlag
LL_retrn InsertElements(linkedlist * theList, unsigned long InsertionPoint, unsigned long NewElements, Boolean if_cleared)
#else
LL_retrn FInsertElements(linkedlist * theList, unsigned long InsertionPoint, unsigned long NewElements, Boolean if_cleared)
#endif
{
    sublistHeader * baseSublist;
    LL_retrn newLL_result;
    linkedlist dummy;
    unsigned long bottom;
    unsigned long ElementsOnTop, CopyPoint, SkipAhead, NewSublistElements;

#ifndef LL_FastFlag
    if (theList->memory == NULL) return init_err;
#endif
    if (InsertionPoint == 1)
    {
        baseSublist = theList->memory;
        bottom = 1;
    }
    else
    {
#ifndef LL_FastFlag
        if (ElementExists(theList, InsertionPoint-1) != yz_true)  return range_err;
#endif
        baseSublist = LL_FindSubList(theList, InsertionPoint-1, &bottom);
    }

    if (NewElements == 0)  return passed;
    ElementsOnTop = baseSublist->num_elements + bottom - InsertionPoint;


        // If necessary, create another sublist.
        // This is a rather hokey way of doing it -- create a new linked list, and from that extract and insert the sublist.

    if (baseSublist->num_elements + NewElements > baseSublist->max_elements)
    {

            // Try to balance it so that both sublists have spare_room number of spare elements.
            // First we allocate more room in a new sublist, then shove everything up to make room for the new inserted elements
            // and then try to fill the lower sublist as full as we want it by copying the top of those higher up into the next sublist
            // and removing the extra elements from the lower one.
            // CopyPoint is in 'local' (to the sublist) coordinates -- refers to where we start shoving stuff upwards.
            // SkipAhead is how far we ought to push everything up so that the first sublist will be as full/empty as we want it.
    
            // In Yazoo, we need to have CopyPoint at the point of the new arrivals.  Since many LLs are 'pinned',
            // AddElements() expects that the ones below stay put.

        CopyPoint = baseSublist->num_elements + 1;
        SkipAhead = baseSublist->num_elements - (CopyPoint - 1);
        if (LONG_MAX-theList->elementnum < NewElements+SkipAhead) return mem_err;


            // Allocate enough room to hold the last max-filled sublist, plus extra spare_room.

        NewSublistElements = SkipAhead+NewElements;    // thus successive sublists get bigger
        if (baseSublist->max_elements > NewSublistElements)  NewSublistElements = baseSublist->max_elements;

#ifndef LL_FastFlag
        newLL_result = NewLinkedList(&dummy, NewSublistElements, theList->elementsize, theList->spare_room, if_cleared);
#else
        newLL_result = FNewLinkedList(&dummy, NewSublistElements, theList->elementsize, theList->spare_room, if_cleared);
#endif
        if (newLL_result != passed) return newLL_result;

        if (NewSublistElements > SkipAhead+NewElements)         // here we really can ignore return value -- never have 0 els
            FDeleteElements(&dummy, SkipAhead+NewElements+1, dummy.elementnum);


            // stitch the new list in

        dummy.memory->next_sublist = baseSublist->next_sublist;
        baseSublist->next_sublist = dummy.memory;
        theList->elementnum += NewElements+SkipAhead;         // fool the computer for now


            // First make room for new arrivals..

#ifndef LL_FastFlag
        CopyElements(theList, InsertionPoint, theList, InsertionPoint+NewElements, ElementsOnTop);
        if (if_cleared == yz_true) ClearElements(theList, InsertionPoint, InsertionPoint+NewElements-1);
#else
        FCopyElements(theList, InsertionPoint, theList, InsertionPoint+NewElements, ElementsOnTop);
        if (if_cleared == yz_true) FClearElements(theList, InsertionPoint, InsertionPoint+NewElements-1);
#endif


            // Now push everything off the top of the lower list, and reset the number of variables it holds.
            // CopyPoint hasn't changed from last CopyEls, since it is tied to a position in the LLs, not a particular element.

#ifndef LL_FastFlag
        CopyElements(theList, CopyPoint+bottom-1, theList, CopyPoint+SkipAhead+bottom-1, SkipAhead+NewElements);
#else
        FCopyElements(theList, CopyPoint+bottom-1, theList, CopyPoint+SkipAhead+bottom-1, SkipAhead+NewElements);
#endif
        baseSublist->num_elements -= SkipAhead;
        theList->elementnum -= SkipAhead;              // unfool the computer
    }
    else           // if we don't need to allocate another list -- just increase num_elements closer to max_elements
    {
        if (LONG_MAX-theList->elementnum < NewElements) return mem_err;
        baseSublist->num_elements += NewElements;
        theList->elementnum += NewElements;
        if (bottom + baseSublist->num_elements > InsertionPoint)
#ifndef LL_FastFlag
            CopyElements(theList, InsertionPoint, theList, InsertionPoint + NewElements, ElementsOnTop);
#else
            FCopyElements(theList, InsertionPoint, theList, InsertionPoint + NewElements, ElementsOnTop);
#endif
    }

                // subtle point -- if InsertionPoint+NewElements-1 == 0, ClearElements returns param_err, which we throw away.  Fine.
#ifndef LL_FastFlag
    if (if_cleared == yz_true) ClearElements(theList, InsertionPoint, InsertionPoint + NewElements - 1);
#else
    if (if_cleared == yz_true) FClearElements(theList, InsertionPoint, InsertionPoint + NewElements - 1);
#endif

    return passed;
}


// [F]AddElements():  a quick way to insert elements at the end of a linked list -- to stretch the list, in other words

#ifndef LL_FastFlag
LL_retrn AddElements(linkedlist * theList, unsigned long NewElements, Boolean if_cleared)
#else
LL_retrn FAddElements(linkedlist * theList, unsigned long NewElements, Boolean if_cleared)
#endif
{
#ifndef LL_FastFlag
    return InsertElements(theList, theList->elementnum+1, NewElements, if_cleared);
#else
    return FInsertElements(theList, theList->elementnum+1, NewElements, if_cleared);
#endif
}


// [F]DeleteElements() removes the memory space associated with certain elements, and pulls the others down
// (so their positions change in the list).  Empty sublists are removed to free up space.

#ifndef LL_FastFlag
LL_retrn DeleteElements(linkedlist * theList, unsigned long FirstDelete, unsigned long LastDelete)
#else
LL_retrn FDeleteElements(linkedlist * theList, unsigned long FirstDelete, unsigned long LastDelete)
#endif
{
    sublistHeader *tempSublist, *currentSublist, **parentPtr, *tempPtr;
    unsigned long bottom, top, firstCopy, copyBottom;

#ifndef LL_FastFlag
    if (theList->memory == NULL) return init_err;
    if ((ElementExists(theList, FirstDelete) != yz_true) || (ElementExists(theList, LastDelete) != yz_true)) return range_err;
#endif
    if (FirstDelete > LastDelete) return passed;
    currentSublist = LL_FindSubList(theList, FirstDelete, &bottom);


        // First, find the sublist in which we stop deleting, then copy what comes after back down towards the bottom.

    while (bottom <= LastDelete)
    {
        top = bottom + currentSublist->num_elements - 1;
        copyBottom = bottom;
        if (bottom < FirstDelete) copyBottom = FirstDelete;      // i.e. if it's the first pass

        firstCopy = top + 1;
        if (LastDelete < top)
        {
            firstCopy = LastDelete + 1;
            memmove((void *) (((char *) currentSublist) + (copyBottom-bottom)*theList->elementsize + sublistHeaderSize),
                    (void *) (((char *) currentSublist) + (firstCopy-bottom)*theList->elementsize + sublistHeaderSize),
                    (top + 1 - firstCopy)*theList->elementsize);
        }

        bottom += currentSublist->num_elements;         // bottom goes by the old indexing; before the elements were deleted
        currentSublist->num_elements -= (firstCopy-copyBottom);
        currentSublist = currentSublist->next_sublist;
    }


        // Now get rid of null lists.

    parentPtr = &(theList->memory);
    while (*parentPtr != NULL)
    {
        tempSublist = *parentPtr;
        if (tempSublist->num_elements == 0)
        {
            tempPtr = tempSublist->next_sublist;
            free((void *) *parentPtr);
            *parentPtr = tempPtr;
        }
        else parentPtr = &((*parentPtr)->next_sublist);
    }


        // If we went too far and deleted the first sublist, recreate it

    if (theList->memory == NULL)
    {
        tempPtr = LL_NewSubList(0, 1, theList->elementsize, uncleared);
        theList->memory = tempPtr;
        if (tempPtr == NULL)  return mem_err;
    }

    theList->elementnum -= LastDelete - FirstDelete + 1;

    return passed;
}


// [F]DeleteElement() is just the singular form of [F]DeleteElements() -- it removes a single element.

#ifndef LL_FastFlag
LL_retrn DeleteElement(linkedlist * theList, unsigned long DeletePoint)
#else
LL_retrn FDeleteElement(linkedlist * theList, unsigned long DeletePoint)
#endif
{
#ifndef LL_FastFlag
    return DeleteElements(theList, DeletePoint, DeletePoint);
#else
    return FDeleteElements(theList, DeletePoint, DeletePoint);
#endif
}



// [F]DefragmentLinkedList() reduces a linked list to a single list, with NO spare room (modified 4/3/04), regardless of the
// spare_room parameter.  Call this (for speed) when the number of elements is fixed, or to get all of the list elements in one block.

#ifndef LL_FastFlag
LL_retrn DefragmentLinkedList(linkedlist * theList)
#else
LL_retrn FDefragmentLinkedList(linkedlist * theList)
#endif
{
    linkedlist oldList;
    LL_retrn DefLL_retrn;

    if (theList->elementnum == theList->memory->num_elements)  return passed;
    oldList = *theList;

#ifndef LL_FastFlag
    if (oldList.memory == NULL) return init_err;
    DefLL_retrn = NewLinkedList(theList, oldList.elementnum, oldList.elementsize, 0, uncleared);
#else
    DefLL_retrn = FNewLinkedList(theList, oldList.elementnum, oldList.elementsize, 0, uncleared);
#endif
    if (DefLL_retrn != passed)
    {
        *theList = oldList;
        return DefLL_retrn;
    }

    theList->spare_room = oldList.spare_room;

#ifndef LL_FastFlag
    if (theList->elementnum > 0)  {
        DefLL_retrn = CopyElements(&oldList, 1, theList, 1, oldList.elementnum);
        if (DefLL_retrn != passed)  return DefLL_retrn;         }

    DefLL_retrn = DeleteLinkedList(&oldList);
    if (DefLL_retrn != passed)  return DefLL_retrn;

    return passed;
#else
    if (theList->elementnum > 0)  {
        FCopyElements(&oldList, 1, theList, 1, oldList.elementnum);         }

    FDeleteLinkedList(&oldList);
    return passed;
#endif
}


// [F]CopyElements() copies a string of elements from one position in one linked list to another position in another linked list.
// Lists must have identically sized elements, of course.
// This routine is long-winded because it is designed to be fast, not simple.

#ifndef LL_FastFlag
LL_retrn CopyElements(linkedlist * SourceList, unsigned long SourceBegin, linkedlist * DestList, unsigned long DestBegin, unsigned long NumberToCopy)
#else
void FCopyElements(linkedlist * SourceList, unsigned long SourceBegin, linkedlist * DestList, unsigned long DestBegin, unsigned long NumberToCopy)
#endif
{
    unsigned long FirstCopied, LastCopied;
    unsigned long SourceBottom, DestBottom;
    unsigned long ElementSize;
    sublistHeader *CurrentSourceSublist;
    sublistHeader *CurrentDestSublist;


        // First, make sure the user is within range.

#ifndef LL_FastFlag
    if ((SourceList->memory == NULL) || (DestList->memory == NULL))  return init_err;
    if (SourceList->elementsize != DestList->elementsize)  return list_mismatch_err;
    if (NumberToCopy == 0) return passed;            // put before el-exists-check, since if NTC == 0 we often can't pass a valid element
    if ((ElementExists(SourceList, SourceBegin) != yz_true) || (ElementExists(DestList, DestBegin) != yz_true)) return range_err;
    if ((ElementExists(SourceList, SourceBegin + NumberToCopy - 1) != yz_true) || (ElementExists(DestList, DestBegin+NumberToCopy-1) != yz_true))
        return range_err;
#else
    if (NumberToCopy == 0) return;
#endif

    ElementSize = SourceList->elementsize;


        // Now start copying the lists.  If we can do it from lowest to highest without overwriting, that's easiest (first clause of 'if' statement).
        // Either way, we iteratively hop down sublists: updating them, restricting the size of the copied memory block, and copying the memory.

    if ((SourceBegin > DestBegin) || (SourceBegin+NumberToCopy <= DestBegin) || (SourceList != DestList))
    {
        CurrentSourceSublist = LL_FindSubList(SourceList, SourceBegin, &SourceBottom);
        CurrentDestSublist = LL_FindSubList(DestList, DestBegin, &DestBottom);
        FirstCopied = 1;
        LastCopied = NumberToCopy - 1;
        while (LastCopied < NumberToCopy)
        {
            if (CurrentSourceSublist->num_elements+SourceBottom-1 < SourceBegin+FirstCopied-1)     //update sublists
            {
                SourceBottom = SourceBottom + CurrentSourceSublist->num_elements;
                CurrentSourceSublist = SublistHeader(CurrentSourceSublist)->next_sublist;
            }
            if (CurrentDestSublist->num_elements+DestBottom-1 < DestBegin+FirstCopied-1)
            {
                DestBottom = DestBottom + CurrentDestSublist->num_elements;
                CurrentDestSublist = SublistHeader(CurrentDestSublist)->next_sublist;
            }

            LastCopied = NumberToCopy;
            if (SourceBottom+CurrentSourceSublist->num_elements < SourceBegin+LastCopied)      // restrict copy length
                LastCopied = SourceBottom + CurrentSourceSublist->num_elements - SourceBegin;
            if (DestBottom+CurrentDestSublist->num_elements < DestBegin+LastCopied)
                LastCopied = DestBottom + CurrentDestSublist->num_elements - DestBegin;

            memmove((void *) (((char *) CurrentDestSublist) + (FirstCopied+DestBegin-DestBottom-1)*ElementSize + sublistHeaderSize),
                    (void *) (((char *) CurrentSourceSublist) + (FirstCopied+SourceBegin-SourceBottom-1)*ElementSize + sublistHeaderSize),
                    ElementSize*(LastCopied-FirstCopied+1));

            FirstCopied = LastCopied+1;
        }
    }
    else if (SourceBegin < DestBegin)       // No, copy from the top down; otherwise we will overwrite something.
    {
        CurrentSourceSublist = LL_FindSubList(SourceList, SourceBegin+NumberToCopy-1, &SourceBottom);
        CurrentDestSublist = LL_FindSubList(DestList, DestBegin+NumberToCopy-1, &DestBottom);
        FirstCopied = NumberToCopy+1;
        LastCopied = NumberToCopy;

        while(FirstCopied > 1)
        {
            FirstCopied = 1;          // how much to copy?
            if (SourceBottom > FirstCopied+SourceBegin-1)
                FirstCopied = SourceBottom - SourceBegin + 1;
            if (DestBottom > FirstCopied+DestBegin-1)
                FirstCopied = DestBottom - DestBegin + 1;

                        // here we copy first, then scale the lists
            memmove((void *) ((char *) CurrentDestSublist + (FirstCopied+DestBegin-DestBottom-1)*ElementSize + sublistHeaderSize),
                    (void *) ((char *) CurrentSourceSublist + (FirstCopied+SourceBegin-SourceBottom-1)*ElementSize + sublistHeaderSize),
                    ElementSize*(LastCopied-FirstCopied+1));

            LastCopied = FirstCopied-1;
            if (FirstCopied == SourceBottom-SourceBegin+1)
                CurrentSourceSublist = LL_FindSubList(SourceList, SourceBegin+LastCopied-1, &SourceBottom);
            if (FirstCopied == DestBottom-DestBegin+1)
                CurrentDestSublist = LL_FindSubList(DestList, DestBegin+LastCopied-1, &DestBottom);
        }
    }

#ifndef LL_FastFlag
    return passed;
#else
    return;
#endif
}


// [F]CompareElements() compares the two subsets of linked lists (of the same length) byte-by-byte to test for equality.

#ifndef LL_FastFlag
LL_retrn CompareElements(linkedlist * SourceList, unsigned long SourceBegin, linkedlist * DestList, unsigned long DestBegin,
                                    unsigned long NumberToCompare)
#else
LL_retrn FCompareElements(linkedlist * SourceList, unsigned long SourceBegin, linkedlist * DestList, unsigned long DestBegin,
                                    unsigned long NumberToCompare)
#endif
{
    unsigned long FirstCompared, LastCompared, SourceBottom, DestBottom, ElementSize;
    sublistHeader *CurrentSourceSublist, *CurrentDestSublist;
    unsigned long retrn;


        // make sure params were passed correctly

#ifndef LL_FastFlag
    if ((SourceList->memory == NULL) || (DestList->memory == NULL))  return init_err;
    if (SourceList->elementsize != DestList->elementsize)  return list_mismatch_err;
#endif
    if (NumberToCompare == 0)  return equal;
#ifndef LL_FastFlag
    if ((ElementExists(SourceList, SourceBegin) != yz_true) || (ElementExists(DestList, DestBegin) != yz_true))
        return range_err;
    if ((ElementExists(SourceList, SourceBegin+NumberToCompare-1) != yz_true) || (ElementExists(DestList, DestBegin+NumberToCompare-1) != yz_true))
        return range_err;
#endif

    ElementSize = SourceList->elementsize;

    CurrentSourceSublist = LL_FindSubList(SourceList, SourceBegin, &SourceBottom);
    CurrentDestSublist = LL_FindSubList(DestList, DestBegin, &DestBottom);

    FirstCompared = 1;
    LastCompared = NumberToCompare-1;


        // Now, by chunks (size of each sublist), do straight comparisons of memory.

    while (LastCompared < NumberToCompare)
    {
        if (CurrentSourceSublist->num_elements + SourceBottom < SourceBegin+FirstCompared)
        {
            SourceBottom += CurrentSourceSublist->num_elements;
            CurrentSourceSublist = SublistHeader(CurrentSourceSublist)->next_sublist;
        }
        if (CurrentDestSublist->num_elements + DestBottom < DestBegin+FirstCompared)
        {
            DestBottom += CurrentDestSublist->num_elements;
            CurrentDestSublist = SublistHeader(CurrentDestSublist)->next_sublist;
        }

        LastCompared = NumberToCompare;
        if (SourceBottom+CurrentSourceSublist->num_elements < SourceBegin+LastCompared)
            LastCompared = SourceBottom + CurrentSourceSublist->num_elements - SourceBegin;
        if (DestBottom+CurrentDestSublist->num_elements < DestBegin+LastCompared)
            LastCompared = DestBottom + CurrentDestSublist->num_elements - DestBegin;

        retrn = memcmp((void *) ((char *) CurrentSourceSublist + (FirstCompared+SourceBegin-SourceBottom-1)*ElementSize + sublistHeaderSize),
                        (void *) ((char *) CurrentDestSublist + (FirstCompared+DestBegin-DestBottom-1)*ElementSize + sublistHeaderSize),
                        (size_t) (LastCompared-FirstCompared+1)*ElementSize);
        if (retrn != 0) return notequal;

        FirstCompared = LastCompared+1;
    }

    return equal;
}


// [F]FillElements() fills elements of a linked list with a byte pattern (FF FF or 4E 4E, etc.).
// Useful for clearing lists -- set pattern to 0.

#ifndef LL_FastFlag
LL_retrn FillElements(linkedlist * theList, unsigned long Start, unsigned long End, char Value)
#else
void FFillElements(linkedlist * theList, unsigned long Start, unsigned long End, char Value)
#endif
{
    unsigned long FirstCleared, LastCleared, bottom, top;
    sublistHeader * CurrentSublist;

#ifndef LL_FastFlag
    if (theList->memory == NULL)  return init_err;
    if ((ElementExists(theList, Start) != yz_true) || (ElementExists(theList, End) != yz_true))
        return range_err;
#endif

    FirstCleared = Start;
    LastCleared = Start-1;
    CurrentSublist = LL_FindSubList(theList, Start, &bottom);

    while (LastCleared < End)
    {
        top = bottom + CurrentSublist->num_elements - 1;
        LastCleared = End;
        if (top < End) LastCleared = top;

        memset((void *) ((char *) CurrentSublist + (FirstCleared-bottom)*theList->elementsize + sublistHeaderSize), Value,
                (size_t) theList->elementsize*(LastCleared - FirstCleared + 1));

        bottom = top+1;
        FirstCleared = LastCleared+1;
        CurrentSublist = SublistHeader(CurrentSublist)->next_sublist;
    }

#ifndef LL_FastFlag
    return passed;
#else
    return;
#endif
}


// [F]ClearElements() calls [F]FillElements(.., .., .., 0) -- just for convenience.

#ifndef LL_FastFlag
LL_retrn ClearElements(linkedlist * theList, unsigned long Start, unsigned long End)
#else
void FClearElements(linkedlist * theList, unsigned long Start, unsigned long End)
#endif
{
#ifndef LL_FastFlag
    return FillElements(theList, Start, End, 0);
#else
    FFillElements(theList, Start, End, 0);
    return;
#endif
}


// [F]SetElements() -- reads data from a buffer into a linked list.

#ifndef LL_FastFlag
LL_retrn SetElements(linkedlist * theList, unsigned long StartElement, unsigned long EndElement, void *ReadAddress)
#else
void FSetElements(linkedlist * theList, unsigned long StartElement, unsigned long EndElement, void *ReadAddress)
#endif
{
    unsigned long FirstSet, LastSet;
    unsigned long bottom, top;
    sublistHeader * currentSublist;

    FirstSet = StartElement;
    LastSet = 0;

#ifndef LL_FastFlag
    if (theList->memory == NULL)  return init_err;
    if (EndElement < StartElement)  return passed;
    if ((ElementExists(theList, StartElement) != yz_true) || (ElementExists(theList, EndElement) != yz_true))
        return range_err;
#else
    if (EndElement < StartElement)  return;
#endif

    currentSublist = LL_FindSubList(theList, StartElement, &bottom);

    while (LastSet < EndElement)
    {
        top = bottom + currentSublist->num_elements - 1;
        LastSet = EndElement;
        if (top < EndElement) LastSet = top;

        memmove((void *) ((char *) currentSublist + (FirstSet-bottom)*theList->elementsize + sublistHeaderSize),
                (void *) ((char *) ReadAddress + (FirstSet-StartElement)*theList->elementsize),
                theList->elementsize*(LastSet-FirstSet+1));

        bottom = top + 1;
        FirstSet = LastSet + 1;
        currentSublist = SublistHeader(currentSublist)->next_sublist;
    }

#ifndef LL_FastFlag
    return passed;
#else
    return;
#endif
}


// [F]SetElement() -- sets a single element.

#ifndef LL_FastFlag
LL_retrn SetElement(linkedlist* theList, unsigned long theElement, void *ReadAddress)
#else
void FSetElement(linkedlist* theList, unsigned long theElement, void *ReadAddress)
#endif
{
#ifndef LL_FastFlag
    return SetElements(theList, theElement, theElement, ReadAddress);
#else
    FSetElements(theList, theElement, theElement, ReadAddress);
    return;
#endif
}


// [F]GetElements():  writes data from a linked list into a buffer

#ifndef LL_FastFlag
LL_retrn GetElements(linkedlist * theList, unsigned long StartElement, unsigned long EndElement, void *WriteAddress)
#else
void FGetElements(linkedlist * theList, unsigned long StartElement, unsigned long EndElement, void *WriteAddress)
#endif
{
    unsigned long FirstGet, LastGet;
    unsigned long bottom, top;
    sublistHeader * currentSublist;

    FirstGet = StartElement;
    LastGet = 0;

#ifndef LL_FastFlag
    if (theList->memory == NULL)  return init_err;
    if (EndElement < StartElement)  return passed;
    if ((ElementExists(theList, StartElement) != yz_true) || (ElementExists(theList, EndElement) != yz_true))
        return range_err;
#else
    if (EndElement < StartElement) return;
#endif

    currentSublist = LL_FindSubList(theList, StartElement, &bottom);

    while (LastGet < EndElement)
    {
        top = bottom + currentSublist->num_elements - 1;
        LastGet = EndElement;
        if (top < EndElement) LastGet = top;

        memmove((void *) ((char *) WriteAddress + (FirstGet-StartElement)*theList->elementsize),
                (void *) ((char *) currentSublist + (FirstGet-bottom)*theList->elementsize + sublistHeaderSize),
                    theList->elementsize*(LastGet - FirstGet + 1));

        bottom = top + 1;
        FirstGet = LastGet + 1;
        currentSublist = SublistHeader(currentSublist)->next_sublist;
    }

#ifndef LL_FastFlag
    return passed;
#else
    return;
#endif
}


// [F]GetElement() just copies one single element into a buffer

#ifndef LL_FastFlag
LL_retrn GetElement(linkedlist * theList, unsigned long theElement, void *WriteAddress)
#else
void FGetElement(linkedlist * theList, unsigned long theElement, void *WriteAddress)
#endif
{
#ifndef LL_FastFlag
    return GetElements(theList, theElement, theElement, WriteAddress);
#else
    FGetElements(theList, theElement, theElement, WriteAddress);
    return;
#endif
}


// [F]ElementExists() tests to make sure the list exists, and then that the element is within range

#ifndef LL_FastFlag
LL_retrn ElementExists(linkedlist * theList, unsigned long TheElement)
#else
LL_retrn FElementExists(linkedlist * theList, unsigned long TheElement)
#endif
{
#ifndef LL_FastFlag
    if (theList->memory == NULL)  return init_err;
#endif
    if ((TheElement > theList->elementnum) || (TheElement == 0)) return yz_false;

    return yz_true;
}


// [F]Element() returns a pointer to the memory location of TheElement (or 0 if it is nonexistent)

#ifndef LL_FastFlag
void *Element(linkedlist * theList, unsigned long TheElement)
#else
void *FElement(linkedlist * theList, unsigned long TheElement)
#endif
{
    sublistHeader * HeaderOfSublist;
    unsigned long bottom;

#ifndef LL_FastFlag
    if (ElementExists(theList, TheElement) != yz_true)  return NULL;
#endif
    HeaderOfSublist = LL_FindSubList(theList, TheElement, &bottom);

    return (void *) ((char *) HeaderOfSublist + sublistHeaderSize + theList->elementsize*(TheElement-bottom));
}


// FSkipElements() (no slow-mode version) skips ahead from a starting element/sublist by N elements.
// The purpose is for speed, so we don't need to call FElement() repeatedly if we are accessing consecutive elements repeatedly.
// Assumes the indices are within range (no range checking).
// LocalIndex is measured relative to the bottom of the sublist, where 0 is the first element.

#ifndef LL_FastFlag
void *FSkipElements(linkedlist *theList, sublistHeader **theSublist, unsigned long *LocalIndex, unsigned long NumberToSkip)
{
    *LocalIndex += NumberToSkip;
    while (*LocalIndex >= (*theSublist)->num_elements)  {
        *LocalIndex -= (*theSublist)->num_elements;
        *theSublist = (*theSublist)->next_sublist;      }

    return (void *) ((char *) (*theSublist) + sublistHeaderSize + theList->elementsize*(*LocalIndex));
}
#endif
