/*
 *  yazooh.h -- misc. Yazoo-wide definitions
 *
 *  Yazoo
 *  Copyright (C) 2005 Brian C. Ross
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of the License, or any later version.

 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#ifndef Yazoo_h
#define Yazoo_h


// Generic definitions

typedef unsigned long Ulong;
typedef signed long Slong;
typedef unsigned char Boolean;


// Libraries used by all Yazoo .c(pp) files

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "lnklst.h"


// Generic 'define's

#define yz_true 1
#define yz_false 0

#define passed 0



// Variable data types

#define datatype char
#define Ubyte_type 0    /* The numeric types must come before block, etc., since Yazoo  */
#define Sbyte_type 1    /* checks to see if a type is numeric by checking if  */
#define Ushort_type 2   /* type <= double_type.  */
#define Sshort_type 3
#define Ulong_type 4
#define Slong_type 5
#define single_type 6
#define double_type 7
#define block_type 8
#define string_type 9
#define composite_type 10
#define proxy_type 11
#define no_type 12


// Shortcuts for reading/writing variously-typed linked lists

#define LL_UByte(a,b) ((unsigned char *) FElement(a,b))
#define LL_UShort(a,b) ((unsigned short *) FElement(a,b))
#define LL_ULong(a,b) ((unsigned long *) FElement(a,b))
#define LL_SByte(a,b) ((signed char *) FElement(a,b))
#define LL_SShort(a,b) ((signed short *) FElement(a,b))
#define LL_SLong(a,b) ((signed long *) FElement(a,b))
#define LL_Double(a,b) ((double *) FElement(a,b))
#define LL_Ptr(a,b) ((void **) FElement(a,b))

#define BundleMember(a,b) (((member *) FElement(&(a->variable_ptr->mem.members), b))->member_ptr)
#define VariableMember(a,b) (((member *) FElement(&(a->mem.members), b))->member_ptr)



// The bytecode instruction words (all 32 bit)
// f denotes floating point (double) arithmetic; i denotes integer

#define null_cmd 0          /* do nothing; marks ends of {}  */
#define func_return 1       /* return from the function  */
#define internal_var 2      /* for example, number of vars, err_code, ...  */
#define built_in_function 3 /* call a built-in function (besides the ones listed here)  */
#define user_function 5     /* (()) call the code of the current variable  */
#define get_args 6          /* call the arguments from the passing routine  */
#define jump_always 7       /* relative addressing  */
#define jump_if_true 8      /* jump if test is false  */
#define jump_if_false 9     /* jump if test is true  */
#define exit_Yazoo 10

#define def_general 11      /* set the code (variable & member)  */
#define forced_equate 12    /* (=!) copy data; only restriction is that data sizes must be the same  */
#define code_number 13      /* (#) return N'th code block of left arg, where N is given by right arg  */
#define sub_code 14         /* (<<) match var (left arg) with code (right arg)  */
#define append_code 15      /* (:) match left var with right arg code  */
#define remove_member 16    /* (remove) remove member  */

#define search_member 17    /* search backwards for var with specified ID  */
#define step_to_memberID 18 /* (.) step to a var with specified ID  */
#define step_to_index 19    /* ([]) step to Nth var  */
#define step_to_indices 20  /* return a range of variables  */
#define step_to_all 21      /* [*] */
#define resize 22           /* ([^#] -- sets a new top */
#define insert_indices_up 23    /* same as step_to_indices, but create new entries, affixed to the member being pushed up  */
#define insert_indices_down 24  /* same as step_to_indices, but create new entries, attached to the member just below  */
#define delete_indices 25   /* deletes a range of indices, but does not remove the associated member  */
#define this_var 27         /* return the variable that contains the code being executed  */
#define that_var 28         /* return the variable to the left of an equate sign  */
#define no_var 29           /* returns nothing -- no bundle, stem, stem member, etc.  */

#define if_eq 30         /* (==)  */
#define if_ne 31         /* (/=)  */
#define if_gt 32         /* (>)  */
#define if_ge 33         /* (>=)  */
#define if_lt 34         /* (<)  */
#define if_le 35         /* (<=)  */
#define if_and 36        /* (and)  */
#define if_or 37         /* (or)  */
#define if_xor 38        /* (xor)  */
#define if_not 39        /* (not)  */
#define if_eq_at 40      /* (== @)  */
#define if_ne_at 41      /* (/= @)  */

#define addf 42         /* (+)  -- P.S.  all numeric commands must be in a row (for DefGeneral())  */
#define subf 43         /* (-)  */
#define mulf 44         /* (*)  */
#define divf 45         /* (/)  */
#define powerf 46       /* (^)  */
#define modi 47         /* (mod) */

#define paren_command 48
#define not_a_command 49

#define constant_Slong 50   /* a constant integer entered in the code */
#define constant_float 51   /* a const. float entered in the code */
#define constant_string 52  /* comes from "" -- basically a block */
#define code_block 53       /* a {} */

#define ubyte_cmd 54
#define sbyte_cmd 55
#define ushort_cmd 56
#define sshort_cmd 57
#define ulong_cmd 58
#define slong_cmd 59
#define float_cmd 60
#define double_cmd 61
#define block_cmd 62
#define string_cmd 63

#define code_marker 65

#define commands_num 66
#define written_commands_num 66
#define control_commands_num 15
#define OoOLevels 17

#define bi_commands_num 44
#define internal_vars_num 10



// Argument types (IDs of jump table) of the left/right arguments to the various operators

#define start_arg 0   /* start_arg is anything that can be used as the first function in an expression */
#define var_arg 1
#define log_arg 2
#define data_arg 3
#define code_arg 4
#define type_arg 5
#define no_arg 6


// Flags used after the def_general command

#define SetFlag(a,b,c) (a = (a&(~(1<<b)))+(c<<b))
#define GetFlag(a,b) ((a>>b)&1)

#define make_proxy 8
#define unjammable 7
#define hidden_member 6
#define run_constructor 5
#define relink_target 4
#define new_target 3
#define can_add_members 2
#define update_members 1
#define post_equate 0


// Prototypes from consts.c(pp)

extern const Ulong PrimitiveSizes[10];
extern const double MaxInteger[6];
extern const double MinInteger[6];
extern const Slong CommandIDs[written_commands_num];
extern const char* CommandNames[written_commands_num];
extern const char OpDirection[OoOLevels];
extern const char* BuiltInCommandNames[bi_commands_num];
extern const char* InternalVarNames[internal_vars_num];
extern char left_args[commands_num];
extern char right_args[commands_num];
extern const char OoOlevel[commands_num];
extern const char* ControlCommandNames[control_commands_num];
extern Boolean BIF_CanStartSentence[bi_commands_num];
extern void(*AllowedCMDJumpTable[6][commands_num])(Slong);
//extern const char *empty_str;
extern const char *ZeroWrittenCode;

#endif
