/*
 *  yzmain.c(pp) - Yazoo main program
 *
 *  Yazoo
 *  Copyright (C) 2005 Brian C. Ross
 *
 *  This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2 of the License, or any later version.

 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 *  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "yzmain.h"
#include <stdio.h>
#include <stdlib.h>
#include "yazooh.h"
#include "lnklst.h"
#include "cmpile.h"
#include "intrpt.h"
#include "hobbsh.h"
#include "extrnl.h"
#include "userfn.h"
#include <time.h>
#include <math.h>
#include <limits.h>


// YazooTest_ON should always be undefined

#ifdef YazooTest_ON
#include "Test/Test.c"
#endif


// Below:  the lists of error messages for cases where the main script (start.zoo or the argument to yazoo) blows up.
// The provided start.zoo file has its own copy of these lists, since it manually flags errors that occur under its watch.

const char *CompileErrorStrings[] = {
    "", "out of memory", "unexpected end of script", "end of line expected", "end of script expected",
        "unknown command", "invalid command", "argument expected", "no left-hand argument allowed", "no right-hand argument allowed",
    "variable or number expected", "variable expected", "']' expected", "'}' expected", "')' expected",
        "overflow", "underflow", "not a number", "not a string", "missing arguments",
    "must follow if", "conditional expected", "\"end if\" expected", "\"end for\" expected", "\"end while\" expected",
        "\"until\" expected", "\"in\" expected", "'[' expected", "',' expected",
    "illegal control command", "unexpected control terminator", "not a statement", "??? -- unknown error"  };
    
const char *HobbishErrorStrings[] = {
    "", "out of memory", "illegal command", "code overflow", "inaccessible code", "jump to middle of sentence",
        "number out of range", "division by zero",
    "member not defined", "variable has no member", "invalid index", "member is void", "proxy is void", "cannot step to multiple members",
        "cannot step to multiple proxies", "unknown error",
    "type mismatch", "argument out of range", "self reference", "illegal target", "no member associated with variable",
        "target was deleted", "incomplete member", "incomplete variable", "invalid block size",
    "not numeric", "not a variable", "unequal data sizes", "error reading number", "no target variable",
        "mismatched indices", "variable is 'jammed' (aliased elsewhere)", "unjammed member", "unjammed proxy",
    "not composite", "undefined string", "multiple indices not allowed", "wrong number of arguments", "string expected",
        "invalid index", "error in argument", "nonexistent register", "nonexistent Yazoo function", "can't find external function",
    "recursion depth too high", "I/O error",
        "return flag (? -- only used internally)", "finished signal (? -- only used internally)"  };

const char *LL_ErrorStrings[] = {
    "unspecified internal error", "out of memory", "bad parameter", "range error", "initialization error", "list_mismatch_error"  };



// yz_main() is the outermost function.  It initializes Yazoo's memory,
// loads and runs the required script (start.zoo or start.hob if no script is provided), flags any error, and quits.-

// main() is used if Yazoo is to be compiled as a stand-alone program (i.e. not as an API).


#ifdef YazooMainProgram
int main(int argc, char **argv)
{
    return run_yazoo(argc-1, argv+1);         // we ignore the first argument
}
#endif


int run_yazoo(int argc, char **argv)
{
    int retrn;
    
    yz_compile_global_struct hold_compile_globals = yz_compile_globals;
    yz_interpret_global_struct hold_interpret_globals = yz_interpret_globals;
    yz_bytecode_global_struct hold_bytecode_globals = yz_bytecode_globals;
    yz_external_global_struct hold_external_globals = yz_external_globals;
    
    retrn = yz_main(argc, argv);
    
    yz_compile_globals = hold_compile_globals;
    yz_interpret_globals = hold_interpret_globals;
    yz_bytecode_globals = hold_bytecode_globals;
    yz_external_globals = hold_external_globals;
    
    return retrn;
}
    
    

int yz_main(int argc, char **argv)
{
    linkedlist CompiledCode, SourceCode, *FileReadLL, ZeroCodeLL, NewlinePosCopy;
    FILE *MyFile;
    char OneChar, *scanPtr, *ZeroWrittenCodeCopy = (char *) ZeroWrittenCode;
    unsigned char retrn;
    int main_error = 0;
    LL_retrn linkedLretrn;
    Ulong counter, CharCounter, ScriptCodeIndex;
    Slong *ScriptCodeEntryPtr;
    code_ref *ScriptCR;
    Boolean if_already_compiled = yz_false;
    
#ifdef YazooTest_ON
    TestYazoo();
#endif
    
        // Initialize the random number generator.
        
    srand((Ulong) time(0));
    rand();  rand();  rand();  rand();      // otherwise the first couple of random numbers are suspiciously consistent
    
    
        // Init the temp. names for def_general
    
    TempDefGenName = LONG_MAX;
    

        // Set up a buffer for loading the script
    
    linkedLretrn = NewLinkedList(&SourceCode, 0, sizeof(char), 0, uncleared);
    if (linkedLretrn != passed)  {  printf("Yazoo:  Memory Error:  %s\n", CompileErrorStrings[memory_err]);  return 1;  }
    
    
        // Open the file specified in the first argument, if given; otherwise, open start.zoo or start.hob
    
    if (argc > 1)  {  printf("usage: yazoo [source_file]\n");  return 1;  }
    if (argc == 0)  {
        MyFile = fopen("./start.zoo", "r");
        if (MyFile == NULL)  {  if_already_compiled = yz_true;  MyFile = fopen("./start.hob", "rb");  }
        if (MyFile == NULL)  {  printf("Error:  Could not open start.hob or start.zoo\n");  return 1;  }   }
    else  {
        MyFile = fopen(argv[0], "rb");
        if (MyFile == NULL)  {  printf("Error:  Could not open %s\n", argv[0]);  return 1;  }
        counter = 0;  while (*(argv[0]+counter) != 0)  counter++;
        if (counter > 4)
            if ((*(char *) (argv[0]+counter-4) == '.') && (*(char *) (argv[0]+counter-3) == 'h') &&
                (*(char *) (argv[0]+counter-2) == 'o') && (*(char *) (argv[0]+counter-1) == 'b'))  if_already_compiled = yz_true;  }


        // Load the script into memory
    
    CharCounter = 1;
    FileReadLL = &SourceCode;
    
    OneChar = ' ';
    while (1 == 1)  {
        if (CharCounter > FileReadLL->elementnum)  {
            linkedLretrn = FAddElements(&SourceCode, FileReadBufferSize, uncleared);
            if (linkedLretrn != passed)  {
                printf("Yazoo:  Memory Error:  %s\n", CompileErrorStrings[memory_err]);  return 1;  }  }

        if (feof(MyFile) == 0)      {
            OneChar = fgetc(MyFile);
            *LL_UByte(FileReadLL, CharCounter) = OneChar;       }
        else    {
            *LL_UByte(FileReadLL, CharCounter) = 0;
            break;      }
        CharCounter++;         }
    
    fclose(MyFile);
    linkedLretrn = DefragmentLinkedList(FileReadLL);
    if (linkedLretrn != passed)  {  printf("Yazoo:  Memory Error:  %s\n", CompileErrorStrings[memory_err]);  return 1;  }
    
    
        // Load and compile the code that initializes Zero, the top-level variable.
    
    retrn = InitCompile(0);
    if (retrn == passed)  retrn = Compile((char **) &ZeroWrittenCodeCopy, &ZeroCodeLL);
    if (retrn != passed)  {  printf("Yazoo:  Compile (Zero) returned \"%s\" at character %i\n",
                            (char *) CompileErrorStrings[retrn],  (int) (errPtr-ZeroWrittenCode+1));  return 1;  }


        // Initialize all the global variables used by Yazoo.
    
    retrn = InitYazoo(LL_SLong(&ZeroCodeLL, 1), ZeroCodeLL.elementnum);     // Compile() calls DefragmentLL()
    if (retrn != passed)  {  printf("Yazoo:  Init returned \"%s\"\n", (char *) HobbishErrorStrings[retrn]);  return 1;  }


        // Sanity check -- make sure Zero compiled correctly.
    
    StartCodePtr = LL_SLong(&ZeroCodeLL, 1);
    EndCodePtr = StartCodePtr + ZeroCodeLL.elementnum;

    PCCodePtr = StartCodePtr;
    CheckHobbish();
    if (err_code != passed)
        {  printf("Yazoo:  Bytecode (Zero) error \"%s\" at index %i, script %i.\n", (char *) HobbishErrorStrings[err_code],
                                        (int) err_index, (int) err_script.PLL_index);  return 1;  }


        // (Compile and) transform the starting script.

    if (if_already_compiled == yz_false) {      // case 1:  the script needs to be compiled
        scanPtr = (char *) FElement(FileReadLL, 1);
        if (retrn == passed)  retrn = Compile(&scanPtr, &CompiledCode);


            // if there's a compile-time error, flag it
            
        if (retrn != passed)  {
            char *FirstLineChar = (char *) FElement(FileReadLL, 1), *loopChar;
            Ulong lines_num = 0;
            
            
                // find the beginning & end of the line at which the error occurred
                
            loopChar = FirstLineChar;
            while (loopChar < errPtr)  {
                FirstLineChar = loopChar;
                while (lettertype(*loopChar != a_null) && ((lettertype(*loopChar) != a_eol) || (*loopChar == ',')))  loopChar++;
                lines_num++;
                
                if ((*loopChar == '\r') && (*(loopChar+1) == '\n'))  loopChar += 2;
                else  loopChar++;
            }
            
            
                // write the error message..
                
            printf("Yazoo:  Compiler returned \"%s\" at line %i in file ", CompileErrorStrings[retrn], (int) lines_num);
            if (argc == 0)  printf("start.zoo\n\n");
            else  printf("%s\n\n", argv[0]);


                // .. and the line in the script that caused it..
                
            loopChar = FirstLineChar;
            while (((lettertype(*loopChar) != a_eol) || (*loopChar == ',')) && (lettertype(*loopChar) != a_null))  {
                if (lettertype(*loopChar) != unprintable)  {
                    if (*loopChar == '\t')  printf(" ");
                    else  printf("%c", *loopChar);         }
                loopChar++;         }
            printf("\n");
            
            
                // .. and finally a marker indicating the offending character in the line.
                
            loopChar = FirstLineChar;
            while (((lettertype(*loopChar) != a_eol) || (*loopChar == ',')) && (lettertype(*loopChar) != a_null) && (loopChar < errPtr))  {
                if (lettertype(*loopChar) != unprintable)  printf(" ");
                loopChar++;         }
            printf("^\n\n");

            return 1;         }
        

            // save a copy of the NewlinePositions[] array in case we have a runtime error, so we can the line that caused it
        
        linkedLretrn = FNewLinkedList(&NewlinePosCopy, NewlinePositions.elementnum, NewlinePositions.elementsize, 0, uncleared);
        if (linkedLretrn != passed)  {  printf("Yazoo:  Memory Error:  %s\n", CompileErrorStrings[memory_err]);  return 1;  }
        FCopyElements(&NewlinePositions, 1, &NewlinePosCopy, 1, NewlinePositions.elementnum);       // since NP is overwritten by compile()
    }
    
    else  {    // case 2:  just copy the pre-compiled code into our CompiledCode list.
        linkedLretrn = NewLinkedList(&CompiledCode, CharCounter/sizeof(Slong), sizeof(Slong), 0, uncleared);
        if (linkedLretrn != passed)  {  printf("Yazoo:  Memory Error:  %s\n", CompileErrorStrings[memory_err]);  return 1;  }
        if (FileReadLL->elementnum > 0)  FSetElements(&CompiledCode, 1, CompiledCode.elementnum, FElement(FileReadLL, 1));   }
    
    
        // check the bytecode of the script (should be OK if we compiled it using the compiler)
    
    StartCodePtr = LL_SLong(&CompiledCode, 1);
    EndCodePtr = StartCodePtr + CompiledCode.elementnum;
    
    PCCodePtr = StartCodePtr;
    CheckHobbish();
    if (err_code != passed)  {
        printf("Yazoo:  Bytecode error \"%s\" at index %i, script %i in file ", HobbishErrorStrings[err_code],
                                                (int) err_index, (int) err_script.PLL_index);  return 1;
        if (argc == 0)  printf("start.zoo\n\n");
        else  printf("%s\n\n", argv[0]);
        return 1;         }
        
    if (warning_code != passed)  {
        printf("Yazoo:  Bytecode warning \"%s\" at index %i, script %i in file ", HobbishErrorStrings[warning_code],
                                                (int) warning_index, (int) warning_script.PLL_index);
        if (argc == 0)  printf("start.zoo\n\n");
        else  printf("%s\n\n", argv[0]);         }


        // Run Zero to set up Yazoo's top-level memory.

    ThatCoordinate._Bundle = NULL;
    BeginExecution(&PCCodeRef, yz_false, 0, 1, 0);    // runs Zero (PCCodeRef was set to point to Zero by InitYazoo())
    if (err_code == return_flag)  err_code = passed;
    if (err_code != passed)    // should never happen
        {  printf("Yazoo:  Zero execution error: \"%s\" at index %i, script %i.\n", HobbishErrorStrings[err_code],
                                                (int) err_index, (int) err_script.PLL_index);  return 1;  }
    DerefFibril(&(SearchCoordinate._Fibril));         // for when we check refs -- SC_F gets overwritten since BE() 2nd arg is false


        // Zero contains the original workspace (where we run our script), and the return variables for the external functions.

    Zero_workspace = ((member *) FElement(&(Zero->variable_ptr->mem.members), _workspace1))->member_ptr;


        // Now run the script inside Zero's workspace
        
    retrn = AddCode((Slong *) FElement(&CompiledCode, 1), &ScriptCodeEntryPtr, &ScriptCodeIndex, CompiledCode.elementnum);
    if (retrn != passed)  return retrn;
    retrn = AddCodeRef(&(Zero_workspace->variable_ptr->code_list), (fibril *) FElement(&(Zero_workspace->variable_ptr->fibril_list.data), 1),
                                                    ScriptCodeEntryPtr, ScriptCodeIndex);
    if (retrn != passed)  return retrn;

#ifdef YazooTest_ON
test_MM_CountRefs();
#endif

    ScriptCR = (code_ref *) FElement(&(Zero_workspace->variable_ptr->code_list), Zero_workspace->variable_ptr->code_list.elementnum);
    BeginExecution(ScriptCR, yz_false, 0, 1, 0);
    if (err_code == return_flag)  err_code = passed;
    

        // if there was an error in the script, print out the line that caused it (if possible)
        
    if ((err_code != passed) && (err_code != finished_signal))  {
        if ((err_script.PLL_index == 2) && (if_already_compiled == yz_false))  {
            Ulong ErrorPosition, loopLine;
            
            
                // search the line positions array to find the line in text corresponding to the offending bytecode where the error happened
                
            ErrorPosition = FindLine(&NewlinePosCopy, 0, err_index + (err_script.code_ptr - ScriptCR->code_ptr));
            printf("Yazoo execution error: \"%s\" in file ", HobbishErrorStrings[err_code]);
            if (argc == 0)  printf("start.zoo\n\n");
            else  printf("%s\n\n", argv[0]);

            if (ErrorPosition != 0)  {
                Ulong FirstLine = 0, ErrorOffset;
                char *BaseChar, *loopChar, *FirstLinePtr = NULL, *LastChar;

                
                    // get the end of the line so we can print it out
                
                ErrorOffset = *LL_ULong(&NewlinePosCopy, ErrorPosition) - 1;
                BaseChar = ((char *) FElement(FileReadLL, 1));

                loopChar = BaseChar;
                while (loopChar <= BaseChar+ErrorOffset)  {
                    FirstLinePtr = loopChar;
                    while (lettertype(*loopChar != a_null) && ((lettertype(*loopChar) != a_eol) || (*loopChar == ',')))  loopChar++;
                    FirstLine++;
                    
                    if ((*loopChar == '\r') && (*(loopChar+1) == '\n'))  loopChar += 2;
                    else  loopChar++;
                }
                
                if (ErrorPosition == NewlinePosCopy.elementnum)  LastChar = BaseChar + CharCounter - 1;
                else  LastChar = BaseChar + *LL_ULong(&NewlinePosCopy, ErrorPosition+1) - 2;
                
                
                    // print out the line number, and the line, where the error occurred
                    
                loopChar = FirstLinePtr;
                loopLine = FirstLine;
                while (loopChar <= LastChar)  {
                    printf("%i:  ", (int) loopLine);
                    while (((lettertype(*loopChar) != a_eol) || (*loopChar == ','))
                                && (lettertype(*loopChar) != a_null) && (loopChar <= LastChar))  {
                        if (*loopChar == '\t')  printf(" ");
                        else if (lettertype(*loopChar) != unprintable)  printf("%c", *loopChar);
                        loopChar++;     }
                    if ((*loopChar == '\r') && (*(loopChar+1) == '\n'))  loopChar += 2;
                    else  loopChar++;
                    printf("\n");
                    loopLine++;
                }
                printf("\n\n");
        }   }
        
        else  {    // we don't have the script that caused the error handy -- just do what we can
            printf("Yazoo execution error: \"%s\" at index %i, script %i in file ID %i\n\n", HobbishErrorStrings[err_code],
                                                    (int) err_index, (int) err_script.PLL_index, (int) err_script.PLL_index);      }
        main_error = 1;         }

#ifdef YazooTest_ON
test_MM_CountRefs();
#endif

    FinishCompile();
    CleanUp();
    FDeleteLinkedList(&NewlinePositions);
    FDeleteLinkedList(&SourceCode);
    FDeleteLinkedList(&ZeroCodeLL);
    FDeleteLinkedList(&CompiledCode);
    FDeleteLinkedList(&NewlinePosCopy);

    return main_error;
}
